package jimm.datavision;
import jimm.datavision.layout.*;
import jimm.datavision.layout.swing.SwingLE;
import jimm.datavision.gui.DesignWin;
import java.io.*;
import java.util.Locale;

/**
 * This is the DataVision application class. It opens a design window
 * for each report specified on the command line. If none are specified,
 * it openes a new design window on a new, empty report.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class DataVision {

protected static final String DEFAULT_CHARACTER_SEPARATOR = ",";

char layoutEngineChoice;
String dbPassword;
String paramXMLFile;
int numReports;

/**
 * This main application method opens a design window for each report
 * specified on the command line. If none are specified, it openes a new
 * design window on a new, empty report.
 *
 * @param args command line array; each element is assumed to be a report
 * file name.
 */
public static void main(String[] args) {
    Getopts g = new Getopts("cdhlxwnp:s:r:g:i:", args);
    if (g.error())		// Any bad command line argument?
	usage(null);		// If so, whine and exit

    DataVision dv = new DataVision();

    // Language
    if (g.hasOption('g') || g.hasOption('i'))
	I18N.setLanguage(new Locale(g.option('g', "").toLowerCase(),
				    g.option('i', "").toUpperCase()));

    dv.layoutEngineChoiceFromOptions(g); // Grab layout engine
    if (g.hasOption('n')) {	// No password/empty password
	if (g.hasOption('p'))
	    usage(I18N.get("DataVision.n_and_p"));
	dv.dbPassword = "";
    }
    else
	dv.dbPassword = g.option('p', null); // Grab password; default is null

    if (dv.hasLayoutEngine() && !g.hasOption('n') && !g.hasOption('p'))
	usage(I18N.get("DataVision.n_or_p"));

    dv.paramXMLFile = g.option('r', null); // Parameter XML file name or null

    if (g.argc() == 0) {
	if (dv.hasLayoutEngine()) // Have layout engine but no file
	    usage(I18N.get("DataVision.xml_req"));
	else
	    new DesignWin(null).show();	// No layout engine, no file
    }
    else {			// Loop over input files
	dv.numReports = g.argc();
	for (int i = 0; i < g.argc(); ++i) {
	    String fileName = g.argv(i);
	    try {
		if (dv.hasLayoutEngine()) {
		    Report report = new Report();
		    if (dv.dbPassword != null)
			report.setDatabasePassword(dv.dbPassword);
		    report.readFile(fileName); // Must come after password set
		    if (dv.paramXMLFile != null) // Must come after file read
			report.setParameterXMLFile(dv.paramXMLFile);
		    report.setLayoutEngine(dv.createLayoutEngine(fileName, g));
		    report.run();
		}
		else {
		    new DesignWin(fileName, dv.dbPassword).show();
		}
	    }
	    catch (Exception e) { // Global catch-all
		ErrorHandler.error(e);
	    }
	}
    }
}

boolean hasLayoutEngine() { return layoutEngineChoice != '\0'; }

void layoutEngineChoiceFromOptions(Getopts g) {    
    String errMsg = I18N.get("DataVision.le_one");

    layoutEngineChoice = '\0';
    if (g.hasOption('c'))
	layoutEngineChoice = 'c';
    if (g.hasOption('d')) {
	if (layoutEngineChoice != '\0') usage(errMsg);
	layoutEngineChoice = 'd';
    }
    if (g.hasOption('h')) {
	if (layoutEngineChoice != '\0') usage(errMsg);
	layoutEngineChoice = 'h';
    }
    if (g.hasOption('l')) {
	if (layoutEngineChoice != '\0') usage(errMsg);
	layoutEngineChoice = 'l';
    }
    if (g.hasOption('x')) {
	if (layoutEngineChoice != '\0') usage(errMsg);
	layoutEngineChoice = 'x';
    }
    if (g.hasOption('w')) {
	if (layoutEngineChoice != '\0') usage(errMsg);
	layoutEngineChoice = 'w';
    }
}

LayoutEngine createLayoutEngine(String fileName, Getopts g) throws IOException
{
    LayoutEngine le = null;

    // Create new file name.
    String fileNameSansExtension;
    int pos = fileName.lastIndexOf('.');
    if (pos == -1)
	fileNameSansExtension = new String(fileName);
    else
	fileNameSansExtension = fileName.substring(0, pos);

    String outFileName = null;
    PrintWriter out = null;
    switch (layoutEngineChoice) {
    case 'c':
	char sep = g.option('s', DEFAULT_CHARACTER_SEPARATOR).charAt(0);
	outFileName = fileNameSansExtension + (sep == ',' ? ".csv" : ".tab");
	if (outFileName.equals(fileName))
	    outFileName = fileNameSansExtension + "_out"
		+ (sep == ',' ? ".csv" : ".tab");
	out = new PrintWriter(new FileWriter(outFileName));
	le = new CharSepLE(out, sep);
	break;
    case 'd':
	outFileName = fileNameSansExtension + ".sgml";
	if (outFileName.equals(fileName))
	    outFileName = fileNameSansExtension + "_out.sgml";
	out = new PrintWriter(new FileWriter(outFileName));
	le = new DocBookLE(out);
	break;
    case 'h':
	outFileName = fileNameSansExtension + ".html";
	if (outFileName.equals(fileName))
	    outFileName = fileNameSansExtension + "_out.html";
	out = new PrintWriter(new FileWriter(outFileName));
	le = new HTMLLE(out);
	break;
    case 'l':
	outFileName = fileNameSansExtension + ".tex";
	if (outFileName.equals(fileName))
	    outFileName = fileNameSansExtension + "_out.tex";
	out = new PrintWriter(new FileWriter(outFileName));
	le = new LaTeXLE(out);
	break;
    case 'x':
	outFileName = fileNameSansExtension + ".xml";
	if (outFileName.equals(fileName))
	    outFileName = fileNameSansExtension + "_out.xml";
	IndentWriter iout = new IndentWriter(new FileWriter(outFileName));
	le = new XMLLE(iout);
	break;
    case 'w':
	le = new SwingLE() {
	    public void close() { // Override close() to possibly exit app
		super.close();
		swingLayoutEngineClosed();
	    }
	    };
	break;
    }
    return le;
}

/**
 * We arrange this method to be called by a Swing layout engine when
 * it closes. If appropriate, exit the app.
 */
protected void swingLayoutEngineClosed() {
    if (--numReports == 0)
	System.exit(0);
}

/**
 * Prints a usage message and an optional extra error message to System.err
 * and exits.
 *
 * @param errMsg string to print; may be <code>null</code>
 */
public static void usage(String errMsg) {
    if (errMsg != null)
	System.err.println(errMsg);
    System.err.println(I18N.get("DataVision.usage"));

    System.exit(1);
}

}
