package jimm.datavision;
import java.io.File;
import java.awt.MediaTracker;
import javax.swing.ImageIcon;
import javax.swing.GrayFilter;

/**
 * Represents an external image. The <code>value</code> instance value
 * stores the images file's path.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class ImageField extends Field {

public static final String TYPE_STRING = "image";

protected File imageFile;
protected ImageIcon imageIcon;
protected ImageIcon hiddenImageIcon;

/**
 * Constructor. We make sure the value (a file path) is an absolute file
 * path.
 *
 * @param id the unique identifier for the new field
 * @param report the report containing this line
 * @param section the section containing this line
 * @param value the value; a file path string
 * @param visible show/hide flag
 */
public ImageField(Long id, Report report, Section section, Object value,
		  boolean visible)
{
    super(id, report, section, value, visible);

    if (value != null) {	// Ensure path is absolute
	String path = (String)value;
	File f = new File(path);
	value = f.getAbsolutePath();
    }
}

public File getImageFile() {
    if (imageFile == null && value != null) {
	imageFile = new File((String)value);
	imageIcon = null;
	hiddenImageIcon = null;
    }
    return imageFile;
}

public ImageIcon getImageIcon() {
    if (isVisible())
	return getVisibleImageIcon();
    else
	return getHiddenImageIcon();
}

public ImageIcon getVisibleImageIcon() {
    if (imageIcon == null && value != null)
	imageIcon = new ImageIcon(getImageFile().getPath());
    return imageIcon;
}

public ImageIcon getHiddenImageIcon() {
    if (hiddenImageIcon == null && value != null && canLoad()) {
	ImageIcon ii = getVisibleImageIcon();
	if (ii != null) {
	    hiddenImageIcon =
		new ImageIcon(GrayFilter.createDisabledImage(ii.getImage()));
	}
    }
    return hiddenImageIcon;
}

public void setValue(Object newValue) {
    imageFile = null;
    imageIcon = null;

    if (newValue != null) {	// Ensure path is absolute
	String path = (String)newValue;
	File f = new File(path);
	newValue = f.getAbsolutePath();
    }

    super.setValue(newValue);	// Notify observers
}



public boolean canLoad() {
    return getVisibleImageIcon() != null
	&& getVisibleImageIcon().getImageLoadStatus() == MediaTracker.COMPLETE;
}

public String dragString() {
    return typeString() + ":" + value;
}

public String typeString() { return TYPE_STRING; }

public String formulaString() { return "{" + value + "}"; }

}
