package jimm.datavision;
import java.io.PrintWriter;
import java.io.Writer;
import java.io.OutputStream;

/**
 * An indent writer is a print writer that outputs indentation whitespace
 * after each newline. Calling <code>indent</code> and <code>outdent</code>
 * changes the indentation level. The default indentation width is 4
 * spaces.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class IndentWriter extends PrintWriter {

protected static final int DEFAULT_INDENTATION_WIDTH = 4;

protected int width;
protected int level;
protected boolean newline;

/**
 * Constructor, same as the <code>PrintWriter</code> version.
 *
 * @param out an output stream
 */
public IndentWriter(OutputStream out) {
    super(out);
    init(DEFAULT_INDENTATION_WIDTH);
}

/**
 * Constructor, same as <code>PrintWriter</code> version without the
 * <i>width</i> parameter.
 *
 * @param out an output stream
 * @param autoFlush if <code>true</code>, the <code>println()</code> methods
 * will flush the output bufferset to flush after every line
 * @param width indentation width in spaces
 */
public IndentWriter(OutputStream out, boolean autoFlush, int width) {
    super(out, autoFlush);
    init(width);
}

/**
 * Constructor, same as the <code>PrintWriter</code> version.
 *
 * @param out a writer
 */
public IndentWriter(Writer out) {
    super(out);
    init(DEFAULT_INDENTATION_WIDTH);
}

/**
 * Constructor, same as the <code>PrintWriter</code> version.
 *
 * @param out a writer
 * @param autoFlush if <code>true</code>, the <code>println()</code> methods
 * will flush the output bufferset to flush after every line
 * @param width indentation width in spaces
 */
public IndentWriter(Writer out, boolean autoFlush, int width) {
    super(out, autoFlush);
    init(width);
}

/**
 * Initializes some instance variables. Called from constructors.
 *
 * @param indentationWidth number of spaces per indentation level
 */
protected void init(int indentationWidth) {
    width = indentationWidth;
    level = 0;
    newline = true;
}

/**
 * Increases the indentation level by one.
 */
public void indent() { ++level; }

/**
 * Decreases the indentation level by one.
 */
public void outdent() { if (--level < 0) level = 0; }

public void print(boolean b) { doIndent(); super.print(b); }
public void print(char c) {
    doIndent();
    super.print(c);
    if (c == '\n') newline = true;
}
public void print(char[] s) {
    for (int i = 0; i < s.length; ++i) print(s[i]);
}
public void print(double d) { doIndent(); super.print(d); }
public void print(float f) { doIndent(); super.print(f); }
public void print(int i) { doIndent(); super.print(i); }
public void print(long l) { doIndent(); super.print(l); }
public void print(Object obj) { print(obj.toString()); }

/**
 * This method does not handle newlines embedded in the string.
 *
 * @param str the string to output
 */
public void print(String str) { doIndent(); super.print(str); }

public void println() {
    super.println();
    newline = true;
}
public void println(boolean b) { doIndent(); super.println(b); }
public void println(char c) {
    doIndent();
    super.println(c);
    if (c == '\n') {
	newline = true;
	doIndent();
	newline = true;
    }
}
public void println(char[] s) {
    for (int i = 0; i < s.length; ++i) print(s[i]);
    println();
}
public void println(double d) { doIndent(); super.println(d); newline = true; }
public void println(float f) { doIndent(); super.println(f); newline = true; }
public void println(int i) { doIndent(); super.println(i); newline = true; }
public void println(long l) { doIndent(); super.println(l); newline = true; }
public void println(Object obj) { println(obj.toString()); }
// FIX: this is not correct, but it is not worth fixing right now.
// It does not handle newlines embedded in the string, but I do not care.
public void println(String str) {
    doIndent();
    super.println(str);
    newline = true;
}

/**
 * Performs indentation by printing the correct number of tabs and spaces.
 */
protected void doIndent() {
    if (newline) {
	int spaces = level * width;
	while (spaces >= 8) {
	    super.print("\t");
	    spaces -= 8;
	}
	super.print("        ".substring(0, spaces));
	newline = false;
    }
}

}
