package jimm.datavision;
import java.util.Collection;
import java.util.Observer;
import java.util.Observable;

/**
 * A parameter field represents a parameter. The value of a parameter field
 * holds a parameter object. In the XML, the parameter field's value is
 * the id of the parameter.
 *
 * @see Parameter
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class ParameterField extends Field implements Observer {

protected Parameter parameter;

/**
 * Constructs a parameter field with the specified id in the specified report
 * section whose parameter's id is <i>value</i>.
 *
 * @param id the new field's id
 * @param report the report containing this element
 * @param section the report section in which the field resides
 * @param value the id of a parameter
 * @param visible show/hide flag
 * @see Parameter
 */
public ParameterField(Long id, Report report, Section section, Object value,
		      boolean visible)
{
    super(id, report, section, value, visible);
    parameter = report.findParameter(value);
    parameter.addObserver(this);
}

public void finalize() {
    parameter.deleteObserver(this);
}

public void update(Observable o, Object arg) {
    setChanged();
    notifyObservers(arg);
}

/**
 * Not really used; we drag parameters, not parameter fields.
 */
public String dragString() {
    return typeString() + ":" + parameter.getId();
}

/**
 * Returns the parameter.
 *
 * @return the parameter
 */
public Parameter getParameter() { return parameter; }

/**
 * Sets the parameter.
 *
 * @param newParameter the new parameter
 */
public void setParameter(Parameter newParameter) {
    if (parameter != newParameter) {
	parameter.deleteObserver(this);
	parameter = newParameter;
	parameter.addObserver(this);
	setChanged();
	notifyObservers();
    }
}

public String typeString() { return "parameter"; }

public String designLabel() { return parameter.designLabel(); }

public String formulaString() { return parameter.formulaString(); }

public boolean refersTo(Parameter p) {
    return parameter == p;
}

/**
 * Returns the value of this field. For parameter fields, this is the
 * value generated by evaluating the parameter.
 *
 * @return the result of evaluating the parameter
 * @see Parameter
 */
public Object getValue() { return parameter.getValue(); }

}
