package jimm.datavision;
import jimm.datavision.sql.Table;
import jimm.datavision.sql.Column;
import java.util.*;
import java.sql.PreparedStatement;
import java.sql.SQLException;

/**
 * Queries build SQL query strings. They contain tables, joins, and
 * where clauses.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class Query implements Writeable {

protected Report report;
protected ArrayList joins;
protected String whereClause;
protected ArrayList sortCols;
protected ArrayList sortOrders;
protected Set tables;
protected ArrayList columns;
protected ArrayList usercols;
protected ArrayList preparedStmtValues;

/**
 * Constructor.
 *
 * @param report the report for which this query will generate SQL
 */
public Query(Report report) {
    this.report = report;
    joins = new ArrayList();
    whereClause = null;
    sortCols = new ArrayList();
    sortOrders = new ArrayList();
    tables = new HashSet();
    columns = new ArrayList();
    usercols = new ArrayList();
}

/**
 * Returns <code>true</code> if the specified parameter exists within this
 * query's where clause.
 *
 * @param p a parameter
 * @return <code>true</code> if the specified parameter exists within
 * the where clause
 */
public boolean containsReferenceTo(Parameter p) {
    return whereClause != null && whereClause.indexOf(p.formulaString()) != -1;
}

/**
 * Adds a join to the list of joins used by this query.
 *
 * @param join a join
 */
public void addJoin(Join join) {
    joins.add(join);
}

/**
 * Adds all joins in the collection to our list.
 *
 * @param coll a collection of joins
 */
public void addAllJoins(Collection coll) {
    joins.addAll(coll);
}

/**
 * Removes a join from the list of joins used by this query.
 *
 * @param join a join
 */
public void removeJoin(Join join) {
    joins.remove(join);
}

/**
 * Removes all joins in the collection from our list.
 */
public void clearJoins() {
    joins.clear();
}

/**
 * Returns an iterator over all the joins used by this query.
 *
 * @return an iterator over all the joins
 */
public Iterator joins() { return joins.iterator(); }

/**
 * Returns the where clause fit for human consumption. This mainly means
 * that we substitute formula, parameter, and user column id numbers with
 * names. Called from a where clause editor. This code assumes that curly
 * braces are never nested.
 *
 * @return the eval string with formula, parameter, and user column id
 * numbers replaced with names
 * @see jimm.datavision.gui.WhereClauseWin
 */
public String getEditableWhereClause() {
    return Utils.formulaToDisplay(report, getWhereClause());
}

/**
 * Returns the raw where clause string; may be <code>null</code>.
 *
 * @return the where clause string; may be <code>null</code>
 * @see #getWhereClauseForPreparedStatement
 */
public String getWhereClause() {
    return whereClause;
}

/**
 * Returns the where clause string; may be <code>null</code>. If there
 * are any parameter values, we return '?' in their place and save the
 * values for later use.
 * <p>
 * This code may also modify the clause. For example, a parameter can
 * change the previous comparison operator ("=", "is") based on its arity.
 *
 * @return the where clause string; may be <code>null</code>
 * @see #getWhereClause
 */
protected String getWhereClauseForPreparedStatement() {
   if (whereClause == null)
	return null;
    if (whereClause.indexOf("{") == -1)
	return whereClause;
    return prepare(whereClause);
}

/**
 * Given a clause (really any string), replace all formulas and parameters
 * with their values. Everything else in curly braces is 
 */
protected String prepare(String clause) {
    StringBuffer buf = new StringBuffer();

    int pos, endPos;
    for (pos = 0, endPos = -1;
	 (pos = clause.indexOf("{", endPos + 1)) >= 0;
	 pos = endPos + 1)
    {
	int oldEndPos = endPos;
	endPos = clause.indexOf("}", pos);
	if (endPos == -1) {
	    buf.append(clause.substring(pos));
	    break;
	}

	switch (clause.charAt(pos + 1)) {
	case '@':		// Formula
	    String idAsString = clause.substring(pos + 2, endPos);
	    preparedStmtValues.add(report.findFormula(idAsString).eval());

	    buf.append(clause.substring(oldEndPos + 1, pos));
	    buf.append(" ? ");
	    break;
	case '?':		// Parameter
	    idAsString = clause.substring(pos + 2, endPos);

	    // Find previous word
	    ParserHelper ph = new ParserHelper(clause, pos);

	    // Append prev text without previous token
	    buf.append(clause.substring(oldEndPos + 1,
					     ph.prevTokenStartPos()));

	    // Append possibly modified previous token and parameter
	    addParameter(buf, ph.prevSQLToken(), idAsString);
	    break;
	default:		// Column field; remove curlies
	    buf.append(clause.substring(oldEndPos + 1, pos));
	    buf.append(clause.substring(pos + 1, endPos));
	    break;
	}
	pos = endPos + 1;
    }

    if ((endPos + 1) < clause.length())
	buf.append(clause.substring(endPos + 1));

    return buf.toString();
}

/**
 * Returns a user column ready for use in a select clause. If the user
 * column contains any parameter values, we return '?' in their place
 * and save the values for later use. May return <code>null</code>.
 *
 * @return the user column as a string; may return <code>null</code>
 */
public String userColumnForPreparedStatement(UserColumn uc) {
    if (uc == null)
	return null;
    String ucString = uc.getUserString();
    if (ucString == null)
	return null;
    if (ucString.indexOf("{") == -1)
	return ucString;
    return prepare(ucString);
}

/**
 * Given a parameter id string, add its value(s) to the parameter list
 * and add prepared statement placeholders in the query buffer. Appends
 * the previous word to the buffer. The previous word may be modified
 * if the circumstances call for it. For example, we want to turn
 * "foo in {?Range Parameter}" into "foo between ? and ?". The value
 * of <var>prevWord</var> here would be "in". We would append "between"
 * to the buffer and return the "word" "? and ".
 *
 * @param buf a string buffer containing the SQL query so far
 * @param prevWord the previous word
 * @param idAsString the parameter id
 * @return a new "previous word"
 */
protected void addParameter(StringBuffer buf, String prevWord,
			    String idAsString)
{
    String word = null;
    Long paramId = new Long(idAsString);
    Parameter param = report.findParameter(paramId);

    // Ask report for parameter value so the report can ask the user to
    // fill in the parameter's value.
    Object val = report.getParameterValue(paramId);
    if (val instanceof List) {
	List list = (List)val;
	if (param.getArity() == Parameter.ARITY_RANGE) {
	    // Modify prev word
	    if ("!=".equals(prevWord) || "<>".equals(prevWord))
		buf.append(" not between ");
	    else if ("=".equals(prevWord)
		     || "in".equals(prevWord.toLowerCase()))
		buf.append(" between ");
	    else {
		buf.append(' ');
		buf.append(prevWord);
		buf.append(' ');
	    }

	    word = "? and ?";
	    preparedStmtValues.add(list.get(0));
	    preparedStmtValues.add(list.get(1));
	}
	else {			// Build "(a,b,c)" list
	    switch (list.size()) {
	    case 0:		// No items in list; "is null"
		buf.append(" is null");
		break;
	    case 1:		// One item in list; simple equality
		if ("in".equals(prevWord) || "<>".equals(prevWord))
		    buf.append(" = ");
		else {
		    buf.append(' ');
		    buf.append(prevWord);
		    buf.append(' ');
		}
		word = "?";
		preparedStmtValues.add(list.get(0));
		break;
	    default:
		if ("!=".equals(prevWord) || "<>".equals(prevWord))
		    buf.append(" not in ");
		else if ("=".equals(prevWord)
			 || "in".equals(prevWord.toLowerCase()))
		    buf.append(" in ");
		else {
		    buf.append(' ');
		    buf.append(prevWord);
		    buf.append(' ');
		}

		StringBuffer wordBuf = new StringBuffer("(");
		boolean first = true;
		int len = list.size();
		for (int i = 0; i < len; ++i) {
		    if (first) first = false;
		    else wordBuf.append(',');
		    wordBuf.append('?');
		}
		wordBuf.append(")");
		word = wordBuf.toString();
		preparedStmtValues.addAll(list);
	    }
	}
    }
    else {
	buf.append(' ');
	buf.append(prevWord);	// Previous word
	preparedStmtValues.add(val);
	word =" ?";		// For prepared statement
    }

    buf.append(word);
}

/**
 * Sets the where clause (may be <code>null</code>). The string passed
 * to us contains parameter and formula display strings, not their
 * "real" <code>formulaString</code> representations. Translate the latter
 * into the former before saving this string.
 *
 * @param newWhereClause a where clause string; may be <code>null</code>
 */
public void setEditableWhereClause(String newWhereClause) {
    setWhereClause(Utils.displayToFormula(report, newWhereClause));
}

/**
 * Sets the where clause (may be <code>null</code>).
 *
 * @param newWhereClause a where clause string; may be <code>null</code>
 */
public void setWhereClause(String newWhereClause) {
    whereClause = newWhereClause;
}

/**
 * Adds a column with the default sort order (descending).
 *
 * @param col a database column
 */
public void addSort(Column col) {
    addSort(col, "desc");
}

/**
 * Adds a sort order for the specified column. The first character
 * of the <i>order</i> string is inspected. If it is a 'd' or 'D',
 * then it's taken to mean "descending". Anything else will result
 * in an "ascending" sort order.
 *
 * @param col a database column
 * @param order a string
 */
public void addSort(Column col, String order) {
    sortCols.add(col);
    sortOrders.add(order.toLowerCase().charAt(0) == 'd' ? "desc" : "asc");
}

/**
 * Removes a sorting from the list.
 *
 * @param clause a where clause
 */
public void removeSort(Column col) {
    // I used to use a map, but then sorts would not keep their order,
    // which is important. That's why this code isn't just a map lookup
    // any more.
    for (int i = 0; i < sortCols.size(); ++i) {
	if (sortCols.get(i) == col) {
	    sortCols.remove(i);
	    sortOrders.remove(i);
	    return;
	}
    }
}

/**
 * Removes all sorts from our list.
 */
public void clearSorts() {
    sortCols = new ArrayList();
    sortOrders = new ArrayList();
}

/**
 * Returns an iterator over all the sorted columns used by this query.
 *
 * @return an iterator
 */
public Iterator sortedColumns() { return sortCols.iterator(); }

/**
 * Returns the sort order string ("asc" or "desc") of the specified column.
 * If the column is not sorted, returns <code>null</code>
 *
 * @param col a database column
 * @return the string "asc" or "desc"; <code>null</code> if the column is
 * not on the list of sorted columns.
 */
public String sortOrderOf(Column col) {
    // I used to use a map, but then sorts would not keep their order,
    // which is important. That's why this code isn't just a map lookup
    // any more.
    for (int i = 0; i < sortCols.size(); ++i) {
	if (sortCols.get(i) == col)
	    return (String)sortOrders.get(i);
    }
    return null;
}

/**
 * Returns the index of the specified column in the select statement.
 *
 * @param column a database column
 */
public int indexOfColumn(Column column) { return columns.indexOf(column); }

/**
 * Returns the index of the specified user column in the select statement.
 *
 * @param user a user column
 */
public int indexOfColumn(UserColumn uc) {
    return columns.size() + usercols.indexOf(uc);
}

/**
 * Builds collections of the report tables and column fields actually used
 * in the report.
 */
public void findColumnsUsed() {
    columns.clear();
    tables.clear();
    usercols.clear();
    report.withFieldsDo(new FieldWalker() {
	public void step(Field f) {
	    if (f instanceof ColumnField) {
		Column col = ((ColumnField)f).getColumn();
		if (!columns.contains(col)) columns.add(col);
		tables.add(col.getTable());
	    }
	    else if (f instanceof FormulaField) {
		FormulaField ff = (FormulaField)f;
		for (Iterator iter = ff.columnsUsed().iterator();
		     iter.hasNext(); )
		{
		    Column col = (Column)iter.next();
		    if (!columns.contains(col)) columns.add(col);
		    tables.add(col.getTable());
		}
	    }
	    else if (f instanceof UserColumnField) {
		UserColumn uc = ((UserColumnField)f).getUserColumn();
		if (!usercols.contains(uc)) usercols.add(uc);
	    }
	}
	});

    // Add all columns and tables used in joins.
    for (Iterator iter = joins.iterator(); iter.hasNext(); ) {
	Join join = (Join)iter.next();
	if (!columns.contains(join.from)) columns.add(join.from);
	if (!columns.contains(join.to)) columns.add(join.to);
	tables.add(join.from.getTable());
	tables.add(join.to.getTable());
    }

    // Add all columns and tables used in where clauses. This is difficult
    // since the where clause is a string. We assume that anything of the
    // form "foo.bar" is a table.column name.
// FIX

    // Add all columns and tables used in sorts.
    for (Iterator iter = sortedColumns(); iter.hasNext(); ) {
	Column col = (Column)iter.next();
	if (!columns.contains(col)) columns.add(col);
	tables.add(col.getTable());
    }
}

/**
 * Returns the number of columns in the query. Does not recalculate the
 * columns used; we assume this is being called after the qurey has been
 * run.
 */
public int getNumColumns() { return columns.size() + usercols.size(); }

/**
 * Returns a collection containing the tables used by this query.
 *
 * @return the collection of tables used by this query
 */
public Collection getTablesUsed() {
    findColumnsUsed();
    return tables;
}

/**
 * Returns the query as a human-readable SQL statement, including parameter,
 * formula, and user column display strings.
 *
 * @return a SQL query string
 */
public String toString() {
    return queryAsString(true);
}

/**
 * Returns the query as a SQL string suitable for building a prepared
 * statement.
 *
 * @return a SQL query string
 */
public String toPreparedStatementString() {
    preparedStmtValues = new ArrayList();
    return queryAsString(false);
}

/**
 * Returns the query as either a human-readable SQL statement or a SQL
 * string suitable for building a prepared statement.
 *
 * @param forDisplay if <code>true</code> return a human-readable string,
 * else return a SQL string suitable for building a prepared statement
 * @return a SQL string
 */
protected String queryAsString(boolean forDisplay) {
    // Rebuild collections of tables, columns, and user columns every time
    // (not just first time) since the list of columns we want to use may
    // have changed since last time.
    findColumnsUsed();

    if (tables.size() == 0 || columns.size() == 0)
	return "";

    // Build list of database columns and user columns
    StringBuffer str = new StringBuffer("select ");
    ArrayList selectCols = new ArrayList(columns);
    for (Iterator iter = usercols.iterator(); iter.hasNext(); ) {
	UserColumn uc = (UserColumn)iter.next();
	String prepped = userColumnForPreparedStatement(uc);
	if (prepped != null)
	    selectCols.add(prepped);
    }
    str.append(Utils.join(selectCols, ", "));

    str.append(" from ");
    boolean first = true;
    for (Iterator iter = tables.iterator(); iter.hasNext(); ) {
	if (first) first = false;
	else str.append(", ");
	str.append(((Table)iter.next()).getName());
    }

    if (!joins.isEmpty() || (whereClause != null && whereClause.length() > 0))
    {
	str.append(" where ");
	if (!joins.isEmpty()) {
	    str.append("(");
	    str.append(Utils.join(joins, ") and ("));
	    str.append(")");
	}
	if (whereClause != null && whereClause.length() > 0) {
	    if (!joins.isEmpty())
		str.append(" and ");

	    str.append("(");
	    if (forDisplay)
		str.append(getEditableWhereClause());
	    else {
		// Call getWhereClauseForPreparedStatement so parameter
		// values are substituted and saved.
		str.append(getWhereClauseForPreparedStatement());
	    }
	    str.append(")");
	}
    }

    if (!report.groups.isEmpty() || !sortCols.isEmpty()) {
	str.append(" order by ");
	ArrayList orders = new ArrayList();
	for (Iterator iter = report.groups.iterator(); iter.hasNext(); ) {
	    Group g = (Group)iter.next();
	    StringBuffer buf = new StringBuffer(g.column.fullName());
	    switch (g.sortOrder) {
	    case Group.SORT_ASCENDING: buf.append(" asc"); break;
	    case Group.SORT_DESCENDING: buf.append(" desc"); break;
	    }
	    orders.add(buf.toString());
	}
	for (Iterator iter = sortedColumns(); iter.hasNext(); ) {
	    Column col = (Column)iter.next();
	    orders.add(col.fullName() + " " + sortOrderOf(col));
	}

	str.append(Utils.join(orders, ", "));
    }

    return str.toString();
}

/**
 * Given a prepared statement created with the text returned by
 * <code>toPreparedStatementString</code>, plug in all the parameter
 * and formula values.
 *
 * @see #toPreparedStatementString
 */
public void setParameters(PreparedStatement stmt) throws SQLException {
    int i = 1;
    for (Iterator iter = preparedStmtValues.iterator(); iter.hasNext(); ++i) {
	// In Oracle, Java Dates are turned into timestamps, or something
	// like that. This is an attempt to fix this problem.
	Object val = iter.next();
	if (val instanceof java.util.Date)
	    stmt.setDate(i,
			 new java.sql.Date(((java.util.Date)val).getTime()));
	else
	    stmt.setObject(i, val);
    }
}

/**
 * Writes this query as an XML tag. Writes all joins, where clauses,
 * and sorts as well.
 *
 * @param out a writer that knows about indenting XML tags
 */
public void writeXML(IndentWriter out) {
    out.println("<query>");
    out.indent();

    for (Iterator iter = joins(); iter.hasNext(); )
	((Join)iter.next()).writeXML(out);

    if (whereClause != null && whereClause.length() > 0)
	out.println("<where><![CDATA[" + whereClause + "]]></where>");

    for (int i = 0; i < sortCols.size(); ++i) {
	out.println("<sort order=\"" + sortOrders.get(i) + "\" column=\""
		    + ((Column)sortCols.get(i)).getId() + "\" />");
    }

    out.outdent();
    out.println("</query>");
}


/**
 * A helper class used while parsing the WHERE clause.
 */
class ParserHelper {

String str;
int startPos;
String prevToken;
int prevTokenStartPos;

ParserHelper(String s, int pos) {
    str = s;
    startPos = pos;

    findPreviousSQLToken();
}

/**
 * Finds the token before the one that starts at <var>startPos</var> in
 * <var>str</var>.
 *
 * @param str a string
 * @param the start of a word in the string
 * @return the token before the word that starts at <var>pos</var>
 */
protected void findPreviousSQLToken() {
    if (startPos == 0) {
	prevToken = "";
	prevTokenStartPos = 0;
	return;
    }

    prevTokenStartPos = startPos - 1;
    while (prevTokenStartPos >= 0
	   && Character.isSpaceChar(str.charAt(prevTokenStartPos)))
	--prevTokenStartPos;

    int tokenEnd = prevTokenStartPos + 1;

    char c = str.charAt(prevTokenStartPos);
    boolean prevWordIsAlpha = (Character.isLetterOrDigit(c) || c == '_');

    while (prevTokenStartPos >= 0
	   && !Character.isSpaceChar(c = str.charAt(prevTokenStartPos))
	   && (prevWordIsAlpha ? (Character.isLetterOrDigit(c) || c == '_')
	       : !(Character.isLetterOrDigit(c) || c == '_')))
	--prevTokenStartPos;
    ++prevTokenStartPos;

    prevToken = str.substring(prevTokenStartPos, tokenEnd);

    // Make prevTokenStartPos point to last non-space character before
    // the start of the previous token.
    --prevTokenStartPos;
    while (prevTokenStartPos >= 0
	   && Character.isSpaceChar(str.charAt(prevTokenStartPos)))
	--prevTokenStartPos;
    ++prevTokenStartPos;
}

String prevSQLToken() { return prevToken; }

int prevTokenStartPos() { return prevTokenStartPos; }

}

}
