package jimm.datavision;
import jimm.datavision.sql.*;
import jimm.datavision.gui.DbPasswordDialog;
import jimm.datavision.gui.DesignWin;
import jimm.datavision.gui.StatusDialog;
import jimm.datavision.gui.parameter.ParamAskWin;
import java.io.*;
import java.util.*;
import java.sql.*;
import javax.swing.JOptionPane;
import org.xml.sax.SAXException;

/**
 * A report holds database information, accepts parameters from the user,
 * runs a query, and uses a layout engine to format the output. It may
 * contain many different sections, each of which can contain logic for
 * surpressing itself.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class Report implements Nameable, Writeable {

/**
 * The string to use when reading and writing XML files. Pass it to
 * OutputStreamWriters and InputStreamReaders.
 */
public static final String XML_JAVA_ENCODING = "UTF8";
/**
 * The string to write at the top of an XML file in the XML decl.
 */
public static final String XML_ENCODING_ATTRIBUTE = "UTF-8";

public static final int AREA_REPORT_HEADER = 0;
public static final int AREA_REPORT_FOOTER = 1;
public static final int AREA_PAGE_HEADER = 2;
public static final int AREA_PAGE_FOOTER = 3;
public static final int AREA_DETAIL = 4;
public static final int AREA_GROUP_HEADER = 5;
public static final int AREA_GROUP_FOOTER = 6;

protected static final double OUTPUT_DTD_VERSION = 0.5;

protected String name;
protected String title;
protected String author;
protected String description;
protected Database database;
protected HashMap formulas;
protected HashMap parameters;
protected HashMap usercols;
protected ArrayList reportHeaders;
protected ArrayList reportFooters;
protected ArrayList pageHeaders;
protected ArrayList pageFooters;
protected ArrayList groups;
protected ArrayList details;
protected Row rset;
protected Query query;
protected LayoutEngine layoutEngine;
protected PaperFormat paperFormat;
protected Collection subtotalFields;
protected String databasePassword;
protected boolean askedForParameters;
protected boolean parametersHaveValues;
protected String paramXMLFileName;

/**
 * Constructs an empty report.
 */
public Report() {
    init();
    initEmptyReport();
}

/**
 * Constructs a report by reading the specified XML file. If the user was
 * prompted for a database password and that dialog box was cancelled,
 * we set our database object to <code>null</code>. This is checked in the
 * <code>jimm.datavision.gui.DesignWin</code> constructor.
 *
 * @param xmlFileName file path to report XML file
 * @see jimm.datavision.gui.DesignWin#DesignWin
 * @see jimm.datavision.gui.DbPasswordDialog
 */
public Report(String xmlFileName) {
    init();
    if (xmlFileName != null) {
	try {
	    readFile(xmlFileName);
	}
	catch (UserCancellationException uce) {
	    database = null;
	}
	catch (Exception e) {
	    ErrorHandler.error(e);
	    initEmptyReport();
	}
    }
    else
	initEmptyReport();
}

/**
 * Initializes the report. Called from constructors.
 */
private void init() {
    query = new Query(this);
    formulas = new HashMap();
    parameters = new HashMap();
    usercols = new HashMap();
    reportHeaders = new ArrayList();
    reportFooters = new ArrayList();
    pageHeaders = new ArrayList();
    pageFooters = new ArrayList();
    groups = new ArrayList();
    details = new ArrayList();
    name = I18N.get("Report.default_name");
    title = I18N.get("Report.default_title");
    askedForParameters = false;
    parametersHaveValues = false;
}

/**
 * Initialize an empty report. Called from constructors.
 */
protected void initEmptyReport() {
    reportHeaders.add(new Section(this, I18N.get("Report.report_header"),
				  false));
    pageHeaders.add(new Section(this, I18N.get("Report.page_header"),
				false));
    pageFooters.add(new Section(this, I18N.get("Report.page_footer"),
				false));
    reportFooters.add(new Section(this, I18N.get("Report.report_footer"),
				  false));
    details.add(new Section(this, I18N.get("Report.detail"), true));
    paperFormat = PaperFormat.getDefault();
}

/**
 * Sets the layout engine to use.
 *
 * @param layoutEngine a layout engine
 */
public void setLayoutEngine(LayoutEngine layoutEngine) {
    this.layoutEngine = layoutEngine;
    this.layoutEngine.report = this;
}

/**
 * Generates and returns a new unique formula id number.
 *
 * @return a long id
 */
public Long generateNewFormulaId() {
    long max = 0;
    for (Iterator iter = formulas.values().iterator(); iter.hasNext(); ) {
	long id = ((Formula)iter.next()).getId().longValue();
	if (id > max)
	    max = id;
    }
    return new Long(max + 1);
}

/**
 * Generates and returns a new unique parameter id number.
 *
 * @return a long id
 */
public Long generateNewParameterId() {
    long max = 0;
    for (Iterator iter = parameters.values().iterator(); iter.hasNext(); ) {
	long id = ((Parameter)iter.next()).getId().longValue();
	if (id > max)
	    max = id;
    }
    return new Long(max + 1);
}

/**
 * Generates and returns a new unique user column id number.
 *
 * @return a long id
 */
public Long generateNewUserColumnId() {
    long max = 0;
    for (Iterator iter = usercols.values().iterator(); iter.hasNext(); ) {
	long id = ((UserColumn)iter.next()).getId().longValue();
	if (id > max)
	    max = id;
    }
    return new Long(max + 1);
}

/**
 * Given an id, returns the field that has that id. If no field with the
 * specified id exists, returns <code>null</code>.
 *
 * @return a field, or <code>null</code> if no field with the specified
 * id exists
 */
public Field findField(Object id) {
    Section s;
    Field f;

    for (Iterator iter = reportHeaders.iterator(); iter.hasNext(); )
	if ((f = ((Section)iter.next()).findField(id)) != null)
	    return f;

    for (Iterator iter = pageHeaders.iterator(); iter.hasNext(); )
	if ((f = ((Section)iter.next()).findField(id)) != null)
	    return f;

    for (Iterator iter = groups.iterator(); iter.hasNext(); )
	for (Iterator iter2 = ((Group)iter.next()).headerSections.iterator();
	     iter2.hasNext(); )
	    if ((f = ((Section)iter2.next()).findField(id)) != null)
		return f;

    for (Iterator iter = details.iterator(); iter.hasNext(); )
	if ((f = ((Section)iter.next()).findField(id)) != null)
	    return f;

    for (Iterator iter = groups.iterator(); iter.hasNext(); )
	for (Iterator iter2 = ((Group)iter.next()).footerSections.iterator();
	     iter2.hasNext(); )
	    if ((f = ((Section)iter2.next()).findField(id)) != null)
		return f;

    for (Iterator iter = reportFooters.iterator(); iter.hasNext(); )
	if ((f = ((Section)iter.next()).findField(id)) != null)
	    return f;

    for (Iterator iter = pageFooters.iterator(); iter.hasNext(); )
	if ((f = ((Section)iter.next()).findField(id)) != null)
	    return f;

    return null;
}

public String getName() { return name; }
public void setName(String newName) { name = newName; }

public String getTitle() { return title; }
public void setTitle(String newTitle) { title = newTitle; }

public String getAuthor() { return author; }
public void setAuthor(String newAuthor) { author = newAuthor; }

public String getDescription() { return description; }
public void setDescription(String newDescription) {
    description = newDescription;
}

public Database getDatabase() { return database; }
public void setDatabase(Database newDatabase) { database = newDatabase; }

/**
 * Given an id (a table name), returns the table that has that id. If no
 * table with the specified id exists, returns <code>null</code>. Calls
 * <code>Database.findTable</code>.
 *
 * @return a table, or <code>null</code> if no table with the specified
 * id exists
 * @see Database#findTable
 */
public Table findTable(String id) { return database.findTable(id); }

/**
 * Given an id (a column name), returns the column that has that id. If no
 * column with the specified id exists, returns <code>null</code>. Calls
 * <code>Database.findColumn</code>.
 *
 * @return acolumn, or <code>null</code> if no column with the specified
 * id exists
 * @see Database#findColumn
 */
public Column findColumn(String id) { return database.findColumn(id); }

public Query getQuery() { return query; }
public void setQuery(Query newQuery) { query = newQuery; }

public PaperFormat getPaperFormat() { return paperFormat; }
public void setPaperFormat(PaperFormat newPaperFormat) {
    paperFormat = newPaperFormat;
}

/**
 * Figure out what <var>obj</var> is and add it.
 */
public void add(Object obj) {
    if (obj instanceof Parameter) addParameter((Parameter)obj);
    else if (obj instanceof Formula) addFormula((Formula)obj);
    else if (obj instanceof UserColumn) addUserColumn((UserColumn)obj);
    else if (obj instanceof Group) addGroup((Group)obj);
    else			// Shouldn't happen
	ErrorHandler.error(I18N.get("Report.add_err_1") + ' '
			   + obj.getClass().getName() + ' '
			   + I18N.get("Report.add_err_2"));
}

/**
 * Figure out what <var>obj</var> is and remove it.
 */
public void remove(Object obj) {
    if (obj instanceof Field)
	removeField((Field)obj);
    else if (obj instanceof Formula)
	removeFormula((Formula)obj);
    else if (obj instanceof Parameter)
	removeParameter((Parameter)obj);
    else if (obj instanceof UserColumn)
	removeUserColumn((UserColumn)obj);
    else if (obj instanceof Group)
	removeGroup((Group)obj);
    else if (obj instanceof Section)
	removeSection((Section)obj);
    else
	ErrorHandler.error(I18N.get("Report.remove_err_1") + ' '
			   + obj.getClass().getName()
			   + I18N.get("Report.remove_err_2"));
}

// ---------------- parameters

/**
 * Returns the parameter with the specified id or <code>null</code>
 * if one is not found. <em>Note</em>: don't use this method if you need
 * a parameter's value. That value is supplied by the user. Call
 * <code>Report.getParameterValue</code> instead, which asks the user to
 * supply the value.
 *
 * @param id the parameter id
 * @return the parameter with that id or <code>null</code> if one is not found
 * @see Report#getParameterValue
 */
public Parameter findParameter(Object id) {
    if (id instanceof String)
	id = new Long((String)id);
    return (Parameter)parameters.get(id);
}

/**
 * Returns the parameter with the specified name or <code>null</code>
 * if one is not found.
 *
 * @param name the name string
 * @return the parameter with that name or <code>null</code> if one is not
 * found
 */
public Parameter findParameterByName(String name) {
    if (name == null || name.length() == 0)
	return null;

    name = name.toLowerCase();
    for (Iterator iter = parameters.values().iterator(); iter.hasNext(); ) {
	Parameter p = (Parameter)iter.next();
	if (name.equals(p.getName().toLowerCase()))
	    return p;
    }
    return null;
}

public void addParameter(Parameter p) {
    parameters.put(p.getId(), p);
}
public void removeParameter(Parameter p) {
    parameters.remove(p.getId());
}
public Iterator parameters() { return parameters.values().iterator(); }

// ---------------- formulas

/**
 * Returns the formula with the specified id or <code>null</code>
 * if one is not found.
 *
 * @param id the formula id
 * @return the formula with that id or <code>null</code> if one is not found
 */
public Formula findFormula(Object id) {
    if (id instanceof String)
	id = new Long((String)id);
    return (Formula)formulas.get(id);
}

/**
 * Returns the formula with the specified name or <code>null</code>
 * if one is not found.
 *
 * @param name the name string
 * @return the formula with that name or <code>null</code> if one is not found
 */
public Formula findFormulaByName(String name) {
    if (name == null || name.length() == 0)
	return null;

    name = name.toLowerCase();
    for (Iterator iter = formulas.values().iterator(); iter.hasNext(); ) {
	Formula f = (Formula)iter.next();
	if (name.equals(f.getName().toLowerCase()))
	    return f;
    }
    return null;
}

public void addFormula(Formula f) {
    formulas.put(f.getId(), f);
}
public void removeFormula(Formula f) {
    formulas.remove(f.getId());
}
public Iterator formulas() { return formulas.values().iterator(); }

// ---------------- user columns

/**
 * Returns the user column with the specified id or <code>null</code>
 * if one is not found.
 *
 * @param id the user column id
 * @return the user column with that id or <code>null</code> if one
 * is not found
 */
public UserColumn findUserColumn(Object id) {
    if (id instanceof String)
	id = new Long((String)id);
    return (UserColumn)usercols.get(id);
}

/**
 * Returns the user column with the specified name or <code>null</code>
 * if one is not found.
 *
 * @param name the name string
 * @return the user column with that name or <code>null</code> if one
 * is not found
 */
public UserColumn findUserColumnByName(String name) {
    if (name == null || name.length() == 0)
	return null;

    name = name.toLowerCase();
    for (Iterator iter = usercols.values().iterator(); iter.hasNext(); ) {
	UserColumn f = (UserColumn)iter.next();
	if (name.equals(f.getName().toLowerCase()))
	    return f;
    }
    return null;
}

public void addUserColumn(UserColumn uc) {
    usercols.put(uc.getId(), uc);
}
public void removeUserColumn(UserColumn uc) {
    usercols.remove(uc.getId());
}
public Iterator userColumns() { return usercols.values().iterator(); }

// ---------------- section manipulation

/**
 * Returns <code>true</code> if this report contains the specified section.
 *
 * @return <code>true</code> if this report contains the specified section
 */
public boolean contains(Section s) {
    switch (s.getArea()) {
    case AREA_REPORT_HEADER: return reportHeaders.contains(s);
    case AREA_REPORT_FOOTER: return reportFooters.contains(s);
    case AREA_PAGE_HEADER: return pageHeaders.contains(s);
    case AREA_PAGE_FOOTER: return pageFooters.contains(s);
    case AREA_DETAIL: return details.contains(s);
    case AREA_GROUP_HEADER:
	for (Iterator iter = groups(); iter.hasNext(); )
	    if (((Group)iter.next()).headerSections.contains(s)) return true;
	return false;
    case AREA_GROUP_FOOTER:
	for (Iterator iter = groups(); iter.hasNext(); )
	    if (((Group)iter.next()).footerSections.contains(s)) return true;
	return false;
    default:			// Should never happen
	return false;
    }
}

/**
 * Returns the first section in the list of the specified type. If the
 * type is <code>GROUP_HEADER</code> or <code>GROUP_FOOTER</code>, return
 * (a) <code>null</code> if there are no groups in the report or
 * (b) the first header or footer of the first group.
 * <p>
 * Used by <code>FieldClipping</code> when trying to paste a field into
 * either some other report or the same report if the original section
 * is no longer in that report.
 *
 * @return a section; may be <code>null</code>
 * @see jimm.datavision.gui.cmd.FieldClipping
 */
public Section getFirstSectionByArea(int area) {
    switch (area) {
    case AREA_REPORT_HEADER: return (Section)reportHeaders.get(0);
    case AREA_REPORT_FOOTER: return (Section)reportFooters.get(0);
    case AREA_PAGE_HEADER: return (Section)pageHeaders.get(0);
    case AREA_PAGE_FOOTER: return (Section)pageFooters.get(0);
    case AREA_DETAIL: return (Section)details.get(0);
    case AREA_GROUP_HEADER:
	if (groups.size() == 0) return null;
	return (Section)((Group)groups.get(0)).headers().next();
    case AREA_GROUP_FOOTER:
	if (groups.size() == 0) return null;
	return (Section)((Group)groups.get(0)).footers().next();
    }
    return null;
}

/**
 * Adds a (possibly created) section after <var>afterThis</var> and
 * returns the section. If <var>section</var> is <code>null</code>,
 * a new section will be created.
 *
 * @param section the section to insert; if <code>null</code>, a new
 * section will be created
 * @param afterThis the new section will be inserted after this one
 * @param list the list in which <var>afterThis</var> is found
 * @param nameKey the new section name's I18N lookup key
 * @see I18N
 */
protected Section insertSectionAfter(Section section, Section afterThis,
				     ArrayList list, String nameKey)
{
    int i = list.indexOf(afterThis);
    if (section == null)
	section = new Section(this, I18N.get(nameKey), false);
    list.add(i + 1, section);
    return section;
}

/**
 * Returns a structure useful only by this report for re-inserting
 * a section. This structure may later be passed to
 * <code>reinsertHeader</code>. Used by <code>DeleteSectionCommand</code>.
 *
 * @param section the section in question
 * @return section location information
 * @see jimm.datavision.gui.cmd.DeleteSectionCommand
 */
public ReportSectionLoc getSectionLocation(Section s) {
    List list = null;
    int index = -1;

    switch (s.getArea()) {
    case AREA_REPORT_HEADER: list = reportHeaders; break;
    case AREA_REPORT_FOOTER: list = reportFooters; break;
    case AREA_PAGE_HEADER: list = pageHeaders; break;
    case AREA_PAGE_FOOTER: list = pageFooters; break;
    case AREA_DETAIL: list = details; break;
    case AREA_GROUP_HEADER:
	for (Iterator iter = groups(); iter.hasNext(); ) {
	    Group g = (Group)iter.next();
	    if (g.headerSections.contains(s))
		list = g.headerSections;
	}
	break;
    case AREA_GROUP_FOOTER:
	for (Iterator iter = groups(); iter.hasNext(); ) {
	    Group g = (Group)iter.next();
	    if (g.footerSections.contains(s))
		list = g.footerSections;
	}
	break;
    }

    return new ReportSectionLoc(s, list, list.indexOf(s));
}

/**
 * Reinserts a section based on the location information previously
 * retrieved by a call to <code>getSectionLocation</code>. Used by
 * <code>DeleteSectionCommand</code>.
 *
 * @param loc a section locatction
 * @see jimm.datavision.gui.cmd.DeleteSectionCommand
 */
public void reinsertSection(ReportSectionLoc loc) {
    loc.list.add(loc.index, loc.section);
}

// ---------------- headers

public void addHeader(Section s) {
    reportHeaders.add(s);
    s.setArea(AREA_REPORT_HEADER);
}
public Section insertHeaderAfter(Section s, Section afterThis) {
    s.setArea(AREA_REPORT_HEADER);
    return insertSectionAfter(s, afterThis, reportHeaders,
			      "Report.report_header");
}
public void removeHeader(Section s) {
    reportHeaders.remove(s);
}
public Iterator headers() { return reportHeaders.iterator(); }
/** Used for testing only. */
public int testCountHeaders() { return reportHeaders.size(); }

// ---------------- footers

public void addFooter(Section s) {
    reportFooters.add(s);
    s.setArea(AREA_REPORT_FOOTER);
}
public Section insertFooterAfter(Section s, Section afterThis) {
    s.setArea(AREA_REPORT_FOOTER);
    return insertSectionAfter(s, afterThis, reportFooters,
			      "Report.report_footer");
}
public void removeFooter(Section s) {
    reportFooters.remove(s);
}
public Iterator footers() { return reportFooters.iterator(); }
/** Used for testing only. */
public int testCountFooters() { return reportFooters.size(); }

// ---------------- page headers

public void addPageHeader(Section s) {
    pageHeaders.add(s);
    s.setArea(AREA_PAGE_HEADER);
}
public Section insertPageHeaderAfter(Section s, Section afterThis) {
    s.setArea(AREA_PAGE_HEADER);
    return insertSectionAfter(s, afterThis, pageHeaders, "Report.page_header");
}
public void removePageHeader(Section s) {
    pageHeaders.remove(s);
}
public Iterator pageHeaders() { return pageHeaders.iterator(); }
/** Used for testing only. */
public int testCountPageHeaders() { return pageHeaders.size(); }

// ---------------- page footers

public void addPageFooter(Section s) {
    pageFooters.add(s);
    s.setArea(AREA_PAGE_FOOTER);
}
public Section insertPageFooterAfter(Section s, Section afterThis) {
    s.setArea(AREA_PAGE_FOOTER);
    return insertSectionAfter(s, afterThis, pageFooters, "Report.page_footer");
}
public void removePageFooter(Section s) {
    pageFooters.remove(s);
}
public Iterator pageFooters() { return pageFooters.iterator(); }
/** Used for testing only. */
public int testCountPageFooters() { return pageFooters.size(); }

// ---------------- groups

public void addGroup(Group g) {
    groups.add(g);
    query.removeSort(g.getColumn());
}
public void removeGroup(Group g) { groups.remove(g); }
public void removeAllGroups() { groups.clear(); }
public Iterator groups() { return groups.iterator(); }
public int countGroups() { return groups.size(); }

/**
 * Returns an iterator over the groups in reverse order. Useful when
 * displaying group footers.
 *
 * @return an iterator over the groups, in reverse order
 */
public Iterator groupsReversed() {
    ArrayList reversed = (ArrayList)groups.clone();
    Collections.reverse(reversed);
    return reversed.iterator();
}

// ---------------- details

public void addDetail(Section s) {
    details.add(s);
    s.setArea(AREA_DETAIL);
}
public Section insertDetailAfter(Section s, Section afterThis) {
    s.setArea(AREA_DETAIL);
    return insertSectionAfter(s, afterThis, details, "Report.detail");
}
public void removeDetail(Section s) { details.remove(s); }
public Iterator details() { return details.iterator(); }
/** Used for testing only. */
public int testCountDetails() { return details.size(); }

public void removeField(Field f) {
    Section s = sectionContaining(f);
    if (s != null)
	s.removeField(f);
}

/**
 * Returns <code>true</code> if the specified field exists within this
 * report.
 *
 * @param f a field
 * @return <code>true</code> if the specified field exists within this
 * report
 */
public boolean contains(Field f) {
    return sectionContaining(f) != null;
}

/**
 * Returns the section containing the specified field, or <code>null</code>
 * if the field is not in the report.
 *
 * @param f a field
 * @return the section containing the specified field, or <code>null</code>
 * if the field is not in the report
 */
public Section sectionContaining(Field f) {
    Section s;
    Iterator iter, iter2;
    for (iter = reportHeaders.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.contains(f)) return s;
    }
    for (iter = pageHeaders.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.contains(f)) return s;
    }
    for (iter = groups.iterator(); iter.hasNext(); ) {
	for (iter2 = ((Group)iter.next()).headerSections.iterator();
	     iter2.hasNext(); ) {
	    s = (Section)iter2.next();
	    if (s.contains(f)) return s;
	}
    }
    for (iter = details.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.contains(f)) return s;
    }
    for (iter = groups.iterator(); iter.hasNext(); ) {
	for (iter2 = ((Group)iter.next()).footerSections.iterator();
	     iter2.hasNext(); ) {
	    s = (Section)iter2.next();
	    if (s.contains(f)) return s;
	}
    }
    for (iter = reportFooters.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.contains(f)) return s;
    }
    for (iter = pageFooters.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.contains(f)) return s;
    }

    return null;
}

/**
 * Returns <code>true</code> if the specified field exists within this
 * report either directly (as a field) or indirectly (as a formula used
 * by a subtotal, parameter, user column, or another formula).
 *
 * @param f a field
 * @return <code>true</code> if the specified field exists within this
 * report
 */
public boolean containsReferenceTo(Field f) {
    Section s;
    Iterator iter, iter2;
    for (iter = reportHeaders.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.containsReferenceTo(f)) return true;
    }
    for (iter = pageHeaders.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.containsReferenceTo(f)) return true;
    }
    for (iter = groups.iterator(); iter.hasNext(); ) {
	for (iter2 = ((Group)iter.next()).headerSections.iterator();
	     iter2.hasNext(); ) {
	    s = (Section)iter2.next();
	    if (s.containsReferenceTo(f)) return true;
	}
    }
    for (iter = details.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.containsReferenceTo(f)) return true;
    }
    for (iter = groups.iterator(); iter.hasNext(); ) {
	for (iter2 = ((Group)iter.next()).footerSections.iterator();
	     iter2.hasNext(); ) {
	    s = (Section)iter2.next();
	    if (s.containsReferenceTo(f)) return true;
	}
    }
    for (iter = reportFooters.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.containsReferenceTo(f)) return true;
    }
    for (iter = pageFooters.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.containsReferenceTo(f)) return true;
    }

    return false;
}

/**
 * Returns <code>true</code> if the specified formula exists within this
 * report either directly (as a formula field) or indirectly (as a formula
 * used by a subtotal or by another formula). This is not the same as
 * answering the question, "Does this report contain this formula?" because
 * we want to know if the visual portion of the report or some other formula
 * accesses the specified formula.
 *
 * @param f a formula
 * @return <code>true</code> if the specified parameter exists within some
 * visual element of the report or within some formula
 */
public boolean containsReferenceTo(Formula f) {
    Section s;
    Iterator iter, iter2;
    for (iter = reportHeaders.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.containsReferenceTo(f)) return true;
    }
    for (iter = pageHeaders.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.containsReferenceTo(f)) return true;
    }
    for (iter = groups.iterator(); iter.hasNext(); ) {
	for (iter2 = ((Group)iter.next()).headerSections.iterator();
	     iter2.hasNext(); ) {
	    s = (Section)iter2.next();
	    if (s.containsReferenceTo(f)) return true;
	}
    }
    for (iter = details.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.containsReferenceTo(f)) return true;
    }
    for (iter = groups.iterator(); iter.hasNext(); ) {
	for (iter2 = ((Group)iter.next()).footerSections.iterator();
	     iter2.hasNext(); ) {
	    s = (Section)iter2.next();
	    if (s.containsReferenceTo(f)) return true;
	}
    }
    for (iter = reportFooters.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.containsReferenceTo(f)) return true;
    }
    for (iter = pageFooters.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.containsReferenceTo(f)) return true;
    }

    return false;
}

/**
 * Returns <code>true</code> if the specified user column exists within
 * this report either directly (as a user column field) or indirectly
 * (inside a subtotal or formula). This is not the same as answering the
 * question, "Does this report contain this user column?" because we want
 * to know if the visual portion of the report or some formula accesses the
 * specified user column.
 *
 * @param uc a user column
 * @return <code>true</code> if the specified parameter exists within some
 * visual element of the report or within some formula
 */
public boolean containsReferenceTo(UserColumn uc) {
    Section s;
    Iterator iter, iter2;
    for (iter = reportHeaders.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.containsReferenceTo(uc)) return true;
    }
    for (iter = pageHeaders.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.containsReferenceTo(uc)) return true;
    }
    for (iter = groups.iterator(); iter.hasNext(); ) {
	for (iter2 = ((Group)iter.next()).headerSections.iterator();
	     iter2.hasNext(); ) {
	    s = (Section)iter2.next();
	    if (s.containsReferenceTo(uc)) return true;
	}
    }
    for (iter = details.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.containsReferenceTo(uc)) return true;
    }
    for (iter = groups.iterator(); iter.hasNext(); ) {
	for (iter2 = ((Group)iter.next()).footerSections.iterator();
	     iter2.hasNext(); ) {
	    s = (Section)iter2.next();
	    if (s.containsReferenceTo(uc)) return true;
	}
    }
    for (iter = reportFooters.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.containsReferenceTo(uc)) return true;
    }
    for (iter = pageFooters.iterator(); iter.hasNext(); ) {
	s = (Section)iter.next();
	if (s.containsReferenceTo(uc)) return true;
    }

    return false;
}

/**
 * Returns <code>true</code> if the specified parameter exists within this
 * report either directly (as a parameter field) or indirectly (as a
 * parameter used by a subtotal or by another parameter or in the query's
 * where clause). This is not the same as answering the question, "Does
 * this report contain this parameter?" because we want to know if the
 * visual portion of the report or some formula accesses the specified
 * parameter.
 *
 * @param p a parameter
 * @return <code>true</code> if the specified parameter exists within some
 * visual element of the report or within some formula or within the
 * query's where clause
 */
public boolean containsReferenceTo(Parameter p) {
    if (query.containsReferenceTo(p))
	return true;

    Section s;
    Iterator iter, iter2;

    for (iter = reportHeaders.iterator(); iter.hasNext(); )
	if (((Section)iter.next()).containsReferenceTo(p)) return true;
    for (iter = pageHeaders.iterator(); iter.hasNext(); )
	if (((Section)iter.next()).containsReferenceTo(p)) return true;
    for (iter = groups.iterator(); iter.hasNext(); ) {
	Group group = (Group)iter.next();
	for (iter2 = group.headerSections.iterator(); iter2.hasNext(); )
	    if (((Section)iter2.next()).containsReferenceTo(p)) return true;
	for (iter2 = group.footerSections.iterator(); iter2.hasNext(); )
	    if (((Section)iter2.next()).containsReferenceTo(p)) return true;
    }
    for (iter = details.iterator(); iter.hasNext(); )
	if (((Section)iter.next()).containsReferenceTo(p)) return true;
    for (iter = reportFooters.iterator(); iter.hasNext(); )
	if (((Section)iter.next()).containsReferenceTo(p)) return true;
    for (iter = pageFooters.iterator(); iter.hasNext(); )
	if (((Section)iter.next()).containsReferenceTo(p)) return true;

    return false;
}

/**
 * Returns a list of all the parameters actually used in the
 * report. Calls <code>containsReferenceTo</code>.
 *
 * @return a list of parameters
 * @see #containsReferenceTo
 */
protected List collectUsedParameters() {
    ArrayList list = new ArrayList();
    for (Iterator iter = parameters.values().iterator(); iter.hasNext(); ) {
	Parameter p = (Parameter)iter.next();
	if (containsReferenceTo(p))
	    list.add(p);
    }
    return list;
}

/**
 * Returns the group associated with the specified column, or
 * <code>null</code> if there isn't one.
 *
 * @param col a database column
 * @return the group associated with this column, or <code>null</code>
 * if there isn't one
 */
public Group findGroup(Column col) {
    for (Iterator iter = groups(); iter.hasNext(); ) {
	Group g = (Group)iter.next();
	if (g.getColumn() == col)
	    return g;
    }
    return null;
}

/**
 * Returns the group associated with the specified section, or
 * <code>null</code> if there isn't one.
 *
 * @param section a section
 * @return the group associated with this section, or <code>null</code>
 * if there isn't one
 */
public Group findGroup(Section section) {
    for (Iterator iter = groups(); iter.hasNext(); ) {
	Group group = (Group)iter.next();
	if (group.contains(section))
	    return group;
    }
    return null;
}

/**
 * Returns <code>true</code> if the specified section is contained in
 * any group. The section may be either a header or a footer section.
 *
 * @param section a section
 * @return <code>true</code> if the section is within some group
 */
public boolean isInsideGroup(Section section) {
    return findGroup(section) != null;
}

/**
 * Returns <code>true</code> if the specified database column is a
 * group column.
 *
 * @return <code>true</code> if the specified database column is a
 * group column
 */
public boolean isGroupColumn(Column col) {
    return findGroup(col) != null;
}

/**
 * Creates and returns a new section below the specified one.
 *
 * @param goBelowThis a section
 * @return a new section
 */
public Section insertSectionBelow(Section goBelowThis) {
    return insertSectionBelow(null, goBelowThis);
}

/**
 * Inserts a (possibly newly created) section below the specified one.
 * Returns the inserted section.
 *
 * @param section the section to insert
 * @param goBelowThis <var>section</var> goes below this one; 
 * @return a new section
 */
public Section insertSectionBelow(Section section, Section goBelowThis) {
    if (section == null)
	section = new Section(this, goBelowThis.getArea(),
			      goBelowThis.getName(),
			      goBelowThis.isDetail);

    if (goBelowThis == null)
	return insertHeaderAfter(section, null);

    if (reportHeaders.contains(goBelowThis))
	return insertHeaderAfter(section, goBelowThis);
    else if (reportFooters.contains(goBelowThis))
	return insertFooterAfter(section, goBelowThis);
    else if (pageHeaders.contains(goBelowThis))
	return insertPageHeaderAfter(section, goBelowThis);
    else if (pageFooters.contains(goBelowThis))
	return insertPageFooterAfter(section, goBelowThis);
    else if (details.contains(goBelowThis))
	return insertDetailAfter(section, goBelowThis);

    for (Iterator iter = groups(); iter.hasNext(); ) {
	Group g = (Group)iter.next();
	if (g.headerSections.contains(goBelowThis))
	    return g.insertHeaderAfter(section, goBelowThis);
	else if (g.footerSections.contains(goBelowThis))
	    return g.insertFooterAfter(section, goBelowThis);
    }

    return null;		// Should not happen
}

/**
 * Removes the specified section.
 *
 * @param s a section
 */
public void removeSection(Section s) {
    if (reportHeaders.contains(s))
	removeHeader(s);
    else if (reportFooters.contains(s))
	removeFooter(s);
    else if (pageHeaders.contains(s))
	removePageHeader(s);
    else if (pageFooters.contains(s))
	removePageFooter(s);
    else if (details.contains(s))
	removeDetail(s);
    else {
	for (Iterator iter = groups(); iter.hasNext(); ) {
	    Group g = (Group)iter.next();
	    if (g.headerSections.contains(s)) {
		g.removeHeader(s);
		return;
	    }
	    if (g.footerSections.contains(s)) {
		g.removeFooter(s);
		return;
	    }
	}
    }
}

/**
 * Returns <code>true</code> if this report contains some field anywhere.
 * Useful when the GUI is enabling menu items, for example.
 *
 * @return <code>true</code> if this report contains some field anywhere
 */
public boolean hasFields() {
    return countTablesUsed() > 0;
}

/**
 * Returns the number of tables used by the report. Useful when the GUI
 * is enabling menu items, for example.
 *
 * @return the number of tables used by this report (by the query)
 */
public int countTablesUsed() {
    return query.getTablesUsed().size();
}

/**
 * Returns <code>true</code> if the specified section is the only section
 * of its kind; that is, the only section in the collection in which it
 * is contained.
 *
 * @param s a report section
 * @return <code>true</code> if this section is a loner
 */
public boolean isOneOfAKind(Section s) {
    if (reportHeaders.contains(s) && reportHeaders.size() == 1)
	return true;
    if (reportFooters.contains(s) && reportFooters.size() == 1)
	return true;
    if (pageHeaders.contains(s) && pageHeaders.size() == 1)
	return true;
    if (pageFooters.contains(s) && pageFooters.size() == 1)
	return true;
    if (details.contains(s) && details.size() == 1)
	return true;

    for (Iterator iter = groups(); iter.hasNext(); ) {
	Group g = (Group)iter.next();
	if (g.headerSections.contains(s) && g.headerSections.size() == 1)
	    return true;
	if (g.footerSections.contains(s) && g.footerSections.size() == 1)
	    return true;
    }

    return false;
}

/**
 * Sets the database password.
 */
public void setDatabasePassword(String pwd) { databasePassword = pwd; }

/**
 * Sets a database's user name and password. If we already have the
 * password, give it to the database. If we don't, ask the user for
 * both the user name (supplied to us) and the password and give them
 * to the database.
 *
 * @param db the database
 */
public void askForPassword(Database db) {
    if (databasePassword != null) {
	db.setPassword(databasePassword);
	return;
    }

    if (ErrorHandler.usingGUI()) {
	DbPasswordDialog dialog =
	    new DbPasswordDialog(DesignWin.findWindowFor(this), db.getName(),
				 db.getUserName());
	// This dialog is modal, so when we return the values have been
	// filled.
	db.setUserName(dialog.getUserName());
	db.setPassword(dialog.getPassword());
    }
    else {
	System.out.print(I18N.get("Report.user_name") + " ["
			 + db.getUserName() + "]: ");
	System.out.flush();
	try {
	    BufferedReader in =
		new BufferedReader(new InputStreamReader(System.in));
	    String username = in.readLine();
	    if (username.length() > 0)
		db.setUserName(username);
	}
	catch (IOException e) {
	    ErrorHandler.error(I18N.get("Report.user_name_err"));
	}

	System.out.print(I18N.get("Report.password") + ' ' + db.getUserName()
			 + ": ");
	System.out.flush();
	try {
	    BufferedReader in =
		new BufferedReader(new InputStreamReader(System.in));
	    databasePassword = in.readLine();
	    db.setPassword(databasePassword);
	}
	catch (IOException e) {
	    ErrorHandler.error(I18N.get("Report.password_err"));
	}
    }
}

/**
 * Returns the value of the specified parameter. First time at the start
 * of each report run, asks user for parameter values.
 *
 * @param paramId a parameter id
 */
public Object getParameterValue(Object paramId) {
    if (!askedForParameters)
	askForParameters();

    return findParameter(paramId).getValue();
}

/**
 * Asks the user for parameter values. If the user cancels, we throw a
 * <code>UserCancellationException</code>.
 */
protected void askForParameters() {
    List usedParameters = null;
    if (askedForParameters
	|| (usedParameters = collectUsedParameters()).size() == 0)
    {
	askedForParameters = true;
	return;
    }

    if (ErrorHandler.usingGUI()) {
	DesignWin win = DesignWin.findWindowFor(this);
	if (parametersHaveValues) {
	    // Ask user about re-using the previous values
	    String msg = I18N.get("Report.use_prev_param_vals");
	    if (JOptionPane.showConfirmDialog(win, msg,
					      I18N.get("Report.use_prev_title"),
					      JOptionPane.YES_NO_OPTION)
		== JOptionPane.YES_OPTION)
		{
		    askedForParameters = true;
		    parametersHaveValues = true;
		    return;
		}
	}

	// This dialog is modal. By the time it returns, the parameters
	// have their new values.
	    
	if (new ParamAskWin(win, usedParameters).userCancelled())
	    throw new UserCancellationException(I18N.get("Report.user_cancelled"));
    }
    else {
	if (paramXMLFileName == null) {
	    ErrorHandler.error(I18N.get("Report.missing_param_xml_file"));
	    throw new UserCancellationException(I18N.get("Report.missing_param_xml_file_short"));
	}
	try {
	    new ParameterReader(this).readFile(paramXMLFileName);
	}
	catch (Exception ex) {
	    ErrorHandler.error(I18N.get("Report.param_file_err_1") + ' '
			       + paramXMLFileName
			       + I18N.get("Report.param_file_err_2"), ex);
	    throw new UserCancellationException(I18N.get("Report.param_file_err_short"));
	}
    }

    askedForParameters = true;
    parametersHaveValues = true;
}

/**
 * Spawns a new thread that runs the report, using the layout engine
 * to generate output. The new thread runs the method
 * <code>runReport</code>.
 * <p>
 * You may ask why this method is called <code>run</code> and it runs
 * another method called <code>runReport</code> when the normal convention
 * is for a thread to run a <code>Runnable</code> object that has a
 * <code>run</code> method. The reason is purely historical. This method
 * started non-threaded, and I don't want anyone else who relies on this
 * method to have to change their code.
 *
 * @see #runReport
 */
public void run() {
    new Thread(new Runnable() {
	public void run() { runReport(); }
	}).start();
}

/**
 * Runs the database query and hands rows to the layout engine. This
 * method is called from <code>run</code>, which spawns a new thred
 * in which this method is run. If you want to run a report, decide
 * if you want it to run in a separate thread or not. If so, call
 * <code>run</code>. If not, use this method.
 *
 * @see #run
 */
public void runReport() {
    askedForParameters = false;	// Ask user again
    try {
	askForParameters();
    }
    catch (UserCancellationException e) {
	return;
    }

    for (Iterator iter = groups.iterator(); iter.hasNext(); )
	((Group)iter.next()).reset();

    collectSubtotalFields();

    Connection conn = null;
    rset = null;
    StatusDialog statusDialog = null;
    try {
	if (ErrorHandler.usingGUI()) {
	    DesignWin win = DesignWin.findWindowFor(this);
	    statusDialog = new StatusDialog(win,
					    I18N.get("Report.status_title"),
					    true,
					    I18N.get("Report.status_running"));
	}

	if (!layoutEngine.wantsMoreData())
	    return;

	rset = database.execute(query);

	boolean layoutStarted = false;
	while (layoutEngine.wantsMoreData() && rset.next()) {
	    if (statusDialog != null) {
		if (statusDialog.isCancelled())
		    throw new UserCancellationException();
		statusDialog.update(I18N.get("Report.processing_row") + ' '
				    + rowNumber());
	    }

	    if (!layoutStarted) {
		layoutEngine.start();
		layoutStarted = true;
	    }
	    processResultRow();
	}

	rset.last();		// Recall last row so we can access the data
	if (!layoutStarted) {	// No rows in report
	    layoutEngine.start();
	    layoutEngine.end();
	}
	else {			// Output group footers and end of report
	    layoutEngine.groupFooters(true);
	    layoutEngine.end();
	}
    }
    catch (UserCancellationException uce) {
	layoutEngine.cancel();
    }
    catch (Exception e) {
	layoutEngine.cancel();
	ErrorHandler.error(e);
    }
    finally {
	if (rset != null) rset.close();

	subtotalFields = null;
	if (statusDialog != null)
	    statusDialog.dispose();
    }
}

/**
 * Processes a single database row. Note that the <code>next</code>
 * method of the result set has already been called.
 */
protected void processResultRow() throws java.sql.SQLException {
    updateGroups();

    // To output footers, bring back the previous row of data
    if (!rset.isFirst()) {
	rset.previous();
	layoutEngine.groupFooters(false);
	rset.next();
    }

    updateSubtotals();

    boolean isLastRow = rset.isLast();
    layoutEngine.groupHeaders(isLastRow);
    layoutEngine.detail(isLastRow);
}

/**
 * Returns the current value of the specified column. Only defined
 * when running a report.
 *
 * @return the string or double value of the column
 */
public Object columnValue(Column column) {
    // Ask query for field number, then get value of that column
    return rset.getObject(query.indexOfColumn(column) + 1);
}

/**
 * Returns the current value of the specified user column. Only defined
 * when running a report.
 *
 * @return the string or double value of the user column
 */
public Object columnValue(UserColumn uc) {
    // Ask query for field number, then get value of that column
    return rset.getObject(query.indexOfColumn(uc) + 1);
}

/**
 * Returns the current page number. Asks the layout engine. Only defined
 * when running a report.
 *
 * @return a page number
 */
public int pageNumber() {
    return layoutEngine.pageNumber;
}

/**
 * Returns the current data row number. Only defined when running a report.
 */
public int rowNumber() {
    return rset.getRow();
}

/**
 * Returns the current row of data. Only defined when running a report.
 */
public Row getCurrentRow() {
    return rset;
}

/**
 * Iterates over all sections in the report, passing the section to the
 * specified <code>SectionWalker</code>. The sections are visited in the
 * order in which they will be displayed.
 *
 * @param s a section walker
 */
public void withSectionsDo(SectionWalker s) {
    for (Iterator iter = reportHeaders.iterator(); iter.hasNext(); )
	s.step((Section)iter.next());
    for (Iterator iter = pageHeaders.iterator(); iter.hasNext(); )
	s.step((Section)iter.next());
    for (Iterator iter = groups.iterator(); iter.hasNext(); )
	for (Iterator iter2 = ((Group)iter.next()).headerSections.iterator();
	     iter2.hasNext(); )
	    s.step((Section)iter2.next());
    for (Iterator iter = details.iterator(); iter.hasNext(); )
	s.step((Section)iter.next());
    for (Iterator iter = groups.iterator(); iter.hasNext(); )
	for (Iterator iter2 = ((Group)iter.next()).footerSections.iterator();
	     iter2.hasNext(); )
	    s.step((Section)iter2.next());
    for (Iterator iter = reportFooters.iterator(); iter.hasNext(); )
	s.step((Section)iter.next());
    for (Iterator iter = pageFooters.iterator(); iter.hasNext(); )
	s.step((Section)iter.next());
}

/**
 * Iterates over all fields in the report, passing the section to the
 * specified <code>FieldWalker</code>.
 *
 * @param f a field walker
 */
public void withFieldsDo(final FieldWalker f) {
    withSectionsDo(new SectionWalker() {
	public void step(Section s) {
	    for (Iterator it = s.fields(); it.hasNext(); )
		f.step((Field)it.next());
	}
	});
}

/**
 * Collects all subtotal fields and lets each one initialize itself.
 * Used once at the beginning of each run.
 */
protected void collectSubtotalFields() {
    subtotalFields = new ArrayList();
    withFieldsDo(new FieldWalker() {
	public void step(Field f) {
	    if (f instanceof SubtotalField) {
		((SubtotalField)f).initialize();
		subtotalFields.add(f);
	    }
	}
	});
}

/**
 * Collects all subtotal fields that are subtotaling the specified field.
 * Used by the report design GUI.
 * 
 * @param field a field
 * @return a list of subtotal fields
 */
public AbstractList getSubtotalFieldsFor(final Field field) {
    final ArrayList subs = new ArrayList();
    withFieldsDo(new FieldWalker() {
	public void step(Field f) {
	    if (f instanceof SubtotalField
		&& ((SubtotalField)f).getField() == field)
	    {
		subs.add(f);
	    }
	}
	});
    return subs;
}

/**
 * Updates each subtotal field.
 */
protected void updateSubtotals() {
    for (Iterator iter = subtotalFields.iterator(); iter.hasNext(); )
	((SubtotalField)iter.next()).updateSubtotal();
}

/**
 * Updates each group's value based on the current value of the column
 * each group uses.
 */
protected void updateGroups() {
    for (Iterator iter = groups.iterator(); iter.hasNext(); ) {
	Group g = (Group)iter.next();
	g.setValue(columnValue(g.column));
    }
}

/**
 * Remembers name of parameter XML file.
 *
 * @param fileName an XML file name
 */
public void setParameterXMLFile(String fileName)
{
    paramXMLFileName = fileName;
}

/**
 * Read an XML file and build the contents of this report. Uses a
 * <code>ReportReader</code>.
 *
 * @param fileName a file name string
 */
public void readFile(String fileName)
    throws FileNotFoundException, IOException, SQLException,
	   ClassNotFoundException, InstantiationException,
	   IllegalAccessException, SAXException, UserCancellationException
{
    new ReportReader(this).readFile(fileName);
}

/**
 * Writes the contents of this report as an XML file.
 *
 * @param fileName a file name string
 */
public void writeFile(String fileName) {
    IndentWriter out = null;
    try {
	out = new IndentWriter(new OutputStreamWriter(new FileOutputStream(fileName),
						      XML_JAVA_ENCODING));
	writeXML(out);
	out.close();
    }
    catch (IOException ioe) {
	ErrorHandler.error(I18N.get("Report.write_err") + ' ' + fileName, ioe,
			   I18N.get("Report.write_err_title"));
    }
    finally {
	if (out != null) out.close();
    }
}

/**
 * Writes the contents of this report as an XML file.
 *
 * @param out an indent writer
 */
public void writeXML(IndentWriter out) {
	out.println("<?xml version=\"1.0\" encoding=\""
		    + XML_ENCODING_ATTRIBUTE + "\"?>");

	out.println("<!-- Generated by DataVision version " + info.Version
		    + " -->");
	out.println("<!-- " + info.URL + " -->");

	out.println("<report dtd-version=\"" + OUTPUT_DTD_VERSION
		    + "\" name=\"" + Utils.escapeXML(name)
		    + "\" title=\"" + Utils.escapeXML(title) + "\"");
	out.indent();
	out.indent();
	out.println("author=\"" + Utils.escapeXML(author) + "\">");
	out.outdent();

	out.println();
	out.println("<description><![CDATA["
		    + (description == null ? "" : description)
		    + "]]></description>");

	out.println();
	paperFormat.writeXML(out);

	out.println();
	database.writeXML(out);

	out.println();
	query.writeXML(out);

	Iterator iter;

	out.println();
	out.println("<parameters>");
	out.indent();
	for (iter = parameters.values().iterator(); iter.hasNext(); )
	     ((Parameter)iter.next()).writeXML(out);
	out.outdent();
	out.println("</parameters>");

	out.println();
	out.println("<formulas>");
	out.indent();
	for (iter = formulas.values().iterator(); iter.hasNext(); )
	     ((Formula)iter.next()).writeXML(out);
	out.outdent();
	out.println("</formulas>");

	out.println();
	out.println("<usercols>");
	out.indent();
	for (iter = usercols.values().iterator(); iter.hasNext(); )
	     ((UserColumn)iter.next()).writeXML(out);
	out.outdent();
	out.println("</usercols>");

	out.println();
	ListWriter.writeList(out, reportHeaders, "headers");
	out.println();
	ListWriter.writeList(out, reportFooters, "footers");

	out.println();
	out.println("<page>");
	out.indent();
	ListWriter.writeList(out, pageHeaders, "headers");
	ListWriter.writeList(out, pageFooters, "footers");
	out.outdent();
	out.println("</page>");

	out.println();
	out.println("<groups>");
	out.indent();
	for (iter = groups.iterator(); iter.hasNext(); ) {
	    out.println();
	    ((Group)iter.next()).writeXML(out);
	}
	out.outdent();
	out.println("</groups>");

	out.println();
	ListWriter.writeList(out, details, "details");

	out.outdent();
	out.println();
	out.println("</report>");
}

}
