package jimm.datavision;
import java.io.*;
import java.util.*;
import java.awt.Color;
import java.sql.SQLException;
import jimm.datavision.sql.*;
import uk.co.wilson.xml.MinML;
import org.xml.sax.*;

/**
 * A report reader reads an XML file and creates the innards of a report.
 * It uses the MinML parser.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class ReportReader extends MinML {

// ================================================================
/**
 * This class is used when we are converting formulas from the old
 * pre-DTD_VERSION_FORMULA_IDS format to the new format.
 * <p>
 * We store a copy of the original eval string because if we were to
 * retrieve it from the formula by calling
 * <code>Formula.getEvalString</code>, the formula would attempt to start
 * observing other formulas within the eval string. This would not work
 * because those formulas are represented the old way (by name instead of
 * id) and thus the code that looks for those formulas in order to observe
 * them fails. We don't have to clone the eval string, just save a
 * reference to it.
 */
class FormulaConversion {
Formula formula;
String evalString;
FormulaConversion(Formula f, String evalStr) {
    formula = f;
    evalString = evalStr;
}
}
// ================================================================

/**
 * If there is no report element dtd-version attribute, this is the
 * default value to use.
 */
protected static final double DEFAULT_DTD_VERSION = 0.2;

/**
 * This is the DTD version where formula ids were introduced. Versions
 * before this one require runtime conversion.
 */
protected static final double DTD_VERSION_FORMULA_IDS = 0.2;

protected Stack tagNameStack;
protected Report report;
protected Parameter parameter;
protected Formula formula;
protected UserColumn usercol;
protected Section section;
protected Group group;
protected Field field;
protected String textData;
protected Border border;
protected Line line;
protected double dtdVersion;
protected HashMap formulasToConvert;
protected int nextSectionLocation;
protected boolean missingColumnSeen;

/**
 * Constructor.
 *
 * @param report the report we are building
 */
public ReportReader(Report report) {
    this.report = report;
    tagNameStack = new Stack();
    dtdVersion = DEFAULT_DTD_VERSION;
}

/**
 * Reads the named XML file and builds the innards of the report.
 *
 * @param fileName the name of the XML file
 */
public void readFile(String fileName)
    throws FileNotFoundException, IOException, SAXException
{
    // Get rid of sections that we created in initEmptyReport()
    report.reportHeaders = new ArrayList();
    report.pageHeaders = new ArrayList();
    report.pageFooters = new ArrayList();
    report.reportFooters = new ArrayList();
    report.groups = new ArrayList();
    report.details = new ArrayList();

    parse(new InputStreamReader(new FileInputStream(fileName),
				Report.XML_JAVA_ENCODING));

    convertFormulas();

    // Headers, footers, and details must have at least one section.
    ensureNotEmpty(report.reportHeaders, I18N.get("Report.report_header"),
		   false);
    ensureNotEmpty(report.pageHeaders, I18N.get("Report.page_header"),
		   false);
    ensureNotEmpty(report.details, I18N.get("Report.detail"), true);
    ensureNotEmpty(report.reportFooters, I18N.get("Report.report_footer"),
		   false);
    ensureNotEmpty(report.pageFooters, I18N.get("Report.page_footer"),
		   false);
}

/**
 * Ensures that the specified collection of sections is not empty. If we
 * do create a section, it is marked as suppressed.
 *
 * @param list collection of sections
 * @param name section name to use if we create a new one
 * @param isDetail if <code>true<code> the list contains report detail
 * sections
 */
protected void ensureNotEmpty(Collection list, String name, boolean isDetail) {
    if (list.isEmpty()) {
	Section s = new Section(report, name, isDetail);
	s.setSuppressed(true);
	list.add(s);
    }
}

public void startElement(String name, AttributeList attributes)
    throws SAXException
{
    String parentTag = tagNameStack.empty() ? null
	: (String)tagNameStack.peek();
    tagNameStack.push(new String(name));

    // Get ready to start collecting text
    if (textData == null || textData.length() > 0)
	textData = new String();

    if ("report".equals(name)) report(attributes);
    else if ("database".equals(name)) database(attributes);
    else if ("query".equals(name)) query(attributes);
    else if ("join".equals(name)) join(attributes);
    else if ("sort".equals(name)) sort(attributes);
    else if ("parameter".equals(name)) parameter(attributes);
    else if ("formula".equals(name)) formula(attributes);
    else if ("usercol".equals(name)) usercol(attributes);
    else if ("headers".equals(name)) header(parentTag);
    else if ("footers".equals(name)) footer(parentTag);
    else if ("group".equals(name)) group(attributes);
    else if ("details".equals(name)) nextSectionLocation = Report.AREA_DETAIL;
    else if ("section".equals(name)) section(attributes);
    else if ("field".equals(name)) field(attributes);
    else if ("bounds".equals(name)) bounds(attributes);
    else if ("edge".equals(name)) edge(attributes);
    else if ("format".equals(name)) format(attributes);
    else if ("border".equals(name)) border(attributes);
    else if ("line".equals(name)) line(attributes);
    else if ("point".equals(name)) point(attributes);
    else if ("paper".equals(name)) paper(attributes);
}

protected void header(String parentTag) {
    if ("report".equals(parentTag))
	nextSectionLocation = Report.AREA_REPORT_HEADER;
    else if ("page".equals(parentTag))
	nextSectionLocation = Report.AREA_PAGE_HEADER;
    else if ("group".equals(parentTag))
	nextSectionLocation = Report.AREA_GROUP_HEADER;
}

protected void footer(String parentTag) {
    if ("report".equals(parentTag))
	nextSectionLocation = Report.AREA_REPORT_FOOTER;
    else if ("page".equals(parentTag))
	nextSectionLocation = Report.AREA_PAGE_FOOTER;
    else if ("group".equals(parentTag))
	nextSectionLocation = Report.AREA_GROUP_FOOTER;
}

/**
 * Handle elements expecting text data.
 */
public void endElement(String name) {
    // If we were paranoid, we could compare name to the top of the stack.
    // Let's not.
    String tagName = (String)tagNameStack.pop();

    if ("description".equals(tagName))
	report.setDescription(textData);
    else if ("default".equals(tagName) && parameter != null)
	parameter.addDefaultValue(textData);
    else if ("formula".equals(tagName)) {
	formula.setEvalString(textData);
	// If we need to convert this formula later, save a copy of
	// this eval string text.
	if (formulasToConvert != null) {
	    FormulaConversion fc =
		(FormulaConversion)formulasToConvert.get(formula.getName());
	    if (fc != null)
		fc.evalString = new String(textData);
	}
	formula = null;
    }
    else if ("usercol".equals(tagName)) {
	usercol.setUserString(textData);
	usercol = null;
    }
    else if ("text".equals(tagName) && field != null)
	field.setValue(textData);
    else if ("where".equals(tagName))
	report.query.setWhereClause(textData);
}

/**
 * Reads text data. Text data inside a single tag can be broken up into
 * multiple calls to this method.
 */
public void characters(char ch[], int start, int length) {
    textData += new String(ch, start, length);
}

/**
 * Reads the report tag.
 */
protected void report(AttributeList attributes) {
    String dtdVersionString = attributes.getValue("dtd-version");
    dtdVersion = (dtdVersionString == null)
	? DEFAULT_DTD_VERSION : Double.parseDouble(dtdVersionString);

    report.setName(attributes.getValue("name"));
    report.setTitle(attributes.getValue("title"));
    report.setAuthor(attributes.getValue("author"));
}

/**
 * Reads the database tag and creates the database object.
 */
protected void database(AttributeList attributes) {
    try {
	report.database =
	    new Database(attributes.getValue("driverClassName"),
			 attributes.getValue("connInfo"),
			 report,
			 attributes.getValue("name"),
			 attributes.getValue("username"));
    }
    catch (UserCancellationException iae) {
	// Thrown by database when user cancelled password dialog.
	// Let the report catch this.
	throw iae;
    }
    catch (Exception e) {
	ErrorHandler.error(I18N.get("ReportReader.db_err"), e,
			   I18N.get("ReportReader.db_err_title"));
    }
}

/** Reads the query. */
protected void query(AttributeList attributes) {
    report.setQuery(new Query(report));
}

protected void join(AttributeList attributes) {
    Column from = findColumn(attributes.getValue("from"));
    Column to = findColumn(attributes.getValue("to"));
    if (from != null && to != null)
	report.query.addJoin(new Join(from, attributes.getValue("relation"),
				      to));
}

protected void sort(AttributeList attributes) {
    Column col = findColumn(attributes.getValue("column"));
    if (col != null)
	report.query.addSort(col, attributes.getValue("order"));
}

protected void parameter(AttributeList attributes) throws SAXException {
    parameter = new Parameter(new Long(attributes.getValue("id")),
			      report,
			      attributes.getValue("type"),
			      attributes.getValue("name"),
			      attributes.getValue("question"),
			      attributes.getValue("arity"));
    report.addParameter(parameter);
}

/**
 * Reads a formula. If the XML format is really old, we need to give
 * each formula an id number and translate its formula text so references
 * to other formulas use the other formula's id number instead of its name.
 */
protected void formula(AttributeList attributes) throws SAXException {
    String idString = attributes.getValue("id");
    String name = attributes.getValue("name");
    Long id = null;
    if (idString == null) {
	if (dtdVersion >= DTD_VERSION_FORMULA_IDS) {
	    String str = I18N.get("ReportReader.the_formula")
		+ ' ' + name + ' '
		+ I18N.get("ReportReader.formula_missing_id_err");
	    throw new SAXException(str);
	}
	// else, we are OK with a null id
    }
    else
	id = new Long(idString);

    formula = new Formula(id, report, name, null);
    report.addFormula(formula);

    // If the id is null, that means we need to convert this formula
    // to the current format.
    if (id == null) {
	if (formulasToConvert == null)
	    formulasToConvert = new HashMap();
	FormulaConversion fc = new FormulaConversion(formula, null);
	formulasToConvert.put(formula.getName(), fc);
    }
}

/**
 * Reads a user column. Value of user column will be read later.
 */
protected void usercol(AttributeList attributes) throws SAXException {
    usercol = new UserColumn(new Long(attributes.getValue("id")), report,
			     attributes.getValue("name"), null);
    report.addUserColumn(usercol);
}

/**
 * Revisits each formula and let it convert formula names to
 * formula id numbers within its eval string.
 */
protected void convertFormulas() throws SAXException {
    if (formulasToConvert != null) {
	for (Iterator iter = formulasToConvert.values().iterator();
	     iter.hasNext(); )
	{
	    FormulaConversion fc = (FormulaConversion)iter.next();
	    Formula f = fc.formula;
	    try {
		f.setEditableEvalString(fc.evalString);
	    }
	    catch (IllegalArgumentException iae) {
		String msg = I18N.get("ReportReader.the_formula")
		    + ' ' + f.getName() + ' '
		    + I18N.get("ReportReader.formula_unknown_name");
		throw new SAXException(msg);
	    }
	}
    }
}

/** Creates a group and adds it to the report. */
protected void group(AttributeList attributes) {
    Column col = findColumn(attributes.getValue("column"));
    if (col == null) {
	group = null;
	return;
    }

    group = new Group(col);
    report.groups.add(group);
}

/** Creates an empty section and adds it to the report. */
protected void section(AttributeList attributes) {
    // The section's area will change when it is added to the report
    section = new Section(report, Report.AREA_REPORT_HEADER, null,
			  Double.parseDouble(attributes.getValue("height")),
			  false);

    String suppressedVal = attributes.getValue("suppressed");
    section.setSuppressed("true".equals(suppressedVal));

    addSectionToReport();
}

/**
 * Adds the last seen section to the report. The value of
 * <code>nextSectionLocation</code> determines where the section belongs.
 */
protected void addSectionToReport() {
    switch (nextSectionLocation) {
    case Report.AREA_REPORT_HEADER:
	section.setName(I18N.get("Report.report_header"));
	section.setIsDetail(false);
	report.reportHeaders.add(section);
	break;
    case Report.AREA_REPORT_FOOTER:
	section.setName(I18N.get("Report.report_footer"));
	section.setIsDetail(false);
	report.reportFooters.add(section);
	break;
    case Report.AREA_PAGE_HEADER:
	section.setName(I18N.get("Report.page_header"));
	section.setIsDetail(false);
	report.pageHeaders.add(section);
	break;
    case Report.AREA_PAGE_FOOTER:
	section.setName(I18N.get("Report.page_footer"));
	section.setIsDetail(false);
	report.pageFooters.add(section);
	break;
    case Report.AREA_GROUP_HEADER:
	section.setName(I18N.get("Report.group_header"));
	section.setIsDetail(false);
	if (group != null)
	    group.headerSections.add(section);
	break;
    case Report.AREA_GROUP_FOOTER:
	section.setName(I18N.get("Report.group_footer"));
	section.setIsDetail(false);
	if (group != null)
	    group.footerSections.add(section);
	break;
    case Report.AREA_DETAIL:
	section.setName(I18N.get("Report.detail"));
	section.setIsDetail(true);
	report.details.add(section);
	break;
    default:
	break;
    }
}

/**
 * Reads and creates a field. If the XML format is really old, we need to
 * convert formula fields by changing their values from the formula name
 * to the formula id.
 */
protected void field(AttributeList attributes) {
    String id = attributes.getValue("id");
    String type = attributes.getValue("type");
    Object value = attributes.getValue("value");
    String visibleString = attributes.getValue("visible");

    boolean visible = true;
    if (visibleString != null) {
	visibleString = visibleString.trim().toLowerCase();
	if (visibleString.length() > 0)
	    visible = "true".equals(visibleString);
    }

    // If this is a database column, make sure the column exists.
    if ("column".equals(type) && findColumn(value.toString()) == null) {
	field = null;
	return;
    }

    // If we are converting formulas without id numbers, change our value
    // (a pointer to the formula itself) from the formula's name to the
    // formula's newly-created id.
    if ("formula".equals(type) && formulasToConvert != null && value != null) {
	Formula f = ((FormulaConversion)formulasToConvert.get(value)).formula;
	if (f != null)
	    value = f.getId();
    }

    field = Field.create(new Long(id), report, section, type, value, visible);
    if (field instanceof SubtotalField && group != null)
	((SubtotalField)field).setGroup(group);

    section.addField(field);
}

/** Reads and sets the current field's bounds rectangle. */
protected void bounds(AttributeList attributes) {
    if (field != null)
	field.getBounds()
	    .setBounds(Double.parseDouble(attributes.getValue("x")),
		       Double.parseDouble(attributes.getValue("y")),
		       Double.parseDouble(attributes.getValue("width")),
		       Double.parseDouble(attributes.getValue("height")));
}

/** * Reads and creates the current field's format. */
protected void format(AttributeList attributes) {
    if (field == null)
	return;

    Format format = new Format();
    String val;

    if ((val = attributes.getValue("font")) != null)
	format.setFont(val);
    if ((val = attributes.getValue("size")) != null)
	format.setSize(Double.parseDouble(val));
    if ((val = attributes.getValue("bold")) != null)
	format.setBold("true".equals(val));
    if ((val = attributes.getValue("italic")) != null)
	format.setItalic("true".equals(val));
    if ((val = attributes.getValue("underline")) != null)
	format.setUnderline("true".equals(val));
    if ((val = attributes.getValue("wrap")) != null)
	format.setWrap("true".equals(val));
    if ((val = attributes.getValue("align")) != null)
	format.setAlign(Format.alignFromString(val));
    if ((val = attributes.getValue("color")) != null)
	format.setColor(parseColor(val));
    if ((val = attributes.getValue("format")) != null)
	format.setFormat(val);

    field.setFormat(format);
}

/** Parses color string and returns a <code>java.awt.Color</code>. */
protected Color parseColor(String val) {
    StringTokenizer tok = new StringTokenizer(val, ";");
    int r = Integer.parseInt(tok.nextToken().trim());
    int g = Integer.parseInt(tok.nextToken().trim());
    int b = Integer.parseInt(tok.nextToken().trim());
    int a = Integer.parseInt(tok.nextToken().trim());
    return new Color(r, g, b, a);
}

/** Reads and creates a new field border. */
protected void border(AttributeList attributes) {
    if (field != null) {
	border = new Border(field);
	field.setBorder(border);

	String val = attributes.getValue("color");
	if (val != null)
	    border.setColor(parseColor(val));
    }
}

/** Reads and creates a new border edge. */
protected void edge(AttributeList attributes) {
    if (field == null)
	return;

    String val = attributes.getValue("style");
    int style = BorderEdge.styleFromString(val);

    val = attributes.getValue("thickness");
    double thickness = (val == null) ? 1.0 : Double.parseDouble(val);

    val = attributes.getValue("number");
    int num = (val == null) ? 1 : Integer.parseInt(val);

    BorderEdge edge = new BorderEdge(style, thickness, num);

    String loc = attributes.getValue("location");
    if ("top".equals(loc)) border.setTop(edge);
    else if ("bottom".equals(loc)) border.setBottom(edge);
    else if ("left".equals(loc)) border.setLeft(edge);
    else if ("right".equals(loc)) border.setRight(edge);
}

/** Reads and creates a new line. */
protected void line(AttributeList attributes) {
    String val = attributes.getValue("thickness");
    double thickness = (val == null) ? 1.0 : Double.parseDouble(val);
    Color color = null;
    if ((val = attributes.getValue("color")) != null)
	color = parseColor(val);

    val = attributes.getValue("visible");
    boolean visible = true;
    if (val != null) {
	val = val.trim().toLowerCase();
	if (val.length() > 0)
	    visible = "true".equals(val);
    }

    section.lines.add(line = new Line(report, section, thickness, color,
				      visible));
}

/** Reads a line's point and adds it to the current line. */
protected void point(AttributeList attributes) {
    line.addEndPoint(Double.parseDouble(attributes.getValue("x")),
		     Double.parseDouble(attributes.getValue("y")));
}

/** Reads paper size name and orientation. */
protected void paper(AttributeList attributes) {
    int orientation = PaperFormat.PORTRAIT; // Default value
    String orientationStr = attributes.getValue("orientation");
    if (orientationStr != null
	&& "landscape".equals(orientationStr.toLowerCase()))
	orientation = PaperFormat.LANDSCAPE;

    PaperFormat pf =
	PaperFormat.get(orientation, attributes.getValue("name"));
    if (pf != null)
	report.setPaperFormat(pf);
}

/**
 * Returns the column identified by its name. The first time we can not find
 * a column, we report an error to the user.
 *
 * @param fullName a column name
 * @return a database column
 */
protected Column findColumn(String fullName) {
    Column col = report.findColumn(fullName);
    if (col == null && !missingColumnSeen) {
	missingColumnSeen = true;
	ErrorHandler.error(I18N.get("ReportReader.the_column")
			   + ' ' + fullName + ' '
			   + I18N.get("ReportReader.column_unknown"));
    }
    return col;
}

}
