package jimm.datavision;
import jimm.datavision.sql.Row;
import java.sql.ResultSet;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Observer;
import java.util.Observable;

/**
 * A section of a report contains a group of fields and suppression
 * information. Sections include page headers and footers, report headers
 * and footers, group headers and footers, and details. Sections contain
 * elements (fields and lines).
 *
 * @see Report
 * @see Field
 * @see Line
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class Section
    extends Observable
    implements Nameable, Writeable, Observer
{

protected static final double DEFAULT_HEIGHT = 20;

protected Report report;
protected int area;
protected String name;
protected double height;
protected ArrayList fields;
protected ArrayList lines;
protected boolean suppressed;
protected SuppressionProc suppressionProc;
protected boolean isDetail;

/**
 * Constructor. Uses <code>Report.AREA_REPORT_HEADER</code> and
 * <code>DEFAULT_HEIGHT</code>.
 *
 * @param report the report containing this section
 * @param name the name of the section (for example, "Group Header")
 * @param isDetail if <code>true<code> this section is a report detail
 * section
 * @see Report
 */
public Section(Report r, String name, boolean isDetail) {
    this(r, Report.AREA_REPORT_HEADER, name, DEFAULT_HEIGHT, isDetail);
}

/**
 * Constructor. Uses <code>DEFAULT_HEIGHT</code>.
 *
 * @param report the report containing this section
 * @param area one of the <code>Report.AREA_</code>* constants
 * @param name the name of the section (for example, "Group Header")
 * @param isDetail if <code>true<code> this section is a report detail
 * section
 * @see Report
 */
public Section(Report r, int area, String name, boolean isDetail) {
    this(r, area, name, DEFAULT_HEIGHT, isDetail);
}

/**
 * Constructor.
 *
 * @param report the report containing this section
 * @param area one of the <code>Report.AREA_</code>* constants
 * @param name the name of the section (for example, "Group Header")
 * @param h the height
 * @param isDetail if <code>true<code> this section is a report detail
 * section
 * @see Report
 */
public Section(Report r, int area, String name, double h, boolean isDetail) {
    report = r;
    this.area = area;
    this.name = name;
    height = h;
    this.isDetail = isDetail;
    fields = new ArrayList();
    lines = new ArrayList();
    suppressed = false;
}

public void update(Observable o, Object arg) {
    setChanged();
    notifyObservers(arg);
}

/**
 * Returns the report containing this section.
 */
public Report getReport() { return report; }

/**
 * Returns the height of this section.
 *
 * @return the height
 */
public double getHeight() { return height; }

/**
 * Sets the height.
 *
 * @param newHeight the new height
 */
public void setHeight(double newHeight) {
    if (height != newHeight) {
	height = newHeight;
	setChanged();
	notifyObservers();
    }
}

/**
 * Returns the width of this section.
 *
 * @return the width
 */
public double getWidth() { return report.getPaperFormat().getWidth(); }

/**
 * Returns the area of the section (one of the <code>Report.AREA_</code>*
 * constants).
 *
 * @return one of the <code>Report.AREA_</code>* constants
 * @see Report
 */
public int getArea() { return area; }

/**
 * Sets the area to one of the <code>Report.AREA_</code>* constants.
 *
 * @param area one of the <code>Report.AREA_</code>* constants
 */
public void setArea(int area) { this.area = area; }

/**
 * Returns the name of this section.
 *
 * @return the section name
 */
public String getName() { return name; }

/**
 * Sets the section name.
 *
 * @param newName a new name
 */
public void setName(String newName) {
    if (name != newName && (name == null || !name.equals(newName))) {
	name = newName;
	setChanged();
	notifyObservers();
    }
}

/**
 * Given an id, returns the field within this section that has that id.
 * If no field with the specified id exists, returns <code>null</code>.
 *
 * @return a field, or <code>null</code> if no field with the specified
 * id exists in this section
 */
public Field findField(Object id) {
    Long lid;
    if (id instanceof String)
	lid = new Long((String)id);
    else
	lid = (Long)id;

    for (Iterator iter = fields.iterator(); iter.hasNext(); ) {
	Field f = (Field)iter.next();
	if (f.getId().equals(lid))
	    return f;
    }
    return null;
}

/**
 * Adds a field to this section.
 *
 * @param f a field
 */
public void addField(Field f) {
    fields.add(f);
    f.addObserver(this);
    f.setSection(this);
    setChanged();
    notifyObservers();
}

/**
 * Removes a field from this section.
 *
 * @param f field
 */
public void removeField(Field f) {
    if (fields.contains(f)) {
	fields.remove(f);
	f.deleteObserver(this);
	f.setSection(null);
	setChanged();
	notifyObservers();
    }
}

/**
 * Returns an iterator over all fields in this section.
 *
 * @return an iterator
 */
public Iterator fields() { return fields.iterator(); }

/**
 * Adds a line to this section.
 *
 * @param f a line
 */
public void addLine(Line l) {
    lines.add(l);
    setChanged();
    notifyObservers();
}

/**
 * Removes a line from this section.
 *
 * @param f line
 */
public void removeLine(Line f) {
    lines.remove(f);
    setChanged();
    notifyObservers();
}

/**
 * Returns an iterator over all lines in this section.
 *
 * @return an iterator
 */
public Iterator lines() { return lines.iterator(); }

public boolean isSuppressed() { return suppressed; }
public void setSuppressed(boolean newSuppressed) {
    if (suppressed != newSuppressed) {
	suppressed = newSuppressed;
	setChanged();
	notifyObservers();
    }
}

/**
 * Returns <code>true</code> if this is a report detail section.
 *
 * @return <code>true</code> if this is a report detail section
 */
public boolean isDetail() { return isDetail; }

/**
 * Tells this section if it is a detail section. Since sections can't be
 * moved around, this is only called by report creation code such as
 * <code>ReportReader.section</code>
 *
 * @param val new boolean value
 */
public void setIsDetail(boolean val) {
    if (isDetail != val) {
	isDetail = val;
	setChanged();
	notifyObservers();
    }
}

/**
 * Returns the supression proc this section uses
 *
 * @return a suppression proc
 */
public SuppressionProc getSuppressionProc() { return suppressionProc; }

/**
 * Sets the suppression proc to use.
 *
 * @param newSuppressionProc a new suppression proc
 */
public void setSuppressionProc(SuppressionProc newSuppressionProc) {
    if (suppressionProc != newSuppressionProc) {
	suppressionProc = newSuppressionProc;
	setChanged();
	notifyObservers();
    }
}

/**
 * Returns <code>true</code> if the specified field is inside this section.
 *
 * @param f a field
 * @return <code>true</code> if the field is within this section
 */
public boolean contains(Field f) {
    for (Iterator iter = fields(); iter.hasNext(); )
	if ((Field)iter.next() == f)
	    return true;
    return false;
}

/**
 * Returns <code>true</code> if the specified field exists within this
 * section either directly (as a field) or indirectly (as a formula used
 * by a subtotal, user column, or formula).
 *
 * @param f a field
 * @return <code>true</code> if the specified field is referenced within
 * this section
 */
public boolean containsReferenceTo(Field f) {
    for (Iterator iter = fields(); iter.hasNext(); ) {
	Field field = (Field)iter.next();
	if (field == f || field.refersTo(f))
	    return true;
    }
    return false;
}

/**
 * Returns <code>true</code> if the specified formula exists within this
 * secion either directly (as a formula field) or indirectly (as a formula
 * used by a subtotal or by another formula).
 *
 * @param f a formula
 * @return <code>true</code> if the specified formula is referenced within
 * this section
 */
public boolean containsReferenceTo(Formula f) {
    for (Iterator iter = fields(); iter.hasNext(); ) {
	Field field = (Field)iter.next();
	if (field.refersTo(f))
	    return true;
    }
    return false;
}

/**
 * Returns <code>true</code> if the specified user column exists within this
 * secion either directly (as a user column field) or indirectly (as a user
 * column used by a subtotal or by a formula).
 *
 * @param f a formula
 * @return <code>true</code> if the specified formula is referenced within
 * this section
 */
public boolean containsReferenceTo(UserColumn uc) {
    for (Iterator iter = fields(); iter.hasNext(); ) {
	Field field = (Field)iter.next();
	if (field.refersTo(uc))
	    return true;
    }
    return false;
}

/**
 * Returns <code>true</code> if the specified parameter exists within this
 * secion either directly (as a parameter field) or indirectly (as a parameter
 * used by a subtotal or by a formula).
 *
 * @param p a parameter
 * @return <code>true</code> if the specified formula is referenced within
 * this section
 */
public boolean containsReferenceTo(Parameter p) {
    for (Iterator iter = fields(); iter.hasNext(); ) {
	Field field = (Field)iter.next();
	if (field.refersTo(p))
	    return true;
    }
    return false;
}

/**
 * Return true if this section should be printed, given this particular
 * row of data and our supressed state and suppression proc.
 *
 * @param row a row of data
 * @return a boolean specifying if the data should be displayed or not
 */
public boolean isActiveGivenData(Row row) {
    if (suppressed) return false;
    if (suppressionProc == null) return true;
    return suppressionProc.suppressGivenData(row);
}

/**
 * Writes this section and all it contains as an XML tag.
 *
 * @param out a writer that knows about indenting XML tags
 */
public void writeXML(IndentWriter out) {
    out.print("<section height=\"" + height + "\"");
    if (suppressed) out.print(" suppressed=\"true\"");
    out.println(">");
    out.indent();

    for (Iterator iter = fields(); iter.hasNext(); )
	((Field)iter.next()).writeXML(out);
    for (Iterator iter = lines(); iter.hasNext(); )
	((Line)iter.next()).writeXML(out);
// FIX
// Suppression proc

    out.outdent();
    out.println("</section>");
}

}
