package jimm.datavision;
import java.sql.Types;

/**
 * A subtotal field represents a running subtotal of a field, either
 * <code>ColumnField</code> or </code>FormulaField</code>. It also may be
 * associated with a group (assigned in <code>ReportReader.field</code> or
 * when editing a report), meaning that the running total is reset whenever
 * the group's value changes. The value of a subtotal field holds the id of
 * some other field whose value we are subtotaling.
 *
 * @see ColumnField
 * @see FormulaField
 * @see ReportReader
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class SubtotalField extends Field {

protected Group group;		// Set by report creation; possibly null
protected double total;		// Read-only
protected Field fieldToSubtotal;

/**
 * Returns <code>true</code> if this is a legal field to subtotal.
 * Called by the GUI to decide who's elligible for subtotaling.
 *
 * @param f a field
 * @return <code>true</code> if this is a legal field to subtotal
 */
public static boolean canSubtotal(Field f) {
    if (f instanceof SubtotalField) // All subtotal fields are OK
	return true;

    // All other fields must be in details. Section may be null during
    // a drag.
    Section s = f.getSection();
    if (s == null || !f.getSection().isDetail())
	return false;

    // All formula fields are OK. Not really (formula fields can evaluate
    // to strings), but we currently have no way of knowing the return type
    // of the formula.
    if (f instanceof FormulaField)
	return true;

    // Parameter fields: OK if numeric, else not OK.
    if (f instanceof ParameterField) {
	Parameter p = ((ParameterField)f).getParameter();
	return p.getType() == Parameter.TYPE_NUMERIC;
    }

    // Else, f must be a column field.
    if (!(f instanceof ColumnField))
	return false;

    // We have a column field. Only return true if it's a numeric type.
    int type = ((ColumnField)f).getColumn().getType();
    return type == Types.BIGINT
	|| type == Types.BIT
	|| type == Types.DECIMAL
	|| type == Types.DOUBLE
	|| type == Types.DOUBLE
	|| type == Types.INTEGER
	|| type == Types.NUMERIC
	|| type == Types.REAL
	|| type == Types.SMALLINT
	|| type == Types.TINYINT;
}

/**
 * Constructs a field with the specified id in the specified report
 * section that subtotals the field with id <i>value</i>.
 *
 * @param id the new field's id
 * @param report the report containing this element
 * @param section the report section in which the field resides
 * @param value the magic string
 * @param visible show/hide flag
 */
public SubtotalField(Long id, Report report, Section section, Object value,
		     boolean visible)
{
    super(id, report, section, value, visible);
    total = 0;

    // The reason I don't grab fieldToSubtotal right now is that this
    // subtotal field may be constructed before the field to which it
    // refers.
}

public void finalize() {
    if (fieldToSubtotal != null)
	fieldToSubtotal.deleteObserver(this);
}

/**
 * Resets this subtotal. Called by the report once at the beginning of
 * each run.
 */
public void initialize() {
    total = 0;
}

public String dragString() {
    return typeString() + ":" + getField().getId();
}

/**
 * Returns the group over which this field is subtotaling. May return
 * <code>null</code> since not all subtotal fields are associated with
 * a group.
 *
 * @return a group; <code>null</code> if this subtotal field is not
 * associated with any group
 */
public Group getGroup() { return group; }

/**
 * Sets the group this field is subtotaling. The group may be
 * <code>null</code>.
 *
 * @param newGroup a group
 */
public void setGroup(Group newGroup) { 
    if (group != newGroup) {
	group = newGroup;
	setChanged();
	notifyObservers();
    }
}

/**
 * Returns the field over which we are subtotaling. We lazily instantiate
 * it because when we are constructed that field may not yet exist. We also
 * start observing the field so we can notify our observers in turn.
 *
 * @return the field we are subtotaling
 */
public Field getField() {
    if (fieldToSubtotal == null) {
	fieldToSubtotal = getReport().findField(value);
	fieldToSubtotal.addObserver(this);
    }
    return fieldToSubtotal;
}

/**
 * Returns the id of the field over which we are subtotaling. The field
 * itself may not yet exist.
 *
 * @return the id of the field we are subtotaling
 */
public Long getFieldId() {
    return (value instanceof Long) ? (Long)value : new Long(value.toString());
}

/**
 * Returns the current running total value.
 *
 * @return a doubleing point total
 */
public double getTotal() { return total; }

public String typeString() { return "subtotal"; }

public String designLabel() {
    return "{sum " + getField().designLabel() + "}";
}

public String formulaString() { return designLabel(); }

public boolean refersTo(Field f) {
    return getField() == f;
}

public boolean refersTo(Formula f) {
    return (getField() instanceof FormulaField)
	&& ((FormulaField)getField()).getFormula() == f;
}

public boolean refersTo(UserColumn uc) {
    return (getField() instanceof UserColumnField)
	&& ((UserColumnField)getField()).getUserColumn() == uc;
}

public boolean refersTo(Parameter p) {
    if ((getField() instanceof ParameterField)
	&& ((ParameterField)getField()).getParameter() == p)
	return true;

    if ((getField() instanceof FormulaField)
	&& ((FormulaField)getField()).refersTo(p))
	return true;

    return false;
}

/**
 * Updates the running subtotal value. Called by the report when a new
 * line of data is retrieved.
 */
public void updateSubtotal() {
    /*
     * Our value field holds the id of some other field. Get that field's
     * value, then convert it to a double.
     */
    Object obj = getField().getValue();
    double value = 0;
    if (obj != null) {
	if (obj instanceof Number)
	    value = ((Number)obj).doubleValue();
	else
	    value = Double.parseDouble(obj.toString());
    }

    // If we are subtotaling within a group and this is a new value,
    // reset the running total value.
    if (group != null && group.isNewValue())
	total = value;
    else
	total += value;
}

/**
 * Returns the value of this field: the subtotal as a Double.
 *
 * @return a Double
 */
public Object getValue() { return new Double(total); }

}
