package jimm.datavision;
import jimm.datavision.sql.Column;
import java.io.*;
import java.util.*;

/**
 * A user column is an arbitrary expression inserted into the SQL query
 * and retrieved as a column value. It may contain database
 * column values, parameters, special values, strings, or
 * numbers. It can't contain formulas because their values may be
 * undefined when the query is run.
 * <p>
 * When used by a query, the following substitutions are made withing
 * the <var>userString</var> of a user column:
 * <ul>
 * <li>{<i>table_name.column_name</i>} is replaced by the column name
 * <i>table_name.column_name</i>.</li>
 * <li> {%<i>special_value_name</i>} is replaced by a special value
 * (report title, report run date, page number, or record number).</li>
 * <li> {?<i>id_number</i>} is replaced by a parameter value (string,
 * number, or date).</li>
 * <ul>
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class UserColumn
    extends Observable
    implements Nameable, Writeable, Draggable, Observer
{

protected Long id;
protected Report report;
protected String name;
protected String userString;
protected ArrayList observedContents;

/**
 * Constructor.
 *
 * @param id the unique identifier for the new user column; if
 * <code>null</code>, generate a new id
 * @param report the report containing this user column
 * @param name the user column name
 */
public UserColumn(Long id, Report report, String name) {
    this(id, report, name, null);
}

/**
 * Constructor. If <i>id</i> is <code>null</code>, generates a new id number.
 * This number is one higher than any previously-seen id number. This does
 * <em>not</em> guarantee that no later user column will be created manually
 * with the same id number.
 *
 * @param id the unique identifier for the new user column; if
 * <code>null</code>, generate a new id
 * @param report the report containing this user column
 * @param name the user column name
 * @param userString the string to evaulate at runtime.
 */
public UserColumn(Long id, Report report, String name, String userString) {
    if (id == null)
	id = report.generateNewUserColumnId();

    this.report = report;
    this.id = id;
    this.name = name;
    this.userString = userString;

    // I'd like to start observing the contents of the eval string here,
    // but the other user columns may not yet be defined (for example, when
    // reading in user columns from an XML file). That's why we start observing
    // contents when someone asks for the eval string.
    observedContents = null;
}

public void finalize() {
    stopObservingContents();
    try {
	super.finalize();
    }
    catch (Throwable t) {}
}

public void update(Observable o, Object arg) {
    setChanged();
    notifyObservers(arg);
}

/**
 * Returns the unique id for this field.
 *
 * @return the unique id
 */
public Long getId() { return id; }

/**
 * Returns the name for this user column.
 *
 * @return the name
 */
public String getName() { return name; }

/**
 * Sets the name.
 *
 * @param newName the new name
 */
public void setName(String newName) {
    if (name != newName && (name == null || !name.equals(newName))) {
	name = newName;
	setChanged();
	notifyObservers();
    }
}

/**
 * Returns the user string for this user column.
 *
 * @return the user string
 */
public String getUserString() {
    // I'd like to start observing the contents of the user string as soon
    // as we are constructed, but the other user columns may not yet be defined
    // (for example, when reading in user columns from an XML file). That's why
    // we start observing contents when someone asks for the user string.
    if (observedContents == null)
	startObservingContents();

    return userString;
}

/**
 * Sets the user string.
 *
 * @param newUserString the new user string
 */
public void setUserString(String newUserString) {
    if (userString != newUserString
	&& (userString == null || !userString.equals(newUserString)))
    {
	stopObservingContents();
	userString = newUserString;

	// Don't start observing contents yet. Wait until someone calls
	// getUserString(). See the comment there.
//  	startObservingContents();

	setChanged();
	notifyObservers();
    }
}

/**
 * Starts observing all user columns and parameters referenced by this
 * user column.
 */
protected void startObservingContents() {
    observedContents = new ArrayList();	// Even if userString is null

    if (userString == null || userString.length() == 0)
	return;

    // Parameters
    int pos = 0;
    while ((pos = userString.indexOf("{?", pos)) != -1) {
	int pos2 = userString.indexOf("}", pos + 1);
	if (pos2 != -1) {
	    Parameter p = report.findParameter(userString.substring(pos + 2,
								    pos2));
	    observedContents.add(p);
	    p.addObserver(this);
	}
	pos = pos2 + 1;
    }
}

/**
 * Stops observing all user columns referenced by this user column.
 */
protected void stopObservingContents() {
    if (observedContents != null) {
	for (Iterator iter = observedContents.iterator(); iter.hasNext(); )
	    ((Observable)iter.next()).deleteObserver(this);
	observedContents = null;
    }
}

/**
 * Returns the user string fit for human consumption. This mainly means
 * that we substitute user column and parameter numbers with names. Called
 * from a user column editor. This code assumes that curly braces are never
 * nested.
 *
 * @return the user string with user column and parameter id numbers replaced
 * with names
 * @see jimm.datavision.gui.FormulaWin
 */
public String getEditableUserString() {
    return Utils.formulaToDisplay(report, getUserString());
}

/**
 * Sets the user string after replacing formula names with formula id
 * numbers. Called from a formula editor.
 * <p>
 * This method will throw an <code>IllegalArgumentException</code> if any
 * formula name or parameter name is not the name of some existing formula
 * or parameter.
 *
 * @param newUserString the new user string
 * @see jimm.datavision.gui.FormulaWin#save
 * @throws NullPointerException
 */
public void setEditableUserString(String newUserString) {
    setUserString(Utils.displayToFormula(report, newUserString));
}

public String dragString() {
    return "usercol:" + getId();
}

public String designLabel() {
    return "{!" + getName() + "}";
}

public String formulaString() {
    return "{!" + getId() + "}";
}

/**
 * Returns <code>true</code> if this user column contains a reference to the
 * specified field.
 *
 * @param f a field
 * @return <code>true</code> if this field contains a reference to the
 * specified field
 * @see UserColumnField
 * @see SubtotalField
 */
public boolean refersTo(Field f) {
    String str = getUserString();
    return str.indexOf(f.formulaString()) != -1;
}

/**
 * Returns <code>true</code> if this user column contains a reference to the
 * specified parameter.
 *
 * @param p a parameter
 * @return <code>true</code> if this user column contains a reference to the
 * specified parameter
 * @see FormulaField
 * @see SubtotalField
 */
public boolean refersTo(Parameter p) {
    String str = getUserString();
    return str.indexOf(p.formulaString()) != -1;
}

/**
 * Returns a collection of the columns used in the user column. This is used
 * by the report's query when it is figuring out what columns and tables
 * are used by the report. Called from <code>FormulaField.columnsUsed</code>.
 *
 * @return a possibly empty collection of database columns
 * @see FormulaField#columnsUsed
 * @see Query#findColumnsUsed
 */
public Collection columnsUsed() {
    ArrayList list = new ArrayList();
    int pos = 0;
    String str = getUserString();
    while ((pos = str.indexOf("{", pos)) != -1) {
	int pos2 = str.indexOf("}", pos + 1);
	char c = str.charAt(pos + 1);
	if (c != '%' && c != '?') {
	    // Ignore special fields and parameters
	    Column col = report.findColumn(str.substring(pos + 1, pos2));
	    if (col != null)	// Should never be null
		list.add(col);
	}
	pos = pos2 + 1;
    }
    return list;
}

/**
 * Returns this user column's SQL text as entered by the user.
 */
public String toString() {
    return getUserString();
}

/**
 * Writes this user column as an XML tag.
 *
 * @param out a writer that knows about indenting XML tags
 */
public void writeXML(IndentWriter out) {
    out.println("<usercol id=\"" + id + "\" name=\"" + Utils.escapeXML(name)
		+ "\"><![CDATA[" + getUserString() + "]]></usercol>");
}

}
