package jimm.datavision;
import java.util.*;

/**
 * Globally available utility classes.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class Utils {

/**
 * Returns a string consisting of all members of a collection separated
 * by the specified string. The <code>toString</code> method of each
 * collection member is called to convert it to a string.
 *
 * @param c a collection of objects
 * @param joinWith the string that will separate each member of the collection
 */
public static String join(Collection c, String joinWith) {
    if (c == null)
	return "";

    StringBuffer buf = new StringBuffer();
    boolean first = true;
    for (Iterator iter = c.iterator(); iter.hasNext(); ) {
	if (first) first = false;
	else if (joinWith != null) buf.append(joinWith);
	buf.append(iter.next().toString());
    }
    return buf.toString();
}

/**
 * Returns a string with HTML special characters replaced by their entity
 * equivalents.
 *
 * @param str the string to escape
 * @return a new string without HTML special characters
 */
public static String escapeHTML(String str) {
    if (str == null || str.length() == 0)
	return "";

    StringBuffer buf = new StringBuffer();
    int len = str.length();
    for (int i = 0; i < len; ++i) {
	char c = str.charAt(i);
	switch (c) {
	case '&': buf.append("&amp;"); break;
	case '<': buf.append("&lt;"); break;
	case '>': buf.append("&gt;"); break;
	case '"': buf.append("&quot;"); break;
	case '\'': buf.append("&apos;"); break;
	default: buf.append(c); break;
	}
    }
    return buf.toString();
}

/**
 * Returns a string with XML special characters replaced by their entity
 * equivalents.
 *
 * @param str the string to escape
 * @return a new string without XML special characters
 */
public static String escapeXML(String str) {
    return escapeHTML(str);
}

/**
 * Returns a string with XML entities replaced by their normal characters.
 *
 * @param str the string to un-escape
 * @return a new normal string
 */
public static String unescapeXML(String str) {
    if (str == null || str.length() == 0)
	return "";

    StringBuffer buf = new StringBuffer();
    int len = str.length();
    for (int i = 0; i < len; ++i) {
	char c = str.charAt(i);
	if (c == '&') {
	    int pos = str.indexOf(";", i);
	    if (pos == -1) {	// Really evil
		buf.append('&');
	    }
	    else if (str.charAt(i+1) == '#') {
		int val = Integer.parseInt(str.substring(i+2, pos), 16);
		buf.append((char)val);
		i = pos;
	    }
	    else {
		String substr = str.substring(i, pos + 1);
		if (substr.equals("&amp;"))
		    buf.append('&');
		else if (substr.equals("&lt;"))
		    buf.append('<');
		else if (substr.equals("&gt;"))
		    buf.append('>');
		else if (substr.equals("&quot;"))
		    buf.append('"');
		else if (substr.equals("&apos;"))
		    buf.append('\'');
		else		// ????
		    buf.append(substr);
		i = pos;
	    }
	}
	else {
	    buf.append(c);
	}
    }
    return buf.toString();
}

/**
 * Given a string, returns a string with all instances of formula,
 * parameter, and user column "formula strings" replaced by "display name"
 * strings. If there are no such strings, the original string is returned.
 *
 * @return a string with all formula strings replaced by display name
 * strings
 */
public static String formulaToDisplay(Report report, String str) {
    if (str == null || str.length() == 0 || str.indexOf("{") == -1)
	return str;

    StringBuffer buf = new StringBuffer();
    int len = str.length();
    for (int i = 0; i < len; ++i) {
	char c = str.charAt(i);
	if (c == '{' && (i + 1) < len) {
	    int nameStart, nameEnd;
	    switch (str.charAt(i + 1)) {
	    case '@':		// Formula
		nameStart = i + 2;
		nameEnd = str.indexOf("}", nameStart);
		if (nameEnd != -1) {
		    String idAsString = str.substring(nameStart, nameEnd);
		    buf.append("{@");
		    buf.append(report.findFormula(idAsString).getName());
		    buf.append("}");
		    i = nameEnd;
		}
		break;
	    case '?':		// Parameter
		nameStart = i + 2;
		nameEnd = str.indexOf("}", nameStart);
		if (nameEnd != -1) {
		    String idAsString = str.substring(nameStart, nameEnd);
		    buf.append("{?");
		    buf.append(report.findParameter(idAsString).getName());
		    buf.append("}");
		    i = nameEnd;
		}
		break;
	    case '!':		// User column
		nameStart = i + 2;
		nameEnd = str.indexOf("}", nameStart);
		if (nameEnd != -1) {
		    String idAsString = str.substring(nameStart, nameEnd);
		    buf.append("{!");
		    buf.append(report.findUserColumn(idAsString).getName());
		    buf.append("}");
		    i = nameEnd;
		}
		break;
	    default:
		buf.append(c);
		break;
	    }
	}
	else {
	    buf.append(c);
	}
    }
    return buf.toString();
}

/**
 * Given a string, returns a string with all instances of formula,
 * parameter, and user column "display names" replaced by "formula
 * strings". If there are no such strings, the original string is returned.
 *
 * @param report a report
 * @param str a string with display names
 * @return a string with all display names replaced by formula strings
 */
public static String displayToFormula(Report report, String str) {
    if (str == null || str.length() == 0
	|| (str.indexOf("{@") == -1 && str.indexOf("{?") == -1))
	return str;

    StringBuffer buf = new StringBuffer();
    int len = str.length();
    for (int i = 0; i < len; ++i) {
	char c = str.charAt(i);
	if (c == '{' && (i + 1) < len) {
	    int nameStart, nameEnd;
	    switch (str.charAt(i + 1)) {
	    case '@':		// Formula
		nameStart = i + 2;
		nameEnd = str.indexOf("}", nameStart);
		if (nameEnd != -1) {
		    String formulaName = str.substring(nameStart, nameEnd);
		    buf.append("{@");
		    Formula formula = report.findFormulaByName(formulaName);
		    if (formula == null) {
			str = I18N.get("Utils.in")
			    + " \"" + str + "\": "
			    + I18N.get("Utils.no_such_formula")
			    + ' ' + formulaName;
			throw new IllegalArgumentException(str);
		    }
		    buf.append(formula.getId());
		    buf.append("}");
		    i = nameEnd;
		}
		break;
	    case '?':		// Parameter
		nameStart = i + 2;
		nameEnd = str.indexOf("}", nameStart);
		if (nameEnd != -1) {
		    String paramName = str.substring(nameStart, nameEnd);
		    buf.append("{?");
		    Parameter param = report.findParameterByName(paramName);
		    if (param == null) {
			str = I18N.get("Utils.in")
			    + " \"" + str + "\": "
			    + I18N.get("Utils.no_such_param")
			    + ' ' + paramName;
			throw new IllegalArgumentException(str);
		    }
		    buf.append(param.getId());
		    buf.append("}");
		    i = nameEnd;
		}
		break;
	    case '!':		// User column
		nameStart = i + 2;
		nameEnd = str.indexOf("}", nameStart);
		if (nameEnd != -1) {
		    String ucName = str.substring(nameStart, nameEnd);
		    buf.append("{!");
		    UserColumn uc = report.findUserColumnByName(ucName);
		    if (uc == null) {
			str = I18N.get("Utils.in")
			    + " \"" + str + "\": "
			    + I18N.get("Utils.no_such_usercol")
			    + ' ' + ucName;
			throw new IllegalArgumentException(str);
		    }
		    buf.append(uc.getId());
		    buf.append("}");
		    i = nameEnd;
		}
		break;
	    default:
		buf.append(c);
		break;
	    }
	}
	else {
	    buf.append(c);
	}
    }

    return buf.toString();
}

}
