package jimm.datavision.gui;
import jimm.datavision.Report;
import jimm.datavision.I18N;
import jimm.datavision.ErrorHandler;
import jimm.datavision.sql.Database;
import jimm.datavision.gui.cmd.DbConnCommand;
import java.io.IOException;
import java.awt.BorderLayout;
import java.awt.event.*;
import javax.swing.*;
import org.xml.sax.SAXException;

/**
 * A database connection editing dialog box. The user can either enter
 * separate values or copy values from an existing report XML file.
 *
 * @see DbConnReader
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class DbConnWin extends EditWin implements ActionListener {

protected static final int TEXT_FIELD_COLS = 32;

protected Report report;
protected JTextField driverClassNameField;
protected JTextField connInfoField;
protected JTextField dbNameField;
protected JTextField userNameField;
protected JPasswordField passwordField;

/**
 * Constructor.
 *
 * @param designWin the window to which this dialog belongs
 * @param database report database connection
 */
public DbConnWin(DesignWin designWin, Report report, boolean modal) {
    super(designWin, I18N.get("DbConnWin.title"), "DbConnCommand.name", modal);

    this.report = report;

    buildWindow();
    pack();
    setVisible(true);
}

/**
 * Builds the window contents.
 */
protected void buildWindow() {
    JPanel editorPanel = buildEditor();

    // OK, Apply, Revert, and Cancel Buttons
    JPanel buttonPanel = closeButtonPanel();

    // Add values and buttons to window
    getContentPane().add(editorPanel, BorderLayout.CENTER);
    getContentPane().add(buttonPanel, BorderLayout.SOUTH);

    new FocusSetter(driverClassNameField);
}

protected JPanel buildEditor() {
    Database db = report.getDatabase();
    EditFieldLayout efl = new EditFieldLayout();
    efl.setBorder(20);

    driverClassNameField =
	efl.addTextField(I18N.get("DbConnWin.driver_class_name"), db == null
			 ? "" : db.getDriverClassName(), TEXT_FIELD_COLS);
    connInfoField = efl.addTextField(I18N.get("DbConnWin.connection_info"),
				     db == null ? "" : db.getConnectionInfo(),
				     TEXT_FIELD_COLS);
    dbNameField = efl.addTextField(I18N.get("DbConnWin.database_name"),
				   db == null ? "" : db.getName(),
				   TEXT_FIELD_COLS);
    userNameField = efl.addTextField(I18N.get("DbConnWin.user_name"),
				     db == null ? "" : db.getUserName(),
				     TEXT_FIELD_COLS);
    String password = (db == null ? "" : db.getPassword());
    if (password == null) password = "";
    passwordField = efl.addPasswordField(I18N.get("DbConnWin.password"),
					 password, TEXT_FIELD_COLS);

    // Click to copy info from another report
    JButton copyButton = new JButton(I18N.get("DbConnWin.copy_settings"));
    copyButton.addActionListener(this);
    JPanel copyPanel = new JPanel();
    copyPanel.add(copyButton);
    efl.add(null, copyPanel);

    return efl.getPanel();
}

protected void fillEditFields() {
    Database db = report.getDatabase();
    if (db == null) {
	driverClassNameField.setText("");
	connInfoField.setText("");
	dbNameField.setText("");
	userNameField.setText("");
    }
    else {
	driverClassNameField.setText(db.getDriverClassName());
	connInfoField.setText(db.getConnectionInfo());
	dbNameField.setText(db.getName());
	userNameField.setText(db.getUserName());
    }
    passwordField.setText("");
}

/**
 * Handles the "Copy Settings..." button.
 *
 * @param e action event
 */
public void actionPerformed(ActionEvent e) {
    String cmd = e.getActionCommand();
    if (I18N.get("DbConnWin.copy_settings").equals(cmd)) {
	JFileChooser chooser = designWin.getChooser();
	int returnVal = chooser.showOpenDialog(this);
	if (returnVal == JFileChooser.APPROVE_OPTION) {
	    String path = chooser.getSelectedFile().getPath();
	    DbConnReader reader = new DbConnReader();
	    try {
		reader.readFile(path);
		driverClassNameField.setText(reader.getDriverClassName());
		connInfoField.setText(reader.getConnectionInfo());
		dbNameField.setText(reader.getDbName());
		userNameField.setText(reader.getUserName());
	    }
	    catch (IOException ioe) {
		ErrorHandler.error(I18N.get("DbConnWin.copy_error"), ioe);
	    }
	    catch (SAXException saxe) {
		ErrorHandler.error(I18N.get("DbConnWin.copy_error"), saxe);
	    }
	}
    }
    else
	super.actionPerformed(e);
}

protected void doSave() {
    DbConnCommand cmd =
	new DbConnCommand(report, driverClassNameField.getText(),
			  connInfoField.getText(), dbNameField.getText(),
			  userNameField.getText(),
			  new String(passwordField.getPassword()));
    cmd.perform();
    commands.add(cmd);
}

protected void doRevert() {
    fillEditFields();
}

}
