package jimm.datavision.gui;
import jimm.datavision.*;
import jimm.datavision.layout.swing.SwingLE;
import jimm.datavision.gui.cmd.*;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.print.Paper;
import java.util.*;
import javax.swing.*;
import javax.swing.event.MouseInputAdapter;

/**
 * The main GUI report design window.
 *
 * @see Report
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class DesignWin extends JFrame implements ActionListener, Observer {

public static final int GRID_SIZE = 8;

public static final int ALIGN_TOP = 0;
public static final int ALIGN_MIDDLE = 1;
public static final int ALIGN_BOTTOM = 2;
public static final int ALIGN_LEFT = 3;
public static final int ALIGN_CENTER = 4;
public static final int ALIGN_RIGHT = 5;
public static final int ALIGN_SNAP_TO_GRID = 6;

public static final int SIZE_SAME_WIDTH = 0;
public static final int SIZE_SAME_HEIGHT = 1;
public static final int SIZE_SAME_SIZE = 2;

protected static final String REPORT_TEMP_DIR = "datavision";
protected static ArrayList designWindows = new ArrayList();

protected Report report;
protected CommandHistory commandHistory;
protected ArrayList sectionWidgets;
protected String reportFileName;
protected JFileChooser chooser;
protected JLayeredPane sectionContainer;
protected ArrayList selectedFields;
protected JScrollPane scroller;
protected JMenuItem undoItem, redoItem, cutItem, copyItem, pasteItem,
    delSelectionItem, delGroupItem, delSectionItem, subtotItem, groupItem,
    sectItem, formatItem, borderItem, boundsItem;
protected JMenu alignSubmenu, sizeSubmenu, paperSizeSubmenu;
protected boolean placingNewTextField;
protected boolean ignoreKeys;

/**
 * Adds a window to the list of open design windows. Also notifies the
 * error handler class that we it should use the GUI to display error
 * messages.
 *
 * @param win a design window
 */
public static void addWindow(DesignWin win) {
    designWindows.add(win);
}

/**
 * Returns the design window associated with the specified report, or
 * <code>null</code> if one is not found.
 *
 * @param r a report
 * @return a design window or <code>null</code> if one is not found
 */
public static DesignWin findWindowFor(Report r) {
    for (Iterator iter = designWindows.iterator(); iter.hasNext(); ) {
	DesignWin win = (DesignWin)iter.next();
	if (win.report == r)
	    return win;
    }
    return null;
}

/**
 * Deletes a window from the list of windows. If there are no more design
 * windows left, exits the application.
 *
 * @param win a design window
 */
protected static void deleteWindow(DesignWin win) {
    designWindows.remove(win);
    if (designWindows.isEmpty())
	System.exit(0);
}

/**
 * Closes each open design window (unless user cancels). If all are closed,
 * quits the application.
 */
protected static void maybeQuit() {
    // Can't use iterator over original list 'cause we're deleting items
    // from the list.
    ArrayList copy = (ArrayList)designWindows.clone();
    for (Iterator iter = copy.iterator(); iter.hasNext(); )
	((DesignWin)iter.next()).maybeClose();
}

/**
 * Constructor. Reads the named report file or, if it's <code>null</code>,
 * creates a new, empty report.
 *
 * @param fileName an XML report file name; may be <code>null</code>
 */
public DesignWin(String fileName) {
    this(fileName, null);
}

/**
 * Constructor. Reads the named report file or, if it's <code>null</code>,
 * creates a new, empty report.
 *
 * @param fileName an XML report file name; may be <code>null</code>
 * @param databasePassword string to give to report; OK if it's
 * <code>null</code>
 */
public DesignWin(String fileName, String databasePassword) {
    super(I18N.get("DesignWin.title"));
    ErrorHandler.useGUI(true);	// Must set before opening report
    selectedFields = new ArrayList();

    // Must create command history before asking for database information
    // because that can cause a command to be created.
    commandHistory = new CommandHistory();

    StatusDialog statusDialog =
	new StatusDialog(this, I18N.get("DesignWin.status_title"), false,
			 fileName == null
			 ? I18N.get("DesignWin.creating_empty")
			 : I18N.get("DesignWin.reading_xml"));

    boolean askForDbConnInfo = readReport(fileName, databasePassword);

    if (report == null || (report.getDatabase() == null && !askForDbConnInfo))
    {
	// User cancelled when asked for password. Close this window and
	// quit if this is the only report open.
	dispose();
	statusDialog.dispose();
	DesignWin.deleteWindow(this); // Will quit if no other windows
	return;
    }

    if (askForDbConnInfo) {
	openDbConnWin(true);	// Modally ask for database connection info
	if (report.getDatabase() == null) {
	    // User cancelled. Close this window and quit if this is the
	    // only report open.
	    dispose();
	    statusDialog.dispose();
	    DesignWin.deleteWindow(this); // Will quit if no other windows
	    return;
	}
    }

    statusDialog.update(I18N.get("DesignWin.building_win"));

    setDefaultCloseOperation(DO_NOTHING_ON_CLOSE);
    buildWindow();
    commandHistory.setMenuItems(undoItem, redoItem);
    enableMenuItems();

    // Make sure we close the window when the user asks to close the window.
    addWindowListener(new WindowAdapter() {
	public void windowClosing(WindowEvent e) { maybeClose(); }
	public void windowActivated(WindowEvent e) { enableMenuItems(); }
    });

    // Might be using this code for dragging to select multiple fields.
//      SelectionHandler h = new SelectionHandler();
//      addMouseListener(h);
//      addMouseMotionListener(h);

    pack();
    statusDialog.dispose();
    show();
    DesignWin.addWindow(this);
}

public void update(Observable o, Object arg) {
    enableMenuItems();
}

/**
 * Performs a command.
 *
 * @param cmd a command
 */
public void performCommand(Command cmd) {
    commandHistory.perform(cmd);
}

/**
 * Adds to the command history a command that has already been performed.
 *
 * @param cmd a command
 */
public void addCommand(Command cmd) {
    commandHistory.add(cmd);
}

/**
 * Reads the named report file or, if it's <code>null</code>, creates
 * a new, empty report. Returns <code>true</code> if we need to ask
 * the user for connection info because this is a new report.
 *
 * @param fileName
 * @param databasePassword string to give to report; OK if it's
 * <code>null</code>
 * @return <code>true</code> if we need to ask the user for connection info
 */
protected boolean readReport(String fileName, String databasePassword) {
    if (fileName != null && fileName.length() > 0) {
	reportFileName = fileName;
	report = new Report();
	report.setDatabasePassword(databasePassword);
	try {
	    report.readFile(fileName);
	    return false;
	}
	catch (UserCancellationException uce) {
	    report = null;	// Signal report open cancelled
	    return true;	// Doesn't matter what we return
	}
	catch (Exception e) {
	    ErrorHandler.error(e);
	    report = new Report();
	    return true;
	}
    }
    else {
	report = new Report();
	reportFileName = null;
	return true;
    }
}

public void setIgnoreKeys(boolean ignore) { ignoreKeys = ignore; }

/**
 * Builds the window components.
 */
protected void buildWindow() {
    buildMenuBar();
    buildSections();
    // Set up key handling (also put inside section canvas?)
    addKeyListener(new KeyAdapter() {
	public void keyPressed(KeyEvent e) {
	    if (!ignoreKeys) {
		int code = e.getKeyCode();
		if (code == KeyEvent.VK_BACK_SPACE
		    || code == KeyEvent.VK_DELETE)
		    deleteSelectedFields();
	    }
	}
	});
}

/**
 * Builds the window menu bar.
 */
protected void buildMenuBar() {
    JMenuBar bar = new JMenuBar();
    bar.add(buildFileMenu());
    bar.add(buildEditMenu());
    bar.add(buildInsertMenu());
    bar.add(buildFormatMenu());
    bar.add(buildDatabaseMenu());
    bar.add(buildReportMenu());
    bar.add(Box.createHorizontalGlue());
    bar.add(buildHelpMenu());
    setJMenuBar(bar);
}

/**
 * Builds and returns the "File" menu.
 *
 * @return a menu
 */
protected JMenu buildFileMenu() {
    JMenu menu = new JMenu(I18N.get("DesignWin.menu_file"));
    menu.setMnemonic(KeyEvent.VK_F);

    MenuUtils.addToMenu(this, menu, I18N.get("DesignWin.menu_file_new"), "new",
			KeyEvent.VK_N);
    MenuUtils.addToMenu(this, menu, I18N.get("DesignWin.menu_file_open"), "open",
			KeyEvent.VK_O);
    menu.addSeparator();
    MenuUtils.addToMenu(this, menu, I18N.get("DesignWin.menu_file_save"), "save",
			KeyEvent.VK_S);
    MenuUtils.addToMenu(this, menu, I18N.get("DesignWin.menu_file_save_as"), "save_as",
			KeyEvent.VK_A, 0);

    menu.addSeparator();
    ButtonGroup orientationGroup = new ButtonGroup();
    ButtonGroup nameGroup = new ButtonGroup();
    menu.add(paperSizeSubmenu =
	     MenuUtils.buildPaperSizeMenu(this,
					  report.getPaperFormat(),
					  orientationGroup, nameGroup,
					  KeyEvent.VK_P, null));

    menu.addSeparator();
    MenuUtils.addToMenu(this, menu, I18N.get("DesignWin.menu_file_close"), "close",
			KeyEvent.VK_C, KeyEvent.VK_W);
    MenuUtils.addToMenu(this, menu, I18N.get("DesignWin.menu_file_quit"), "quit",
			KeyEvent.VK_Q);

    return menu;
}

/**
 * Builds and returns the "Edit" menu.
 *
 * @return a menu
 */
protected JMenu buildEditMenu() {
    JMenu menu = new JMenu(I18N.get("DesignWin.menu_edit"));
    menu.setMnemonic(KeyEvent.VK_E);

    undoItem = MenuUtils.addToMenu(this, menu,
				   I18N.get("DesignWin.menu_edit_undo"),
				   "undo", KeyEvent.VK_U, KeyEvent.VK_Z);
    undoItem.setEnabled(false);
    redoItem = MenuUtils.addToMenu(this, menu,
				   I18N.get("DesignWin.menu_edit_redo"),
				   "redo", KeyEvent.VK_R, 0);
    redoItem.setEnabled(false);
    menu.addSeparator();
    cutItem = MenuUtils.addToMenu(this, menu,
				  I18N.get("DesignWin.menu_edit_cut"),
				  "cut", KeyEvent.VK_C, KeyEvent.VK_X);
    copyItem = MenuUtils.addToMenu(this, menu,
				   I18N.get("DesignWin.menu_edit_copy"),
				   "copy", KeyEvent.VK_O, KeyEvent.VK_C);
    copyItem.setEnabled(false);
    pasteItem = MenuUtils.addToMenu(this, menu,
				    I18N.get("DesignWin.menu_edit_paste"),
				    "paste", KeyEvent.VK_P, KeyEvent.VK_V);
    pasteItem.setEnabled(false);
    menu.addSeparator();
    delSelectionItem = MenuUtils.addToMenu(this, menu,
					   I18N.get("DesignWin.menu_edit_del_fields"),
					   "del_selection", KeyEvent.VK_F,
					   KeyEvent.VK_DELETE, 0);
    menu.addSeparator();
    delGroupItem = MenuUtils.addToMenu(this, menu,
				       I18N.get("DesignWin.menu_edit_del_group"),
				       "del_group", KeyEvent.VK_G, 0);
    delSectionItem = MenuUtils.addToMenu(this, menu,
					 I18N.get("DesignWin.menu_edit_del_section"),
					 "del_section", KeyEvent.VK_S, 0);

    return menu;
}

/**
 * Builds and returns the "Insert" menu.
 *
 * @return a menu
 */
protected JMenu buildInsertMenu() {
    JMenu menu = new JMenu(I18N.get("DesignWin.menu_insert"));
    menu.setMnemonic(KeyEvent.VK_I);

    MenuUtils.addToMenu(this, menu, I18N.get("DesignWin.menu_insert_column"),
			"database_field", KeyEvent.VK_D, 0);
    MenuUtils.addToMenu(this, menu, I18N.get("DesignWin.menu_insert_text"),
			"text", KeyEvent.VK_T, 0);
    MenuUtils.addToMenu(this, menu, I18N.get("DesignWin.menu_insert_formula"),
			"formula_field", KeyEvent.VK_F, 0);
    MenuUtils.addToMenu(this, menu, I18N.get("DesignWin.menu_insert_param"),
			"parameter_field", KeyEvent.VK_P, 0);
    MenuUtils.addToMenu(this, menu, I18N.get("DesignWin.menu_insert_usercol"),
			"usercol_field", KeyEvent.VK_U, 0);
    MenuUtils.addToMenu(this, menu, I18N.get("DesignWin.menu_insert_special"),
			"special_field", KeyEvent.VK_S, 0);
    subtotItem = MenuUtils.addToMenu(this, menu,
				     I18N.get("DesignWin.menu_insert_subtot"),
				     "running_total_field", KeyEvent.VK_R, 0);
    menu.addSeparator();
    MenuUtils.addToMenu(this, menu, I18N.get("DesignWin.menu_insert_image"),
			"image", KeyEvent.VK_I, 0);
    MenuUtils.addToMenu(this, menu, I18N.get("DesignWin.menu_insert_line"),
			"line", KeyEvent.VK_L, 0)
	.setEnabled(false);
    menu.addSeparator();
    groupItem = MenuUtils.addToMenu(this, menu, I18N.get("DesignWin.menu_insert_group"),
				    "add_group", KeyEvent.VK_G, 0);
    sectItem = MenuUtils.addToMenu(this, menu, I18N.get("DesignWin.menu_insert_section"),
				   "ins_section", KeyEvent.VK_E, 0);

    return menu;
}

/**
 * Builds and returns the "Format" menu.
 *
 * @return a menu
 */
protected JMenu buildFormatMenu() {
    JMenu menu = new JMenu(I18N.get("DesignWin.menu_format"));
    menu.setMnemonic(KeyEvent.VK_O);

    formatItem = MenuUtils.addToMenu(this, menu,
				     I18N.get("DesignWin.menu_format_format"),
				     "format_field", KeyEvent.VK_F, 0);
    borderItem = MenuUtils.addToMenu(this, menu,
				     I18N.get("DesignWin.menu_format_border"),
				     "border", KeyEvent.VK_B, 0);
    menu.addSeparator();
    boundsItem = MenuUtils.addToMenu(this, menu,
				     I18N.get("DesignWin.menu_format_bounds"),
				     "position", KeyEvent.VK_P, 0);
    menu.add(alignSubmenu = MenuUtils.buildAlignMenu(this, true, null));
    menu.add(sizeSubmenu = MenuUtils.buildSizeMenu(this, true, null));

    return menu;
}

/**
 * Builds and returns the "Database" menu.
 *
 * @return a menu
 */
protected JMenu buildDatabaseMenu() {
    JMenu menu = new JMenu(I18N.get("DesignWin.menu_database"));
    menu.setMnemonic(KeyEvent.VK_D);

    MenuUtils.addToMenu(this, menu, I18N.get("DesignWin.menu_database_linker"),
			"visual_table_linker", KeyEvent.VK_V, 0);
    menu.addSeparator();
//      MenuUtils.addToMenu(this, menu, I18N.get("DesignWin.menu_database_join"),
//  			"join_tables", KeyEvent.VK_J, 0)
//  	.setEnabled(false);
//      menu.addSeparator();
    MenuUtils.addToMenu(this, menu, I18N.get("DesignWin.menu_database_sql"),
			"sql_query_text", KeyEvent.VK_S, 0);
    menu.addSeparator();
    MenuUtils.addToMenu(this, menu, I18N.get("DesignWin.menu_connection"),
			"db_connection", KeyEvent.VK_C, 0);

    return menu;
}

/**
 * Builds and returns the "Report" menu.
 *
 * @return a menu
 */
protected JMenu buildReportMenu() {
    JMenu menu = new JMenu(I18N.get("DesignWin.menu_report"));
    menu.setMnemonic(KeyEvent.VK_R);

    MenuUtils.addToMenu(this, menu, I18N.get("DesignWin.menu_report_run"),
			"run", KeyEvent.VK_R);
    MenuUtils.addToMenu(this, menu, I18N.get("DesignWin.menu_report_export"),
			"export", KeyEvent.VK_E, 0);
    menu.addSeparator();
    MenuUtils.addToMenu(this, menu, I18N.get("DesignWin.menu_report_select"),
			"select_records", KeyEvent.VK_S, 0);
    MenuUtils.addToMenu(this, menu, I18N.get("DesignWin.menu_report_sort"),
			"sort_by", KeyEvent.VK_O, 0);
    MenuUtils.addToMenu(this, menu, I18N.get("DesignWin.menu_report_group"),
			"group_by", KeyEvent.VK_G, 0);
    menu.addSeparator();
    MenuUtils.addToMenu(this, menu, I18N.get("DesignWin.menu_report_summary"),
			"summary", KeyEvent.VK_U, 0);

    return menu;
}

/**
 * Builds and returns the "Help" menu.
 *
 * @return a menu
 */
protected JMenu buildHelpMenu() {
    JMenu menu = new JMenu(I18N.get("DesignWin.menu_help"));
    menu.setMnemonic(KeyEvent.VK_H);

    MenuUtils.addToMenu(this, menu, I18N.get("DesignWin.menu_help_help"), "help",
			KeyEvent.VK_H, 0);
    menu.addSeparator();
    MenuUtils.addToMenu(this, menu, I18N.get("DesignWin.menu_help_about"), "about",
			KeyEvent.VK_A, 0);

    return menu;
}

/**
 * Builds the sections and adds them to the section container.
 */
protected void buildSections() {
    getContentPane().setLayout(new BorderLayout());

    sectionWidgets = new ArrayList();
    sectionContainer = new JLayeredPane();
    sectionContainer.setLayout(new DesignWinLayout());

    // Can't just call report.each_section because we want to append
    // '(a)', '(b)', etc. if there is more than one section in the group.
    buildSectionsInList(report.headers());
    buildSectionsInList(report.pageHeaders());

    // Add headers for each group.
    for (Iterator iter = report.groups(); iter.hasNext(); ) {
	Group g = (Group)iter.next();
	buildSectionsInList(g.headers());
    }

    buildSectionsInList(report.details());

    // Add footers for each group, where the groups are in reverse order.
    for (Iterator iter = report.groupsReversed(); iter.hasNext(); ) {
	Group g = (Group)iter.next();
	buildSectionsInList(g.footers());
    }

    buildSectionsInList(report.footers());
    buildSectionsInList(report.pageFooters());

    renameSectionWidgets();	// Muck with section names (e.g., add " (a)")

    // Put sectionContainer within scroller and add scroller to content
    // pane.
    int width = SectionWidget.LHS_WIDTH
	+ (int)report.getPaperFormat().getWidth();
    scroller = new JScrollPane(sectionContainer);
    scroller.setPreferredSize(new Dimension(width + 4, 400));
    getContentPane().add(scroller, BorderLayout.CENTER);
}

/**
 * Builds the section widgets for the sections in a collection and adds
 * them to the section container. We name the widgets with a separate
 * call to <code>renameSectionWidgets</code>.
 * <p>
 * We also start observing each section. Though our caller is called
 * more than once, it is OK to call addObserver() on an Observable
 * multiple times.
 *
 * @param iter an iterator over a collection of sections
 * @see #renameSectionWidgets
 */
protected void buildSectionsInList(Iterator iter) {
    while (iter.hasNext()) {
	Section sect = (Section)iter.next();
	sect.addObserver(this);	// Start observing changes

	SectionWidget sw = new SectionWidget(this, sect, "", "");
	sectionWidgets.add(sw);

	// Add to palette layer (one above default layer) because we want
	// to move field widgets below section widgets temporarily at the
	// end of a drag (see putDown()).
	sectionContainer.add(sw, JLayeredPane.PALETTE_LAYER);
    }
}

/**
 * Recalculates section names for all sections in the report. Calls
 * <code>renameSectionWidgetsIn</code> for each group of sections.
 *
 * @see #renameSectionWidgetsIn
 */
protected void renameSectionWidgets() {
    renameSectionWidgetsIn(report.headers(), I18N.get("Report.report_header"),
			   null);
    renameSectionWidgetsIn(report.pageHeaders(),
			   I18N.get("Report.page_header"), null);
    int i = 1;
    for (Iterator iter = report.groups(); iter.hasNext(); ++i) {
	Group g = (Group)iter.next();
	renameSectionWidgetsIn(g.headers(), I18N.get("DesignWin.group") + " #"
			       + i + ' ' + I18N.get("DesignWin.header"), g);
    }
    renameSectionWidgetsIn(report.details(), I18N.get("Report.detail"), null);
    i = 1;
    for (Iterator iter = report.groups(); iter.hasNext(); ++i) {
	Group g = (Group)iter.next();
	renameSectionWidgetsIn(g.footers(), I18N.get("DesignWin.group") + " #"
			       + i + ' ' + I18N.get("DesignWin.footer"), g);
    }
    renameSectionWidgetsIn(report.footers(), I18N.get("Report.report_footer"),
			   null);
    renameSectionWidgetsIn(report.pageFooters(),
			   I18N.get("Report.page_footer"), null);
}

/**
 * Recalculates section names for a collection of sections. Called
 * from <code>renameSectionWidgets</code>. Also sets the section's popup
 * menu's first item text.
 *
 * @param iter an iterator over a collection of sections
 * @param prefix prepended to all section names in the collection
 * @param group the group containing this section list; may be
 * <code>null</code>
 * @see #renameSectionWidgets
 */
protected void renameSectionWidgetsIn(Iterator iter, String prefix,
				      Group group)
{
    SectionWidget firstSectionWidget = null;
    boolean firstSectionFixed = false;
    for (int i = 0; iter.hasNext(); ++i) {
	Section s = (Section)iter.next();
	SectionWidget sw = findSectionWidgetFor(s);
	if (firstSectionWidget == null) {
	    sw.setDisplayName(prefix);
	    firstSectionWidget = sw;
	}
	else {
	    if (!firstSectionFixed) {
		firstSectionWidget.setDisplayName(prefix + " (a)");
		firstSectionFixed = true;
	    }
	    sw.setDisplayName(prefix + " (" + (char)('a' + i) + ")");
	}
	sw.setPopupName(group == null ? prefix : group.getColumn().fullName());
    }
}

/**
 * Enables or disables menu items based on field and window state.
 */
public void enableMenuItems() {
    int numSelected = countSelectedFields();
    boolean someFieldSelected = numSelected > 0;
    boolean multipleFieldsSelected = numSelected > 1;
    FieldWidget first = someFieldSelected ? (FieldWidget)selectedFields.get(0)
	: null;

    // Edit menu
    Section s = (first == null) ? null : first.getSectionWidget().getSection();

    cutItem.setEnabled(someFieldSelected);
    copyItem.setEnabled(someFieldSelected);
    pasteItem.setEnabled(!Clipboard.instance().isEmpty());
    delSelectionItem.setEnabled(someFieldSelected);
    delGroupItem.setEnabled(someFieldSelected && report.isInsideGroup(s));
    delSectionItem.setEnabled(someFieldSelected && !report.isOneOfAKind(s));

    // Insert menu
    if (numSelected == 1) {	// One field is selected
	// Only enable running subtotals if the selected field is a field
	// for which subtotals make sense.
	subtotItem.setEnabled(SubtotalField.canSubtotal(first.getField()));
    }
    else
	subtotItem.setEnabled(false);
    groupItem.setEnabled(report.hasFields());
    sectItem.setEnabled(someFieldSelected);

    // Format menu
    if (someFieldSelected) {
	// Only enable if some field is formattable
	boolean enable = someSelectedFieldUsesFormat();
	formatItem.setEnabled(enable);
    }
    else
	formatItem.setEnabled(false);
    borderItem.setEnabled(someFieldSelected);
    boundsItem.setEnabled(someFieldSelected);
    alignSubmenu.setEnabled(multipleFieldsSelected);
    sizeSubmenu.setEnabled(multipleFieldsSelected);
}

/**
 * Returns <code>true</code> if there is some selected field that can
 * be formatted.
 *
 * @return <code>true</code> if there is some selected field that can
 * be formatted
 */
public boolean someSelectedFieldUsesFormat() {
    for (Iterator iter = selectedFields.iterator(); iter.hasNext(); )
	if (((FieldWidget)iter.next()).usesFormat())
	    return true;
    return false;
}


/**
 * Handles user actions.
 */
public void actionPerformed(ActionEvent e) {
    String cmd = e.getActionCommand();

    // File menu
    if ("new".equals(cmd)) newReport();
    else if ("open".equals(cmd)) openReport();
    else if ("save".equals(cmd)) saveReport();
    else if ("save_as".equals(cmd)) saveReportAs();
    else if ("export".equals(cmd)) exportReport();
    else if ("close".equals(cmd)) maybeClose();
    else if ("quit".equals(cmd)) maybeQuit();

    // Edit menu
    else if ("undo".equals(cmd)) commandHistory.undo();
    else if ("redo".equals(cmd)) commandHistory.redo();
    else if ("cut".equals(cmd))
	commandHistory.perform(new CutCommand(this, selectedFields));
    else if ("copy".equals(cmd)) copySelectedFields();
    else if ("paste".equals(cmd)) paste();
    else if ("del_selection".equals(cmd)) deleteSelectedFields();
    else if ("del_group".equals(cmd)) {
	FieldWidget fw = (FieldWidget)selectedFields.get(0);
	deleteGroupContaining(fw.getSectionWidget().getSection());
    }
    else if ("del_section".equals(cmd)) {
	FieldWidget fw = (FieldWidget)selectedFields.get(0);
	deleteSection(fw.getSectionWidget().getSection());
    }

    // Insert menu
    else if ("database_field".equals(cmd))
	openFieldPickerWin(FieldPickerWin.REPORT_DATABASE_FIELDS);
    else if ("text".equals(cmd)) placeNewTextField();
    else if ("formula_field".equals(cmd))
	openFieldPickerWin(FieldPickerWin.FORMULAS);
    else if ("usercol_field".equals(cmd))
	openFieldPickerWin(FieldPickerWin.USERCOLS);
    else if ("parameter_field".equals(cmd))
	openFieldPickerWin(FieldPickerWin.PARAMETERS);
    else if ("running_total_field".equals(cmd)) openSubtotalWin();
    else if ("special_field".equals(cmd))
	openFieldPickerWin(FieldPickerWin.SPECIAL_FIELDS);
    else if ("image".equals(cmd)) createImageField();
//      else if ("line".equals(cmd)) placeNewLine();
    else if ("add_group".equals(cmd)) openNewGroupWin();
    else if ("ins_section".equals(cmd)) insertSection();

    // Format menu
    else if ("format_field".equals(cmd)) openFormatWin(0);
    else if ("border".equals(cmd)) openFormatWin(1);
    else if ("position".equals(cmd)) openBoundsWin();

    // Format/Align submenu
    else if ("align_top".equals(cmd)) align(ALIGN_TOP);
    else if ("align_middle".equals(cmd)) align(ALIGN_MIDDLE);
    else if ("align_bottom".equals(cmd)) align(ALIGN_BOTTOM);
    else if ("align_left".equals(cmd)) align(ALIGN_LEFT);
    else if ("align_center".equals(cmd)) align(ALIGN_CENTER);
    else if ("align_right".equals(cmd)) align(ALIGN_RIGHT);
    else if ("snap_to_grid".equals(cmd)) align(ALIGN_SNAP_TO_GRID);

    // Format/Size submenu
    else if ("size_width".equals(cmd)) size(SIZE_SAME_WIDTH);
    else if ("size_height".equals(cmd)) size(SIZE_SAME_HEIGHT);
    else if ("size_size".equals(cmd)) size(SIZE_SAME_SIZE);

    // Database menu
    else if ("visual_table_linker".equals(cmd)) openVisTableWin();
    else if ("sql_query_text".equals(cmd)) showSQL();
    else if ("db_connection".equals(cmd)) openDbConnWin(false);

    // Report menu
    else if ("run".equals(cmd)) runReport();
    else if ("select_records".equals(cmd)) openWhereClauseEditor();
    else if ("sort_by".equals(cmd)) openSortWin();
    else if ("group_by".equals(cmd)) openGroupWin();
    else if ("summary".equals(cmd)) openDescripWin();

    // Help menu
    else if ("help".equals(cmd)) help();
    else if ("about".equals(cmd)) about();

    // Paper size menu
    else if ("Portrait".equals(cmd))
	changePaperOrientation(PaperFormat.PORTRAIT);
    else if ("Landscape".equals(cmd))
	changePaperOrientation(PaperFormat.LANDSCAPE);
    else {
	PaperFormat p =
	    PaperFormat.get(report.getPaperFormat().getOrientation(), cmd);
	if (p != null)
	    changePaperSize(p);
    }
}

protected void changePaperSize(PaperFormat p) {
    report.setPaperFormat(p);

    // Select menu items (both orientation and name) based on paper format.
    // We select both because this method is also called by
    // changePaperOrientation().
    paperSizeSubmenu.getItem(p.getOrientation() == PaperFormat.PORTRAIT
			     ? 0 : 1)
	.setSelected(true);
    int i = 3;
    for (Iterator iter = PaperFormat.names(); iter.hasNext(); ++i) {
	String name = (String)iter.next();
	if (name.equals(p.getName()))
	    paperSizeSubmenu.getItem(i).setSelected(true);
    }

    for (Iterator iter = sectionWidgets.iterator(); iter.hasNext(); )
	((SectionWidget)iter.next()).paperSizeChanged();

    Dimension d = new Dimension(SectionWidget.LHS_WIDTH
				+ (int)report.getPaperFormat().getWidth(),
				sectionContainer.getHeight());
    sectionContainer.setPreferredSize(d);
    sectionContainer.invalidate();
    validate();
}

protected void changePaperOrientation(int orientation) {
    PaperFormat p = PaperFormat.get(orientation,
				    report.getPaperFormat().getName());
    if (p != null)
	changePaperSize(p);
}

JFileChooser getChooser() {
    if (chooser == null)
	chooser = new JFileChooser();
    return chooser;
}

public Report getReport() { return report; }

/**
 * Creates a new report in a new design window.
 */
protected void newReport() {
    new DesignWin(null);	// Open a new window
}

/**
 * Opens an existing report in a new design window.
 */
protected void openReport() {
    int returnVal = getChooser().showOpenDialog(this);
    if (returnVal == JFileChooser.APPROVE_OPTION)
	new DesignWin(getChooser().getSelectedFile().getPath());
}

/**
 * Saves the current report.
 */
protected void saveReport() {
    if (reportFileName == null)
	saveReportAs();
    else
	writeReportFile(reportFileName);
}

/**
 * Saves the current report in a different file.
 */
protected void saveReportAs() {
    int returnVal = getChooser().showSaveDialog(this);
    if (returnVal == JFileChooser.APPROVE_OPTION) {
	reportFileName = getChooser().getSelectedFile().getPath();
	writeReportFile(reportFileName);
    }
}

/**
 * Writes the current report to the specified file. Also tells the
 * command history the report has been saved so it knows how to report
 * if any changes have been made from this point on.
 *
 * @param fileName a file name
 */
protected void writeReportFile(String fileName) {
    report.writeFile(fileName);
    commandHistory.setBaseline();
}

/**
 * Exports the report output using one of the layout engines.
 */
protected void exportReport() {
    new ExportWin(this, report);
}

/**
 * Saves the report if it is changed (some command has been performed) and
 * closes the current design window. If there are no more open design
 * windows, exist the application.
 */
protected void maybeClose() {
    if (commandHistory.isChanged()) {
	String str = I18N.get("DesignWin.save_question");
	switch (JOptionPane.showConfirmDialog(this, str)) {
	case JOptionPane.YES_OPTION:
	    saveReport();
	case JOptionPane.NO_OPTION:
	    break;
	case JOptionPane.CANCEL_OPTION:
	    return;		// Don't close window
	}
    }
    dispose();
    DesignWin.deleteWindow(this);
}

/**
 * Runs and displays the report.
 */
protected void runReport() {
    report.setLayoutEngine(new SwingLE());
    report.run();
}

/**
 * Returns the number of selected fields.
 *
 * @return the number of selected fields
 */
int countSelectedFields() {
    return selectedFields.size();
}

/**
 * Hands each field widget to the specified <code>FieldWidgetWalker</code>.
 *
 * @param perambulator a field walker
 */
void withWidgetsDo(FieldWidgetWalker perambulator) {
    for (Iterator iter = sectionWidgets.iterator(); iter.hasNext(); ) {
	SectionWidget sw = (SectionWidget)iter.next();
	Object[] kids = sw.fieldPanel.getComponents();
	for (int i = 0; i < kids.length; ++i) {
	    FieldWidget fw = FieldWidget.findFieldWidgetOwning(kids[i]);
	    if (fw != null)
		perambulator.step(fw);
	}
    }
}

/**
 * Hands each selected field widget to the specified
 * <code>FieldWidgetWalker</code>.
 *
 * @param perambulator a field walker
 */
void withSelectedWidgetsDo(FieldWidgetWalker perambulator) {
    for (Iterator iter = selectedFields.iterator(); iter.hasNext(); )
	perambulator.step((FieldWidget)iter.next());
}

/**
 * Deletes the group that contains the specified section.
 *
 * @param section a report section
 */
void deleteGroupContaining(Section section) {
    Group group = report.findGroup(section);
    if (group != null)
	commandHistory.perform(new DeleteGroupCommand(this, report, group));
}

/**
 * Inserts a new section below the section containing the first selected
 * field widget. Only called when at least one widget is selected.
 *
 * @param section a report section
 */
void insertSection() {
    Field f = firstSelectedField().getField();
    if (f != null)
	insertSectionBelow(f.getSection());
}

/**
 * Inserts a new section below the specified section.
 *
 * @param s a report section
 */
void insertSectionBelow(Section s) {
    commandHistory.perform(new NewSectionCommand(this, report, s));
}

/**
 * Inserts a section widget at the specified position in the list. Also
 * starts observing the section and renames all section widgets (adding "(a",
 * for example. Called from <code>NewSectionCommand.perform</code>.
 *
 * @param sw the section widget to insert
 * @param putAfter <var>sw</var> goes after this widget; may be
 * <code>null</code>
 * @see NewSectionCommand#perform
 */
public void insertSectionWidgetAfter(SectionWidget sw, SectionWidget putAfter)
{
    int insertIndex = sectionWidgets.indexOf(putAfter);
    sectionWidgets.add(insertIndex + 1, sw);

    // Add to palette layer (one above default layer) because
    // we want to move field widgets below section widgets
    // temporarily at the end of a drag (see putDown()).
    sectionContainer.add(sw, JLayeredPane.PALETTE_LAYER, insertIndex + 1);

    sw.getSection().addObserver(this);
    renameSectionWidgets(); // Muck with names (e.g., add " (a)")
}

/**
 * Deletes the specified section. Does nothing if this section is the only
 * one of its kind. (This method should only be called if the section is
 * not one-of-a-kind. No harm done if you do, though.)
 *
 * @param s a report section
 */
public void deleteSection(Section s) {
    if (s != null && !report.isOneOfAKind(s))
	commandHistory.perform(new DeleteSectionCommand(this, report, s));
}

/**
 * Deletes a section from the report and the design window and returns
 * the section widget <em>above</em> the section's. If <var>s</var>
 * is the first section in the report, returns <code>null</code>.
 * Called from commands; don't call this yourself. Insetad call
 * <code>deleteSection</code>.
 *
 * @param s the section to delete
 * @return the section widget above the section's, or <code>null</code>
 * if there is none
 * @see #deleteSection
 * @see DeleteSectionCommand
 * @see NewSectionCommand
 */
public SectionWidget doDeleteSection(Section s) {
    SectionWidget sw = findSectionWidgetFor(s);
    int index = sectionWidgets.indexOf(sw);
    SectionWidget widgetBefore = (index == 0) ? null
	: (SectionWidget)sectionWidgets.get(index - 1);

    s.deleteObserver(this);
    report.removeSection(s);
    sectionContainer.remove(sw);
    sectionWidgets.remove(sw);
    renameSectionWidgets();

    return widgetBefore;
}

/**
 * Rebuilds the group sections and redisplays the report.
 */
public void rebuildGroups() {
    deselectAll();
    remove(scroller);
    buildSections();
    pack();
}


/**
 * Creates and adds a new text field to the first section of the page header.
 *
 * @param x where to place the title
 * @param width how wide it should be
 * @param title the string to display
 * @return the newly-created widget
 */
public FieldWidget addTitleField(int x, int width, String title) {
    Section s = report.getFirstSectionByArea(Report.AREA_PAGE_HEADER);

    // Create the field.
    Field f = Field.create(null, report, s, "text", title, true);
    jimm.datavision.Rectangle b = f.getBounds();
    b.setBounds(x, 0, width, Field.DEFAULT_HEIGHT);

    // Make the field underlined and bold.
    Format fmt = new Format();
    fmt.setBold(true);
    fmt.setUnderline(true);
    f.setFormat(fmt);

    // Add the field to the report section.
    s.addField(f);

    // Create widget and add to section widget.
    TextFieldWidget tfw = new TextFieldWidget(null, f);
    tfw.moveToSection(findSectionWidgetFor(s));
    return tfw;
}

/**
 * Aligns the selected fields to the first selected field (chronologically
 * speaking).
 *
 * @param which alignment constant.
 */
protected void align(int which) {
    if (selectedFields.isEmpty())
	return;

    // Align the widgets based on the position of the first selected field.
    CompoundCommand cmd =
	new CompoundCommand(I18N.get("FieldAlignCommand.name"));
    FieldWidget first = firstSelectedField();
    for (Iterator iter = selectedFields.iterator(); iter.hasNext(); )
	cmd.add(new FieldAlignCommand((FieldWidget)iter.next(), which, first));

    commandHistory.perform(cmd);
}

/**
 * Resize the selected fields based on the first selected field
 * (chronologically speaking) and the specified resize (width, height, both).
 *
 * @param which size constant
 */
protected void size(int which) {
    if (selectedFields.isEmpty())
	return;

    // Resize the widgets based on the position of the first selected field.
    CompoundCommand cmd =
	new CompoundCommand(I18N.get("FieldResizeCommand.name"));
    FieldWidget first = firstSelectedField();
    for (Iterator iter = selectedFields.iterator(); iter.hasNext(); )
	cmd.add(new FieldResizeCommand((FieldWidget)iter.next(), which,
				       first));

    commandHistory.perform(cmd);
}

/**
 * Returns first selected field widget (chronologically speaking). May
 * return <code>null</code> if no fields are selected
 *
 * @return field widget selected the earliest, or <code>null</code> if
 * no fields are selected
 */
FieldWidget firstSelectedField() {
    return selectedFields.isEmpty()
	? null : (FieldWidget)selectedFields.get(0);
}

/**
 * Selects a field widget, possibly deselecting all others everywhere.
 * Called from section widget.
 *
 * @param FieldWidget a field widget
 * @param makeSelected new selection state
 * @param deselectAll if <code>true</code>, all other fields in all sections
 * are deselected first
 */
public void select(FieldWidget fieldWidget, boolean makeSelected,
	    boolean deselectOthers)
{
    if (deselectOthers)
	deselectAll();

    fieldWidget.doSelect(makeSelected);
    if (makeSelected) {
	if (!selectedFields.contains(fieldWidget)) // Don't add it twice
	    selectedFields.add(fieldWidget);
    }
    else {
	selectedFields.remove(fieldWidget);
    }

    enableMenuItems();
}

/**
 * Deselect all fields. Called from SectionWidget#deselectAll.
 *
 * @see SectionWidget#deselectAll
 */
public void deselectAll() {
    for (Iterator iter = selectedFields.iterator(); iter.hasNext(); )
	((FieldWidget)iter.next()).doSelect(false);
    selectedFields.clear();

    enableMenuItems();
}

/**
 * Copies the selected fields to the clipboard. We need to create
 * pasteable objects.
 *
 * @see jimm.datavision.gui.cmd.Pasteable
 */
protected void copySelectedFields() {
    ArrayList pasteables = new ArrayList();
    for (Iterator iter = selectedFields.iterator(); iter.hasNext(); )
	pasteables.add(new FieldClipping((FieldWidget)iter.next()));
    Clipboard.instance().setContents(pasteables);

    pasteItem.setEnabled(true);
}

protected void paste() {
    CompoundCommand cmd = new CompoundCommand(I18N.get("PasteCommand.name"));
    if (selectedFields.size() > 0)
	cmd.add(new DeleteCommand(this, selectedFields));
    cmd.add(new PasteCommand(this));
    commandHistory.perform(cmd);
}

/**
 * Delete selected fields.
 */
protected void deleteSelectedFields() {
    deleteSelectedFieldsAnd(null);
}

/**
 * Delete specified field and all selected fields.
 *
 * @param oneMore an additional field to delete; may be <code>null</code>
 */
protected void deleteSelectedFieldsAnd(FieldWidget oneMore) {
    ArrayList fields = new ArrayList(selectedFields);
    if (oneMore != null && !fields.contains(oneMore))
	fields.add(oneMore);
    commandHistory.perform(new DeleteCommand(this, fields));
}

/**
 * Toggles the visibility of all selected fields plus the one passed in.
 * Called by a section widget.
 *
 * @see FieldWidget#doSetVisibility
 */
void setFieldVisibility(boolean newVisiblity, FieldWidget fw) {
    if (selectedFields.isEmpty() && fw == null)
	return;

    String nameKey = newVisiblity ? "FieldShowCommand.name"
	: "FieldHideCommand.name";

    CompoundCommand cmd = new CompoundCommand(I18N.get(nameKey));
    for (Iterator iter = selectedFields.iterator(); iter.hasNext(); )
	cmd.add(new FieldShowHideCommand((FieldWidget)iter.next(), nameKey,
					 newVisiblity));
    if (fw != null && !selectedFields.contains(fw))
	cmd.add(new FieldShowHideCommand(fw, nameKey, newVisiblity));

    commandHistory.perform(cmd);
}

/**
 * Asks design window to create and accepts a new text field. Called from
 * <code>SectionWidget.createNewTextField</code>.
 *
 * @see DesignWin#createNewTextField
 */
void createNewTextField(SectionWidget sw, MouseEvent e) {
    commandHistory.perform(new NewTextFieldCommand(sw, e));

    // Accept the drop
    acceptNewTextField();
}

/**
 * Picks up the field widget because field dragging is starting.
 * Called from SectionWidget#pickUp.
 *
 * @param f field widget
 * @see SectionWidget#pickUp
 */
void pickUp(FieldWidget f) {
    Dimension size = sectionContainer.getBounds().getSize();
    sectionContainer.setPreferredSize(size);
    sectionContainer.setMinimumSize(size);

    for (Iterator iter = selectedFields.iterator(); iter.hasNext(); ) {
	FieldWidget fw = (FieldWidget)iter.next();
	fw.pickUp();

	// Add to the drag layer of our section container.
	sectionContainer.add(fw.getComponent(), JLayeredPane.DRAG_LAYER);

	// Recalculate bounds relative to the section container.
	jimm.datavision.Rectangle b = fw.getField().getBounds();
	b.setBounds(b.x + SectionWidget.LHS_WIDTH,
		    b.y + fw.getSectionWidget().getBounds().y,
		    b.width, b.height);
    }
}

/**
 * Puts the dragged field widgets down inside the sections they are
 * floating above. Called from SectionWidget#putDown.
 *
 * @param f the field widget being dragged; all other selected fields
 * have been dragged along with it
 * @param origScreenPos the original location of the field in screen
 * coordinates
 * @param mouseScreenPos the current mouse position in screen coordinates;
 * note 
 * @see SectionWidget#putDown
 */
void putDown(FieldWidget f, java.awt.Point origScreenPos,
	     java.awt.Point mouseScreenPos)
{
    // Move all dragged fields under everything else so getComponentAt()
    // will not return this field.
    for (Iterator iter = selectedFields.iterator(); iter.hasNext(); )
	sectionContainer.setLayer(((FieldWidget)iter.next()).getComponent(),
				  JLayeredPane.DEFAULT_LAYER.intValue());

    CompoundCommand cmd =
	new CompoundCommand(I18N.get("FieldMoveCommand.name"));

    // Move to new section. Each field may be dropped into a different
    // section.
    for (Iterator iter = selectedFields.iterator(); iter.hasNext(); ) {
	FieldWidget fw = (FieldWidget)iter.next();
	SectionWidget sw =
	    getSectionWidgetUnder(fw.getComponent().getLocationOnScreen());
	if (sw == null)		// Field snaps back to orig pos if sw is null
	    fw.snapBack();
	else
	    cmd.add(new FieldMoveCommand(fw, sw));
    }

    if (cmd.numCommands() > 0)
	commandHistory.perform(cmd);
}

/**
 * Returns the section widget under the mouse. Returns <code>null</code>
 * unless the mouse is over a section field panel (the white area on
 * which fields belong).
 *
 * @param screenPos a position in screen coordinates
 * @return the SectionWidget under the mouse
 */
protected SectionWidget getSectionWidgetUnder(java.awt.Point screenPos) {
    // Translate screenPos to sectionContainer coords
    java.awt.Point scScreenPos = sectionContainer.getLocationOnScreen();
    java.awt.Point scPos = new java.awt.Point(screenPos.x - scScreenPos.x,
					      screenPos.y - scScreenPos.y);

    if (scPos.x < SectionWidget.LHS_WIDTH) // Reject if in LHS name
	return null;

    // We've landed on a section field panel, a field in the section, or a
    // section name label. Crawl up the parent hierarchy until we find
    // first the section field panel (insuring we are in the right place)
    // and then the section.
    Component c = sectionContainer.getComponentAt(scPos);
    while (c != null && !(c instanceof SectionWidget))
	c = c.getParent();
    return (SectionWidget)c;
}

/**
 * Drags the selected field widgets a few pixels. Called from section that
 * contains field being dragged.
 *
 * @param action a <code>FieldWidget.ACTION_*</code> constant
 * @param dx delta x
 * @param dy delta y
*/
protected void dragSelectedWidgets(int action, int dx, int dy) {
    for (Iterator iter = selectedFields.iterator(); iter.hasNext(); )
	((FieldWidget)iter.next()).doDrag(action, dx, dy);
}

/**
 * Opens a new or existing field picker window.
 *
 * @param startingType the index of the starting type to display
 */
protected void openFieldPickerWin(int startingType) {
    new FieldPickerWin(this, report, startingType);
}

/**
 * Opens a new or existing field running subtotal window. We should
 * only get here if there is exactly one selected field.
 */
protected void openSubtotalWin() {
    new SubtotalsWin(this, (FieldWidget)selectedFields.get(0));
}

/**
 * Opens a dialog that asks the user to select an image file. Creates an
 * image in the report header.
 */
protected void createImageField() {
    int returnVal = getChooser().showOpenDialog(this);
    if (returnVal == JFileChooser.APPROVE_OPTION) {
	Section s = report.getFirstSectionByArea(Report.AREA_REPORT_HEADER);
	NewImageFieldCommand cmd =
	    new NewImageFieldCommand(findSectionWidgetFor(s),
				     getChooser().getSelectedFile());
	commandHistory.perform(cmd);
    }
}

/**
 * Opens a new or existing new group window.
 */
protected void openNewGroupWin() {
    new NewGroupWin(this, report);
}

/**
 * Sets the flag that tells everyone else that the user wants to place
 * a new text field.
 */
protected void placeNewTextField() {
    placingNewTextField = true;
    setCursor(Cursor.getPredefinedCursor(Cursor.TEXT_CURSOR));
}

/**
 * Returns <code>true</code> if the user is trying to place a new text field.
 *
 * @return <code>true</code> if the user is trying to place a new text field
 */
boolean isPlacingNewTextField() {
    return placingNewTextField;
}

/**
 * The caller has accepted the new text field.
 */
void acceptNewTextField() {
    placingNewTextField = false;
    setCursor(null);
}

/**
 * The caller has rejected the new text field.
 */
void rejectNewTextField() {
    acceptNewTextField();
}

/**
 * Opens a new or existing field format window.
 *
 * @param whichTab the index of the tab to display when opened
 */
protected void openFormatWin(int whichTab) {
    FieldWidget fw = firstSelectedField();
    if (fw != null)
	new FormatWin(this, whichTab);
}

/**
 * Opens a new or existing bounds editor window.
 */
protected void openBoundsWin() {
    FieldWidget fw = firstSelectedField();
    if (fw != null)
	new BoundsWin(this);
}

/**
 * Opens a new or existing visible table joiner window.
 */
protected void openVisTableWin() {
    new VisTableWin(this, report);
}

/**
 * Opens a new or existing where clause editor.
 */
protected void openWhereClauseEditor() {
    new WhereClauseWin(this, report);
}

/**
 * Opens a new or existing sort order window.
 */
protected void openSortWin() {
    new SortWin(this, report);
}

/**
 * Opens a new or existing group order window.
 */
protected void openGroupWin() {
    new GroupWin(this, report);
}

/**
 * Opens a new or existing sort order window.
 */
protected void openDescripWin() {
    new DescripWin(this, report);
}

/**
 * Opens the help window.
 */
protected void help() {
    HelpWin helpWin = HelpWin.instance();
    helpWin.setState(Frame.NORMAL); // De-iconify. Why is this necessary?
    helpWin.setVisible(true);
    helpWin.toFront();
}

/**
 * Opens the about box.
 */
protected void about() {
    String msg = I18N.get("DesignWin.about_1") + info.Version + "\n"
	+ I18N.get("DesignWin.about_2") + "\n"
	+ info.URL + "\n\n"
	+ info.Copyright + ".\n\n"
	+ I18N.get("DesignWin.about_3");
    JOptionPane.showMessageDialog(null, msg, I18N.get("DesignWin.about_title"),
				  JOptionPane.PLAIN_MESSAGE);
}

/**
 * Opens a new or existing database connection info window.
 *
 * @param modal passed on to dialog constructor
 */
protected void openDbConnWin(boolean modal) {
    new DbConnWin(this, report, modal);
}

/**
 * Opens a new window containing the SQL query text.
 */
protected void showSQL() {
    new SQLQueryWin(this, report.getQuery().toString());
}

/**
 * Returns the section widget containing the specified section.
 *
 * @param s section
 * @return the section widget containing the section
 */
public SectionWidget findSectionWidgetFor(Section s) {
    for (Iterator iter = sectionWidgets.iterator(); iter.hasNext(); ) {
	SectionWidget sw = (SectionWidget)iter.next();
	if (sw.section == s)
	    return sw;
    }
    return null;
}


/**
 * Snaps the rectangle to the grid.
 *
 * @param r a rectangle
 */
public void snapToGrid(jimm.datavision.Rectangle r) {
    int coord = (int)r.x;
    int mod = coord % GRID_SIZE;
    if (mod != 0) {
	if (mod <= GRID_SIZE / 2)
	    r.setX(coord - mod);
	else
	    r.setX(coord + GRID_SIZE - mod);
    }

    coord = (int)r.y;
    mod = coord % GRID_SIZE;
    if (mod != 0) {
	if (mod < GRID_SIZE / 2)
	    r.setY(coord - mod);
	else
	    r.setY(coord + GRID_SIZE - mod);
    }
}

// Might be using this code for dragging to select multiple fields. It
// should eventually live in a separate file.

//  class SelectionHandler extends MouseInputAdapter {

//  protected java.awt.Rectangle rect;
//  protected HashSet fields;
//  protected JPanel selectionOutline;

//  public void mousePressed(MouseEvent e) {
//      int x = e.getX() - SectionWidget.LHS_WIDTH;
//      if (x < 0)
//  	return;

//      deselectAll();

//      int y = e.getY();
//      rect = new java.awt.Rectangle(x, y, 0, 0);
//      selectionOutline = new JPanel();
//      selectionOutline.setBounds(rect);
//      getContentPane().add(selectionOutline, 0); // Add to top of visual stack
//  }

//  public void mouseDragged(MouseEvent e) {
//      updateSize(e);
//  }

//  public void mouseReleased(MouseEvent e) {
//      updateSize(e);
//      remove(selectionOutline);
//  }

//  /* 
//   * Updates the size of the current rectangle, changes selection list, and
//   * calls <code>repaint</code>. Because rect always has the same origin,
//   * translate it if the width or height is negative.
//   * 
//   * For efficiency, specify the painting region using arguments to the
//   * <code>repaint</code> call.
//   */
//  void updateSize(MouseEvent e) {
//      if (selectionOutline == null)
//  	return;

//      int x = e.getX();
//      int y = e.getY();
//      rect.setSize(x - rect.x, y - rect.y);
//      selectionOutline.setSize(x - rect.x, y - rect.y);

//  //      selectFieldsWithin(rect);

//  //      java.awt.Rectangle totalRepaint = rectToDraw.union(previousRectDrawn);
//  //      repaint(totalRepaint.x, totalRepaint.y,
//  //  	    totalRepaint.width, totalRepaint.height);
//  }

//  }

}
