package jimm.datavision.gui;
import jimm.datavision.*;
import jimm.datavision.layout.*;
import java.io.PrintWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.awt.BorderLayout;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.util.ArrayList;
import javax.swing.*;

/**
 * This dialog lets the user export report results using one of the
 * layout engines.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class ExportWin extends JDialog implements ActionListener {


protected Report report;
protected DesignWin designWin;
protected JComboBox combo;

/**
 * Constructor.
 *
 * @param designWin the window to which this dialog belongs
 * @param initialText the initial text to edit
 * @param title the window title
 */
public ExportWin(DesignWin designWin, Report report) {
    super(designWin, I18N.get("ExportWin.title"));
    this.report = report;
    this.designWin = designWin;
    buildWindow();
    pack();
    setVisible(true);
}

/**
 * Builds the window contents.
 *
 * @param whichTab the index of the tab to display when opened
 */
protected void buildWindow() {
    // Panel containing list of layout engines
    combo = new JComboBox(layoutNames());
    JPanel comboPanel = new JPanel();
    comboPanel.setBorder(BorderFactory.createEmptyBorder(20, 20, 20, 20));
    comboPanel.add(combo);

    // OK and Cancel Buttons
    JPanel buttonPanel = buildButtonPanel();

    // Add list, cards, and buttons to window
    getContentPane().setLayout(new BorderLayout());
    getContentPane().add(comboPanel, BorderLayout.CENTER);
    getContentPane().add(buttonPanel, BorderLayout.SOUTH);
}

/**
 * Builds and returns a panel containing the OK and Cancel
 *
 * @return a panel
 */
protected JPanel buildButtonPanel() {
    JPanel buttonPanel = new JPanel();
    JButton button;

    buttonPanel.add(button = new JButton(I18N.get("GUI.ok")));
    button.addActionListener(this);
    button.setDefaultCapable(true);

    buttonPanel.add(button = new JButton(I18N.get("GUI.cancel")));
    button.addActionListener(this);

    return buttonPanel;
}

/**
 * Handles the OK and Cancel buttons.
 *
 * @param e action event
 */
public void actionPerformed(ActionEvent e) {
    String cmd = e.getActionCommand();
    if (I18N.get("GUI.ok").equals(cmd)) {
	String choice = (String)combo.getSelectedItem();
	dispose();
	runReport(choice);
    }
    else if (I18N.get("GUI.cancel").equals(cmd)) {
	dispose();
    }
}

protected String[] layoutNames() {
    String[] names = new String[6];
    int i = 0;
    names[i++] = I18N.get("ExportWin.le_comma");
    names[i++] = I18N.get("ExportWin.le_tab");
    names[i++] = I18N.get("ExportWin.le_docbook");
    names[i++] = I18N.get("ExportWin.le_html");
    names[i++] = I18N.get("ExportWin.le_latex");
    names[i++] = I18N.get("ExportWin.le_xml");
    return names;
}

/**
 * Given the user's choice of layout engine, ask user for output file
 * and run the report.
 *
 * @param choice the combo box string that the user selected
 */
protected void runReport(String choice) {
    String extension = null;
    if (I18N.get("ExportWin.le_comma").equals(choice))
	extension = ".csv";
    else if (I18N.get("ExportWin.le_tab").equals(choice))
	extension = ".tab";
    else if (I18N.get("ExportWin.le_docbook").equals(choice))
	extension = ".sgml";
    else if (I18N.get("ExportWin.le_html").equals(choice))
	extension = ".html";
    else if (I18N.get("ExportWin.le_latex").equals(choice))
	extension = ".tex";
    else if (I18N.get("ExportWin.le_xml").equals(choice))
	extension = ".xml";

    PrintWriter out = openOutput(extension);
    if (out == null)
	return;

    LayoutEngine le = null;
    if (I18N.get("ExportWin.le_comma").equals(choice))
	le = new CharSepLE(out, ',');
    else if (I18N.get("ExportWin.le_tab").equals(choice))
	le = new CharSepLE(out, '\t');
    else if (I18N.get("ExportWin.le_docbook").equals(choice))
	le = new DocBookLE(out);
    else if (I18N.get("ExportWin.le_html").equals(choice))
	le = new HTMLLE(out);
    else if (I18N.get("ExportWin.le_latex").equals(choice))
	le = new LaTeXLE(out);
    else if (I18N.get("ExportWin.le_xml").equals(choice))
	le = new XMLLE(new IndentWriter(out));

    if (le != null) {
	report.setLayoutEngine(le);
	report.run();
    }
}

/**
 * Lets user select output file. I'd like to be able to give a default
 * file extension (thus the argument), but that will have to come later.
 *
 * @param extension default file name extension (unused)
 */
protected PrintWriter openOutput(String extension) {
    JFileChooser chooser =
	designWin == null ? new JFileChooser() : designWin.getChooser();
    int returnVal = chooser.showSaveDialog(this.getOwner());
    if (returnVal == JFileChooser.APPROVE_OPTION) {
	String path = chooser.getSelectedFile().getPath();
	PrintWriter out = null;
	try {
	    out = new PrintWriter(new FileWriter(path));
	}
	catch (IOException e) {
	    out = null;
	    ErrorHandler.error(I18N.get("ExportWin.err_msg"),
			       e, I18N.get("ExportWin.err_title"));
	}
	finally {
	    return out;
	}
    }
    else
	return null;
}

}
