package jimm.datavision.gui;
import jimm.datavision.*;
import jimm.datavision.sql.*;
import jimm.datavision.gui.DesignWin;
import jimm.datavision.gui.parameter.ParamEditWin;
import jimm.datavision.gui.cmd.NameableRenameCommand;
import jimm.datavision.gui.cmd.FPCutCommand;
import java.awt.Dimension;
import java.awt.BorderLayout;
import java.awt.event.*;
import java.awt.datatransfer.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.tree.*;

/**
 * A window that lets the user drag any field available onto the report
 * and create, edit, delete, and rename formulas, parameters, and user
 * columns.
 * <p>
 * Uses a field picker tree. The classes used to store information about
 * leaf nodes are subclasses of <code>FPLeafInfo</code> and are found
 * in FPLeafInfo.java.
 *
 * @see FieldPickerTree
 * @see FPLeafInfo
 * @see ColumnInfo
 * @see FormulaInfo
 * @see ParameterInfo
 * @see UserColumnInfo
 * @see SpecialInfo
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
class FieldPickerWin
    extends JDialog
    implements ActionListener, TreeSelectionListener, Observer
{

public static final int REPORT_DATABASE_FIELDS = 0;
public static final int FORMULAS = 1;
public static final int PARAMETERS = 2;
public static final int USERCOLS = 3;
public static final int SPECIAL_FIELDS = 4;
public static final int ALL_DATABASE_FIELDS = 5;
//  public static final int GROUP_NAME_FIELDS = XXXX;

protected Report report;
protected DesignWin designWin;
protected JMenuItem cutItem, editFormulaItem, renameFormulaItem,
    editParameterItem, renameParameterItem,
    editUserColumnItem, renameUserColumnItem;
protected FieldPickerTree tree;
protected DefaultTreeModel treeModel;
protected DefaultMutableTreeNode formulaCategoryNode;
protected DefaultMutableTreeNode parameterCategoryNode;
protected DefaultMutableTreeNode userColumnCategoryNode;
protected DefaultMutableTreeNode selectedNode;
protected FPLeafInfo selectedInfo;
protected Comparator nameComparator;

/**
 * Constructor.
 *
 * @param designWin the design window to which this dialog belongs
 * @param report the report
 * @param startingType the index of the starting type to display
 */
FieldPickerWin(DesignWin designWin, Report report, int startingType) {
    super(designWin, I18N.get("FieldPickerWin.title"));
    this.report = report;
    this.designWin = designWin;

    // Note: this comparator imposes orderings that are inconsistent with
    // equals.
    nameComparator = new Comparator() {
	public int compare(Object o1, Object o2) {
	    String name1 = ((Nameable)o1).getName();
	    String name2 = ((Nameable)o2).getName();
	    return name1.compareTo(name2);
	}};

    buildWindow(startingType);

    pack();
    show();
}

public void update(Observable o, Object arg) {
    // Primitive and overkill, but it works.
    tree.repaint();
}

/**
 * Builds the contents of the window.
 *
 * @param startingType the index of the starting type to display
 */
protected void buildWindow(int startingType) {
    buildMenuBar();

    DefaultMutableTreeNode top = new DefaultMutableTreeNode();
    createNodes(top);
    treeModel = new DefaultTreeModel(top);

    tree = new FieldPickerTree(treeModel);
    tree.setRootVisible(false);
    tree.getSelectionModel().setSelectionMode
	(TreeSelectionModel.SINGLE_TREE_SELECTION);

    // Make specified starting type visible and expanded
    tree.expandRow(startingType);

    // Listen for selection changes
    tree.addTreeSelectionListener(this);

    // Create the scroll pane and add the tree to it. 
    JScrollPane treeView = new JScrollPane(tree);
    treeView.setMinimumSize(new Dimension(100, 100));
    treeView.setPreferredSize(new Dimension(200, 300));

    getContentPane().add(treeView, BorderLayout.CENTER);
}

/**
 * Builds the window menu bar.
 */
protected void buildMenuBar() {
    JMenuBar bar = new JMenuBar();
    bar.add(buildFileMenu());
    bar.add(buildEditMenu());
    bar.add(buildFieldMenu());
    setJMenuBar(bar);
}

/**
 * Builds and returns the "File" menu.
 *
 * @return a menu
 */
protected JMenu buildFileMenu() {
    JMenu menu = new JMenu(I18N.get("FieldPickerWin.menu_file"));
    menu.setMnemonic(KeyEvent.VK_F);

    MenuUtils.addToMenu(this, menu, I18N.get("FieldPickerWin.menu_file_close"),
			"close", KeyEvent.VK_C, KeyEvent.VK_W);

    return menu;
}

/**
 * Builds and returns the "Edit" menu.
 *
 * @return a menu
 */
protected JMenu buildEditMenu() {
    JMenu menu = new JMenu(I18N.get("FieldPickerWin.menu_edit"));
    menu.setMnemonic(KeyEvent.VK_E);

    MenuUtils.addToMenu(this, menu, I18N.get("FieldPickerWin.menu_edit_undo"),
			"undo", KeyEvent.VK_U, KeyEvent.VK_Z)
	.setEnabled(false);
    MenuUtils.addToMenu(this, menu, I18N.get("FieldPickerWin.menu_edit_redo"),
			"redo", KeyEvent.VK_R, 0)
	.setEnabled(false);
    menu.addSeparator();
    cutItem =
	MenuUtils.addToMenu(this, menu,
			    I18N.get("FieldPickerWin.menu_edit_cut"),
			    "cut", KeyEvent.VK_C, KeyEvent.VK_X);
    cutItem.setEnabled(false);
    MenuUtils.addToMenu(this, menu, I18N.get("FieldPickerWin.menu_edit_copy"),
			"copy", KeyEvent.VK_O, KeyEvent.VK_C)
	.setEnabled(false);
    MenuUtils.addToMenu(this, menu, I18N.get("FieldPickerWin.menu_edit_paste"),
			"paste", KeyEvent.VK_P, KeyEvent.VK_V)
	.setEnabled(false);

    return menu;
}

/**
 * Builds and returns the "Field" menu.
 *
 * @return a menu
 */
protected JMenu buildFieldMenu() {
    JMenu menu = new JMenu(I18N.get("FieldPickerWin.menu_field"));
    menu.setMnemonic(KeyEvent.VK_I);

    MenuUtils.addToMenu(this, menu,
			I18N.get("FieldPickerWin.menu_field_new_formula"),
			"new_formula", KeyEvent.VK_N, KeyEvent.VK_N);
    editFormulaItem =
	MenuUtils.addToMenu(this, menu,
			    I18N.get("FieldPickerWin.menu_field_edit_formula"),
			    "edit_formula", KeyEvent.VK_E, KeyEvent.VK_E);
    renameFormulaItem =
	MenuUtils.addToMenu(this, menu,
			    I18N.get("FieldPickerWin.menu_field_rename_formula"),
			    "rename_formula", KeyEvent.VK_R, KeyEvent.VK_R);
    editFormulaItem.setEnabled(false);
    renameFormulaItem.setEnabled(false);

    menu.addSeparator();

    MenuUtils.addToMenu(this, menu,
			I18N.get("FieldPickerWin.menu_field_new_param"),
			"new_parameter", KeyEvent.VK_P, KeyEvent.VK_P);
    editParameterItem =
	MenuUtils.addToMenu(this, menu,
			    I18N.get("FieldPickerWin.menu_field_edit_param"),
			    "edit_parameter", KeyEvent.VK_D, KeyEvent.VK_D);
    renameParameterItem =
	MenuUtils.addToMenu(this, menu,
			    I18N.get("FieldPickerWin.menu_field_rename_param"),
			    "rename_parameter", KeyEvent.VK_A, KeyEvent.VK_A);
    editParameterItem.setEnabled(false);
    renameParameterItem.setEnabled(false);

    menu.addSeparator();

    MenuUtils.addToMenu(this, menu,
			I18N.get("FieldPickerWin.menu_field_new_usercol"),
			"new_user_column", KeyEvent.VK_U, KeyEvent.VK_U);
    editUserColumnItem =
	MenuUtils.addToMenu(this, menu,
			    I18N.get("FieldPickerWin.menu_field_edit_usercol"),
			    "edit_user_column", KeyEvent.VK_I, KeyEvent.VK_I);
    renameUserColumnItem =
	MenuUtils.addToMenu(this, menu,
			    I18N.get("FieldPickerWin.menu_field_rename_usercol"),
			    "rename_user_column", KeyEvent.VK_M,
			    KeyEvent.VK_M);
    editUserColumnItem.setEnabled(false);
    renameUserColumnItem.setEnabled(false);

    return menu;
}

/**
 * Creates tree nodes.
 *
 * @param top top-level tree node
 */
protected void createNodes(DefaultMutableTreeNode top) {
    createUsedDatabaseTables(top);
    createFormulas(top);
    createParameters(top);
    createUserColumns(top);
    createSpecialFields(top);
    createAllDatabaseTables(top);
}

/**
 * Creates nodes representing tables and columns for columns used by report.
 *
 * @param top top-level tree node
 */
protected void createUsedDatabaseTables(DefaultMutableTreeNode top) {
    DefaultMutableTreeNode categoryNode =
	new DefaultMutableTreeNode(I18N.get("FieldPickerWin.db_fields"));
    top.add(categoryNode);

    // Store list of tables actually used by the report in a sorted set.
    final TreeSet tables = new TreeSet(nameComparator);

    // Walk the list of all columns used by the report, adding the table
    // to the sorted set of tables.
    report.withFieldsDo(new FieldWalker() {
	public void step(Field f) {
	    if (f instanceof ColumnField) {
		Column col = ((ColumnField)f).getColumn();
		tables.add(col.getTable());
	    }
	}
	});

    for (Iterator iter = tables.iterator(); iter.hasNext(); ) {
	Table table = (Table)iter.next();
	DefaultMutableTreeNode tableNode =
	    new DefaultMutableTreeNode(table.getName());
	categoryNode.add(tableNode);
	for (Iterator c_iter = table.columns(); c_iter.hasNext(); ) {
	    Column column = (Column)c_iter.next();
	    ColumnInfo info = new ColumnInfo(column, designWin);
	    tableNode.add(new DefaultMutableTreeNode(info, false));
	}
    }
}

/**
 * Creates nodes representing formula fields.
 *
 * @param top top-level tree node
 */
protected void createFormulas(DefaultMutableTreeNode top) {
    formulaCategoryNode =
	new DefaultMutableTreeNode(I18N.get("FieldPickerWin.formulas"));
    top.add(formulaCategoryNode);

    TreeSet formulas = new TreeSet(nameComparator);

    for (Iterator iter = report.formulas(); iter.hasNext(); )
	formulas.add(iter.next());

    for (Iterator iter = formulas.iterator(); iter.hasNext(); ) {
	Formula f = (Formula)iter.next();
	FormulaInfo info = new FormulaInfo(report, f, designWin);
	formulaCategoryNode.add(new DefaultMutableTreeNode(info));
	f.addObserver(this);
    }
}

/**
 * Creates nodes representing parameter fields.
 *
 * @param top top-level tree node
 */
protected void createParameters(DefaultMutableTreeNode top) {
    parameterCategoryNode =
	new DefaultMutableTreeNode(I18N.get("FieldPickerWin.parameters"));
    top.add(parameterCategoryNode);

    TreeSet parameters = new TreeSet(nameComparator);

    for (Iterator iter = report.parameters(); iter.hasNext(); )
	parameters.add(iter.next());

    for (Iterator iter = parameters.iterator(); iter.hasNext(); ) {
	Parameter p = (Parameter)iter.next();
	ParameterInfo info = new ParameterInfo(report, p, designWin);
	parameterCategoryNode.add(new DefaultMutableTreeNode(info));
	p.addObserver(this);
    }
}

/**
 * Creates nodes representing user column fields.
 *
 * @param top top-level tree node
 */
protected void createUserColumns(DefaultMutableTreeNode top) {
    userColumnCategoryNode =
	new DefaultMutableTreeNode(I18N.get("FieldPickerWin.usercols"));
    top.add(userColumnCategoryNode);

    TreeSet usercols = new TreeSet(nameComparator);

    for (Iterator iter = report.userColumns(); iter.hasNext(); )
	usercols.add(iter.next());

    for (Iterator iter = usercols.iterator(); iter.hasNext(); ) {
	UserColumn uc = (UserColumn)iter.next();
	UserColumnInfo info = new UserColumnInfo(report, uc, designWin);
	userColumnCategoryNode.add(new DefaultMutableTreeNode(info));
	uc.addObserver(this);
    }
}

/**
 * Creates nodes representing each possible special field.
 *
 * @param top top-level tree node
 */
protected void createSpecialFields(DefaultMutableTreeNode top) {
    DefaultMutableTreeNode categoryNode =
	new DefaultMutableTreeNode(I18N.get("FieldPickerWin.specials"));
    top.add(categoryNode);

    HashMap strs = SpecialField.specialFieldNames();
    TreeSet sortedKeys = new TreeSet(strs.keySet());

    for (Iterator iter = sortedKeys.iterator(); iter.hasNext(); ) {
	String key = (String)iter.next();
	String val = (String)strs.get(key);
	key = SpecialField.TYPE_STRING + ':' + key;
	SpecialInfo info = new SpecialInfo(val, key, designWin);
					   
	DefaultMutableTreeNode node = new DefaultMutableTreeNode(info);
	categoryNode.add(node);
    }
}

/**
 * Creates notes representing tables and columns for all tables in the
 * database.
 *
 * @param top top-level tree node
 */
protected void createAllDatabaseTables(DefaultMutableTreeNode top) {
    DefaultMutableTreeNode categoryNode =
	new DefaultMutableTreeNode(I18N.get("FieldPickerWin.all"));
    top.add(categoryNode);

    for (Iterator iter = report.getDatabase().tables(); iter.hasNext(); ) {
	Table table = (Table)iter.next();
	DefaultMutableTreeNode tableNode =
	    new DefaultMutableTreeNode(table.getName());
	categoryNode.add(tableNode);
	for (Iterator c_iter = table.columns(); c_iter.hasNext(); ) {
	    Column column = (Column)c_iter.next();
	    ColumnInfo info = new ColumnInfo(column, designWin);
	    tableNode.add(new DefaultMutableTreeNode(info));
	}
    }
}

/**
 * Opens a name editor to (re)name a nameable object. Returns <code>true</code>
 * if the user clicked OK, <code>false</code> if the user clicked Cancel.
 *
 * @param nameable a nameable object
 * @param editTitleKey I18N lookup key for "edit" title
 * @param newTitleKey I18N lookup key for "new" title
 * @param promptKey I18N lookup key for prompt
 * @param unnamedKey I18N lookup key for "unnamed" name
 * @return <code>true</code> if the user clicked OK, <code>false</code>
 * if the user clicked Cancel
 */
protected boolean rename(Nameable nameable, String newTitleKey,
			 String editTitleKey, String promptKey,
			 String unnamedKey)
{
    String oldName = nameable.getName();
    String title = (oldName == null || oldName.length() == 0)
	? I18N.get(newTitleKey) : I18N.get(editTitleKey);
    String name = new AskStringDialog((JFrame)this.getOwner(), title,
				      I18N.get(promptKey), oldName)
	.getString();

    if (name == null)		// User cancelled
	return false;

    if (name.length() == 0)
	name = I18N.get(unnamedKey);
    designWin.performCommand(new NameableRenameCommand(nameable, oldName,
						       name));
    return true;
}

/**
 * Adds a newly created editable object to the tree, makes it visible,
 * and opens its editor.
 *
 * @param info describes what is being added to add to the tree
 * @param categoryNode where in the tree to put the new item
 */
protected void addEditableToTree(FPLeafInfo info,
				 DefaultMutableTreeNode categoryNode)
{
    // Add to tree
    DefaultMutableTreeNode node = new DefaultMutableTreeNode(info);
    treeModel.insertNodeInto(node, categoryNode, categoryNode.getChildCount());

    // Make leaf visible and make it the current selection
    TreePath path = new TreePath(node.getPath());
    tree.scrollPathToVisible(path);
    tree.setSelectionPath(path);

    editSelection();		// Open editor
}


/**
 * Creates a new formula, adds it to the report and the tree, and opens
 * the formula editor.
 */
protected void newFormula() {
    // Create formula and let user enter name
    Formula f = new Formula(null, report, "", "");
    if (renameFormula(f) == false)
	return;			// User cancelled

    // Add to report and to tree and open editor
    report.addFormula(f);
    addEditableToTree(new FormulaInfo(report, f, designWin),
		      formulaCategoryNode);
}

/**
 * Opens a name editor to rename the currently select formula. Returns
 * <code>true</code> if the user clicked OK, <code>false</code> if the
 * user clicked Cancel.
 *
 * @return <code>true</code> if the user clicked OK, <code>false</code>
 * if the user clicked Cancel
 */
protected boolean renameFormula() {
    return renameFormula((Formula)selectedInfo.getLeaf());
}

/**
 * Opens a name editor to (re)name a formula. Returns <code>true</code>
 * if the user clicked OK, <code>false</code> if the user clicked Cancel.
 *
 * @param f a formula
 * @return <code>true</code> if the user clicked OK, <code>false</code>
 * if the user clicked Cancel
 */
protected boolean renameFormula(Formula f) {
    return rename(f, "FieldPickerWin.new_formula_name_title",
		  "FieldPickerWin.edit_formula_name_title",
		  "FieldPickerWin.formula_name_prompt",
		  "FieldPickerWin.unnamed_formula");
}

/**
 * Opens an editor on the currently selected item.
 */
protected void editSelection() {
    selectedInfo.openEditor();
}

/**
 * Creates a new parameter, adds it to the report and the tree, and opens
 * the parameter editor.
 */
protected void newParameter() {
    // Create parameter and let user edit it.
    Parameter p = new Parameter(null, report);
    if (renameParameter(p) == false)
	return;			// User cancelled

    // Add to report and to tree and open editor
    report.addParameter(p);
    addEditableToTree(new ParameterInfo(report, p, designWin),
		      parameterCategoryNode);
}

/**
 * Opens a name editor to rename the currently select parameter. Returns
 * <code>true</code> if the user clicked OK, <code>false</code> if the
 * user clicked Cancel.
 *
 * @return <code>true</code> if the user clicked OK, <code>false</code>
 * if the user clicked Cancel
 */
protected boolean renameParameter() {
    return renameParameter((Parameter)selectedInfo.getLeaf());
}

/**
 * Opens a name editor to (re)name a parameter. Returns <code>true</code>
 * if the user clicked OK, <code>false</code> if the user clicked Cancel.
 *
 * @param f a parameter
 * @return <code>true</code> if the user clicked OK, <code>false</code>
 * if the user clicked Cancel
 */
protected boolean renameParameter(Parameter p) {
    return rename(p, "FieldPickerWin.new_param_name_title",
		  "FieldPickerWin.edit_param_name_title",
		  "FieldPickerWin.param_name_prompt",
		  "FieldPickerWin.unnamed_parameter");
}

/**
 * Creates a new user column, adds it to the report and the tree, and opens
 * the user column editor.
 */
protected void newUserColumn() {
    // Create user column and let user enter name
    UserColumn uc = new UserColumn(null, report, "");
    if (renameUserColumn(uc) == false)
	return;			// User cancelled

    // Add to report and to tree and open editor
    report.addUserColumn(uc);
    addEditableToTree(new UserColumnInfo(report, uc, designWin),
		      userColumnCategoryNode);
}

/**
 * Opens a name editor to rename the currently select user column. Returns
 * <code>true</code> if the user clicked OK, <code>false</code> if the
 * user clicked Cancel.
 *
 * @return <code>true</code> if the user clicked OK, <code>false</code>
 * if the user clicked Cancel
 */
protected boolean renameUserColumn() {
    return renameUserColumn((UserColumn)selectedInfo.getLeaf());
}

/**
 * Opens a name editor to (re)name a user column. Returns <code>true</code>
 * if the user clicked OK, <code>false</code> if the user clicked Cancel.
 *
 * @param f a user column
 * @return <code>true</code> if the user clicked OK, <code>false</code>
 * if the user clicked Cancel
 */
protected boolean renameUserColumn(UserColumn f) {
    return rename(f, "FieldPickerWin.new_usercol_name_title",
		  "FieldPickerWin.edit_usercol_name_title",
		  "FieldPickerWin.usercol_name_prompt",
		  "FieldPickerWin.unnamed_usercol");
}

/**
 * Handles user actions. Actions are only allowed when legal. For example,
 * the "Cut" menu item will only be enabled when a delete operation is
 * possible.
 */
public void actionPerformed(ActionEvent e) {
    String cmd = e.getActionCommand();

    // File menu
    if ("close".equals(cmd)) dispose();

    // Edit menu
    else if ("cut".equals(cmd))
	designWin.performCommand(new FPCutCommand(report, designWin, tree,
						  selectedNode));

    // Field menu
    else if ("new_formula".equals(cmd)) newFormula();
    else if ("edit_formula".equals(cmd)) editSelection();
    else if ("rename_formula".equals(cmd)) renameFormula();
    else if ("new_parameter".equals(cmd)) newParameter();
    else if ("edit_parameter".equals(cmd)) editSelection();
    else if ("rename_parameter".equals(cmd)) renameParameter();
    else if ("new_user_column".equals(cmd)) newUserColumn();
    else if ("edit_user_column".equals(cmd)) editSelection();
    else if ("rename_user_column".equals(cmd)) renameUserColumn();
}

/**
 * Modifies the menu in response to a change in the tree's selection.
 */
public void valueChanged(TreeSelectionEvent e) {
    if (e.isAddedPath()) {
	selectedNode
	    = (DefaultMutableTreeNode)e.getPath().getLastPathComponent();
	Object obj = selectedNode.getUserObject();
	if (obj instanceof FPLeafInfo)
	    selectedInfo = (FPLeafInfo)obj;
	else
	    selectedInfo = null;

	if (selectedInfo == null) {
	    cutItem.setEnabled(false);
	    editFormulaItem.setEnabled(false);
	    editParameterItem.setEnabled(false);
	    editUserColumnItem.setEnabled(false);
	    renameFormulaItem.setEnabled(false);
	    renameParameterItem.setEnabled(false);
	    renameUserColumnItem.setEnabled(false);
	}
	else {
	    boolean isFormula = selectedInfo instanceof FormulaInfo;
	    boolean isParameter = selectedInfo instanceof ParameterInfo;
	    boolean isUserColumn = selectedInfo instanceof UserColumnInfo;

	    cutItem.setEnabled(selectedInfo.isDeletable());
	    editFormulaItem.setEnabled(isFormula);
	    editParameterItem.setEnabled(isParameter);
	    editUserColumnItem.setEnabled(isUserColumn);
	    renameFormulaItem.setEnabled(isFormula);
	    renameParameterItem.setEnabled(isParameter);
	    renameUserColumnItem.setEnabled(isUserColumn);
	}

    }
    else {
	selectedInfo = null;
	cutItem.setEnabled(false);
	editFormulaItem.setEnabled(false);
	editParameterItem.setEnabled(false);
	editUserColumnItem.setEnabled(false);
	renameFormulaItem.setEnabled(false);
	renameParameterItem.setEnabled(false);
	renameUserColumnItem.setEnabled(false);
    }
}

}
