package jimm.datavision.gui;
import jimm.datavision.I18N;
import jimm.datavision.PaperFormat;
import java.awt.Font;
import java.awt.event.*;
import java.util.Map;
import java.util.Iterator;
import javax.swing.*;

/**
 * Menu creation utilities.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class MenuUtils {

/**
 * Adds a single item to a popup menu.
 *
 * @param listener action listener for item; may be <code>null</code>
 * @param menu the menu
 * @param title the menu item' title
 * @param font font; may be <code>null</code>
 */
public static JMenuItem addToMenu(ActionListener listener, JPopupMenu menu,
				  String title, Font font)
{
    JMenuItem item = new JMenuItem(title);
    if (listener != null)
	item.addActionListener(listener);
    if (font != null)
	item.setFont(font);
    menu.add(item);
    return item;
}

/**
 * Adds a single item to a menu.
 *
 * @param listener action listener for item; may be <code>null</code>
 * @param menu the menu
 * @param title the menu item's title; also used as the command action string
 */
public static JMenuItem addToMenu(ActionListener listener, JMenu menu,
				  String title)
{
    return addToMenu(listener, menu, title, title, 0, 0, 0);
}

/**
 * Adds a single item to a menu.
 *
 * @param listener action listener for item; may be <code>null</code>
 * @param menu the menu
 * @param title the menu item's title
 * @param action the action string
 */
public static JMenuItem addToMenu(ActionListener listener, JMenu menu,
				  String title, String action)
{
    return addToMenu(listener, menu, title, action, 0, 0, 0);
}

/**
 * Adds a single item to a menu.
 *
 * @param listener action listener for item; may be <code>null</code>
 * @param menu the menu
 * @param title the menu item's title
 * @param action the action string
 * @param accelKey the accelerator key
 */
public static JMenuItem addToMenu(ActionListener listener, JMenu menu,
				  String title, String action, int accelKey)
{
    return addToMenu(listener, menu, title, action, accelKey, accelKey,
		     ActionEvent.CTRL_MASK);
}

/**
 * Adds a single item to a menu.
 *
 * @param listener action listener for item; may be <code>null</code>
 * @param menu the menu
 * @param title the menu item's title
 * @param action the action string
 * @param accelKey the accelerator key
 * @param shortcutKey the shortcut key
 */
public static JMenuItem addToMenu(ActionListener listener, JMenu menu,
				  String title, String action, int accelKey,
				  int shortcutKey)
{
    return addToMenu(listener, menu, title, action, accelKey, shortcutKey,
		     ActionEvent.CTRL_MASK);
}

/**
 * Adds a single item to a menu.
 *
 * @param listener action listener for item; may be <code>null</code>
 * @param menu the menu
 * @param title the menu item's title
 * @param action the action string
 * @param accelKey the accelerator key
 * @param shortcutKey the shortcut key
 * @param mask the modifier mask (for example, ActionEvent.CTRL_MASK)
 */
public static JMenuItem addToMenu(ActionListener listener, JMenu menu,
				  String title, String action, int accelKey,
				  int shortcutKey, int mask)
{
    JMenuItem item = new JMenuItem(title);
    setKeys(item, accelKey, shortcutKey, mask);
    if (action != null)
	item.setActionCommand(action);
    item.addActionListener(listener);
    menu.add(item);
    return item;
}

/**
 * Sets accelerator and shortcut keys. The shortcut key will be the same
 * as the accelerator key, and the shortcut key will use the default
 * modifier mask of <code>ActionEvent.CTRL_MASK</code>.
 *
 * @param accelKey the accelerator key
 */
public static void setKeys(JMenuItem item, int accelKey)
{
    setKeys(item, accelKey, accelKey, ActionEvent.CTRL_MASK);
}

/**
 * Sets accelerator and shortcut keys. The shortcut key will use the default
 * modifier mask of <code>ActionEvent.CTRL_MASK</code>.
 *
 * @param accelKey the accelerator key
 * @param shortcutKey the shortcut key
 */
public static void setKeys(JMenuItem item, int accelKey, int shortcutKey)
{
    setKeys(item, accelKey, shortcutKey, ActionEvent.CTRL_MASK);
}

/**
 * Sets accelerator and shortcut keys.
 *
 * @param accelKey the accelerator key
 * @param shortcutKey the shortcut key
 * @param mask the modifier mask (for example, ActionEvent.CTRL_MASK)
 */
public static void setKeys(JMenuItem item, int accelKey, int shortcutKey,
			   int mask)
{
    if (accelKey != 0)
	item.setMnemonic(accelKey);
    if (shortcutKey != 0)
	item.setAccelerator(KeyStroke.getKeyStroke(shortcutKey, mask));
}

/**
 * Returns a new align menu.
 *
 * @param listener action listener for item; may be <code>null</code>
 * @param useMnemonicKeys if <code>true</code>, mnemonic keys will be
 * assigned to the menu and its items
 * @param font font; may be <code>null</code>
 * @return a new menu
 */
public static JMenu buildAlignMenu(ActionListener listener,
				   boolean useMnemonicKeys, Font font)
{
    JMenu menu = new JMenu(I18N.get("Align.menu"));
    menu.setMnemonic(useMnemonicKeys ? KeyEvent.VK_A : 0);

    JMenuItem item = null;
    item = MenuUtils.addToMenu(listener, menu, I18N.get("Align.menu_tops"),
			       "align_top",
			       useMnemonicKeys ? KeyEvent.VK_T : 0, 0);
    if (font != null) item.setFont(font);
    item = MenuUtils.addToMenu(listener, menu, I18N.get("Align.menu_middles"),
			       "align_middle",
			       useMnemonicKeys ? KeyEvent.VK_M : 0, 0);
    if (font != null) item.setFont(font);
    item = MenuUtils.addToMenu(listener, menu, I18N.get("Align.menu_bottoms"),
			       "align_bottom",
			       useMnemonicKeys ? KeyEvent.VK_B : 0, 0);
    if (font != null) item.setFont(font);
    item = MenuUtils.addToMenu(listener, menu, I18N.get("Align.menu_lefts"),
			       "align_left",
			       useMnemonicKeys ? KeyEvent.VK_L : 0, 0);
    if (font != null) item.setFont(font);
    item = MenuUtils.addToMenu(listener,menu, I18N.get("Align.menu_centers"),
			       "align_center",
			       useMnemonicKeys ? KeyEvent.VK_C : 0, 0);
    if (font != null) item.setFont(font);
    item = MenuUtils.addToMenu(listener, menu, I18N.get("Align.menu_rights"),
			       "align_right",
			       useMnemonicKeys ? KeyEvent.VK_R : 0, 0);
    if (font != null) item.setFont(font);
    item = MenuUtils.addToMenu(listener, menu, I18N.get("Align.menu_snap"),
			       "snap_to_grid",
			       useMnemonicKeys ? KeyEvent.VK_S : 0, 0);
    if (font != null) item.setFont(font);

    return menu;
}

/**
 * Returns a new size menu.
 *
 * @param listener action listener for item; may be <code>null</code>
 * @param useMnemonicKeys if <code>true</code>, mnemonic keys will be
 * assigned to the menu and its items
 * @param font font; may be <code>null</code>
 * @return a new menu
 */
public static JMenu buildSizeMenu(ActionListener listener,
				  boolean useMnemonicKeys, Font font)
{
    JMenu menu = new JMenu(I18N.get("Size.menu"));
    menu.setMnemonic(useMnemonicKeys ? KeyEvent.VK_S : 0);

    JMenuItem item = null;
    item = MenuUtils.addToMenu(listener, menu,
			       I18N.get("Size.menu_same_width"), "size_width",
			       useMnemonicKeys ? KeyEvent.VK_W : 0, 0);
    if (font != null) item.setFont(font);
    item = MenuUtils.addToMenu(listener, menu,
			       I18N.get("Size.menu_same_height"),
			       "size_height",
			       useMnemonicKeys ? KeyEvent.VK_H : 0, 0);
    if (font != null) item.setFont(font);
    item = MenuUtils.addToMenu(listener, menu,
			       I18N.get("Size.menu_same_size"), "size_size",
			       useMnemonicKeys ? KeyEvent.VK_S : 0, 0);
    if (font != null) item.setFont(font);

    return menu;
}

/**
 * Returns a new paper size menu.
 *
 * @param listener action button listener for item; may be <code>null</code>
 * @param currChoice current paper choice (its orientation and name will be
 * pre-selected)
 * @param group a radio button group
 * @param mnemonicKey the menu's mnemonic key; may be 0
 * @param font font; may be <code>null</code>
 * @return a new menu
 */
public static JMenu buildPaperSizeMenu(ActionListener listener,
				       PaperFormat currChoice,
				       ButtonGroup orientationGroup,
				       ButtonGroup nameGroup, int mnemonicKey,
				       Font font)
{
    JMenu menu = new JMenu(I18N.get("MenuUtils.menu_paper_size"));
    menu.setMnemonic(mnemonicKey);

    JRadioButtonMenuItem item = null;

    // Portrait/landscape
    String str = I18N.get("MenuUtils.menu_paper_size_portrait");
    item = new JRadioButtonMenuItem(str);
    orientationGroup.add(item);
    if (listener != null) item.addActionListener(listener);
    if (font != null) item.setFont(font);
    if (currChoice.getOrientation() == PaperFormat.PORTRAIT)
	item.setSelected(true);
    menu.add(item);

    str = I18N.get("MenuUtils.menu_paper_size_landscape");
    item = new JRadioButtonMenuItem(str);
    orientationGroup.add(item);
    if (listener != null) item.addActionListener(listener);
    if (font != null) item.setFont(font);
    if (currChoice.getOrientation() == PaperFormat.LANDSCAPE)
	item.setSelected(true);
    menu.add(item);

    menu.addSeparator();

    // Paper sizes
    for (Iterator iter = PaperFormat.names(); iter.hasNext();) {
	String name = (String)iter.next();
	item = new JRadioButtonMenuItem(name);
	nameGroup.add(item);
	if (listener != null) item.addActionListener(listener);
	if (font != null) item.setFont(font);

	if (name.equals(currChoice.getName()))
	    item.setSelected(true);

	menu.add(item);
    }

    return menu;
}
}
