package jimm.datavision.gui;
import jimm.datavision.*;
import jimm.datavision.gui.cmd.Command;
import jimm.datavision.gui.cmd.SectionResizeCommand;
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

/**
 * A section widget is the visual representation of a report section.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class SectionWidget extends JPanel implements ActionListener {

public static final int LHS_WIDTH = 125;

protected static Font DEFAULT_POPUP_FONT = new Font("Serif", Font.PLAIN, 10);

static final Color NORMAL_COLOR = Color.white;
static final Color SUPPRESSED_COLOR = Color.lightGray;

protected String name;
protected String popupName;
protected DesignWin designWin;
protected Section section;
protected SectionNameLabel label;
protected SectionFieldPanel fieldPanel;
protected JPopupMenu popup;
protected JMenuItem nameItem, toggleSuppress, shrink;
protected JMenuItem deleteGroup, addGroup, deleteSection, insertSection;

/**
 * An inner class that handles display of the popup menu.
 */
class PopupListener extends MouseAdapter {
public void mousePressed(MouseEvent e) { maybeShowPopup(e); }
public void mouseReleased(MouseEvent e) { maybeShowPopup(e); }
private void maybeShowPopup(MouseEvent e) {
    if (e.isPopupTrigger())
	showPopup(e);
}
}

/**
 * Constructor.
 *
 * @param win parent design window
 * @param sect the report section
 * @param name a name such as "Report Header (a)"
 * @param popupName the string to display as the first, disabled popup item
 */
public SectionWidget(DesignWin win, Section sect, String name,
		     String popupName)
{
    super();

    designWin = win;
    section = sect;
    this.name = name;
    this.popupName = popupName;

    setLayout(new SectionLayout());
    setPreferredSize(new Dimension(getTotalWidth(), getHeight()));

    buildPopupMenu();
    addMouseListener(new PopupListener());

    // LHS name of section
    buildDisplayName();

    // Fields
    fieldPanel = new SectionFieldPanel(this);
    fieldPanel.setLayout(null);
    fieldPanel.setBackground(section.isSuppressed()
			     ? SUPPRESSED_COLOR : NORMAL_COLOR);
    add(fieldPanel);
    for (Iterator iter = section.fields(); iter.hasNext(); ) {
	Field f = (Field)iter.next();
	FieldWidget fw = makeWidgetFor(f);
	fieldPanel.add(fw.getComponent(), 0);	// Add to top of visual stack.
    }

    // If design window has initiated the drop of a new text field, create
    // one. Else deselect all fields.
    fieldPanel.addMouseListener(new PopupListener() {
	public void mouseClicked(MouseEvent e) {
	    deselectAll();
	    if (designWin.isPlacingNewTextField())
		createNewTextField(e); // Calls acceptNewTextField()
	}
	});

    // Resizer bar
    add(new SectionResizer(this, designWin.sectionContainer));
}

/**
 * Given a field, this factory method creates and returns the appropriate
 * widget.
 *
 * @param f a report field
 */
public FieldWidget makeWidgetFor(Field f) {
    FieldWidget fw = null;
    if (f instanceof jimm.datavision.TextField)
	fw = new TextFieldWidget(this, f);
    else if (f instanceof FormulaField)
	fw = new FormulaWidget(this, f);
    else if (f instanceof ParameterField)
	fw = new ParameterWidget(this, f);
    else if (f instanceof UserColumnField)
	fw = new UserColumnWidget(this, f);
    else if (f instanceof ImageField)
	fw = new ImageFieldWidget(this, (ImageField)f);
    else
	fw = new FieldWidget(this, f);
    return fw;
}

/**
 * Builds the popup menu.
 */
protected void buildPopupMenu() {
    popup = new JPopupMenu(popupName);

    nameItem = addPopupItem(popupName);
    nameItem.setEnabled(false);
    popup.addSeparator();
    toggleSuppress = addPopupItem(I18N.get("SectionWidget.popup_suppress"));
    popup.addSeparator();
    shrink = addPopupItem(I18N.get("SectionWidget.popup_shrink"));
    popup.addSeparator();
    deleteGroup = addPopupItem(I18N.get("SectionWidget.popup_delete_group"));
    addGroup = addPopupItem(I18N.get("SectionWidget.popup_add_group"));
    popup.addSeparator();
    deleteSection =
	addPopupItem(I18N.get("SectionWidget.popup_delete_section"));
    insertSection =
	addPopupItem(I18N.get("SectionWidget.popup_insert_section"));
}

/**
 * Adds a single menu item to the popup menu.
 *
 * @param text the menu item string
 * @return the created item
 */
protected JMenuItem addPopupItem(String text) {
    JMenuItem item = new JMenuItem(text);
    item.setFont(DEFAULT_POPUP_FONT);
    item.addActionListener(this);
    popup.add(item);
    return item;
}

/**
 * Performs some action based on the action command string (the menu
 * item text).
 */
public void actionPerformed(ActionEvent e) {
    String command = e.getActionCommand();
    if (command == null) return;
    if (command.equals(I18N.get("SectionWidget.popup_suppress")))
	toggleSuppression();
    else if (command.equals(I18N.get("SectionWidget.popup_delete_group")))
	designWin.deleteGroupContaining(section);
    else if (command.equals(I18N.get("SectionWidget.popup_add_group")))
	designWin.openNewGroupWin();
    else if (command.equals(I18N.get("SectionWidget.popup_delete_section")))
	designWin.deleteSection(section);
    else if (command.equals(I18N.get("SectionWidget.popup_insert_section")))
	designWin.insertSectionBelow(section);
    else if (command.equals(I18N.get("SectionWidget.popup_shrink")))
	shrinkToFit();
}

/**
 * Modifies menu items based on the state of the section.
 */
protected void enablePopupMenuItems() {
    nameItem.setText(popupName);
    toggleSuppress.setText(section.isSuppressed()
			   ? I18N.get("SectionWidget.popup_dont_suppress")
			   : I18N.get("SectionWidget.popup_suppress"));
    deleteSection.setEnabled(!section.getReport().isOneOfAKind(section));
    deleteGroup.setEnabled(section.getReport().isInsideGroup(section));
}

/**
 * Constructs the section name widget that is displayed to the left of
 * the section.
 */
protected void buildDisplayName() {
    label = new SectionNameLabel(name, this);
    add(label);
}

/**
 * Set the section display name.
 *
 * @param name the new name
 */
public void setDisplayName(String name) {
    this.name = name;
    label.setText(name);
}

/**
 * Set the popup menu name, also displayed as first, disabled menu item.
 *
 * @param popuNpame the new name
 */
public void setPopupName(String popupName) {
    this.popupName = popupName;
    nameItem.setText(popupName);
}

/**
 * Returns the report we are representing.
 *
 * @return a report
 */
public Report getReport() { return section.getReport(); }

/**
 * Returns the section we are representing.
 *
 * @return a report section
 */
public Section getSection() { return section; }

/**
 * Returns the area of the section (report header, page footer, etc.)
 *
 * @return one of the <code>Report.AREA_</code>* constants that defines
 * section areas
 * @see jimm.datavision.Section
 * @see jimm.datavision.Report
 */
public int getSectionArea() { return section.getArea(); }

/**
 * Returns the design window containing this section widget
 *
 * @return a design window
 */
public DesignWin getDesignWin() { return designWin; }

/**
 * Returns the width of the report paper (the white part upon which fields
 * are placed).
 *
 * @return the paper width
 */
public int getPaperWidth() {
    return (int)section.getWidth();
}

/**
 * Returns the width of the section, including the left-hand side name.
 *
 * @return the total width
 */
public int getTotalWidth() {
    return LHS_WIDTH + getPaperWidth();
}

/**
 * Returns the height of the section, including the resizer bar.
 *
 * @return the total height
 */
public int getHeight() {
    return (int)section.getHeight() + SectionResizer.HEIGHT;
}

/**
 * Returns the height of the report section.
 *
 * @return report section height
 */
public int getSectionHeight() { return (int)section.getHeight(); }

/**
 * Returns the minimum height the section needs to fit all of its fields.
 *
 * @return minimum height
 */
public int getMinSectionHeight() {
    int minY = 0;
    Component[] fieldWidgets = fieldPanel.getComponents();
    for (int i = 0; i < fieldWidgets.length; ++i) {
	int y = fieldWidgets[i].getBounds().y
	    + fieldWidgets[i].getBounds().height;
	if (y > minY)
	    minY = y;
    }
    return minY;
}

/**
 * Resizes this widget. Called by the design window whenever the user
 * selects a new paper size.
 */
void paperSizeChanged() {
    setPreferredSize(new Dimension(getTotalWidth(), getHeight()));
    invalidate();
}

/**
 * Toggles the suppressed flag of the section.
 */
void toggleSuppression() {
    boolean newState = !section.isSuppressed();
    section.setSuppressed(newState);
    fieldPanel.setSuppressed(newState);	// Background color change
}

/**
 * Shrinks this section widget to the minimum height required. This method
 * is only called from the popup menu. It should not be called as part
 * of a larger operation because it creates a command that allows undo/redo.
 */
public void shrinkToFit() {
    SectionResizeCommand cmd = new SectionResizeCommand(this);
    growBy(getMinSectionHeight() - getSectionHeight());
    performCommand(cmd);
}

/**
 * Grows this section widget to the minimum height required. This method,
 * unlike <code>shrinkToFit</code>, is always called as part of some other
 * operation.
 */
public void growToFit() {
    int dy = getMinSectionHeight() - getSectionHeight();
    if (dy > 0)
	growBy(dy);
}

/**
 * Resizes the section. Called by resizer bar and by commands that
 * grow sections as a side effect.
 *
 * @param dy delta y
 */
public void growBy(int dy) {
    if (dy == 0) return;

    // Make sure section fits all fields.
    int origHeight = (int)section.getHeight();
    int newHeight = origHeight + dy;
    int minHeight = getMinSectionHeight();
    if (newHeight < minHeight) {
	newHeight = minHeight;
	if (newHeight == origHeight)
	    return;
    }

    section.setHeight(newHeight); // Hight not including resizer
    setPreferredSize(new Dimension(getTotalWidth(), getHeight())); // Incl. resizer
    revalidate();
}

/**
 * Grows or shrinks the section widget and executes a command that allows
 * this action to be undone/redone. Calls <code>growBy</code> to grow
 * or shrink, then lets our window execute the command that remembers
 * the size change for later undo/redo.
 * <p>
 * The command does not change our height. It remembers the old and new
 * heights for later undo/redo.
 *
 * @param dy delta height
 * @param cmd a section resize command
 * @see #growBy
 * @see #performCommand
 */
public void resizeBy(int dy, SectionResizeCommand cmd) {
    growBy(dy);
    performCommand(cmd);
}

/**
 * Passes a command up to the design window for execution.
 *
 * @param cmd a command
 * @see DesignWin#performCommand
 */
public void performCommand(Command cmd) {
    designWin.performCommand(cmd);
}

/**
 * Passes responsiblity up to the design window.
 * @see DesignWin.setIgnoreKeys
 */
public void setIgnoreKeys(boolean ignoreKeys) {
    designWin.setIgnoreKeys(ignoreKeys);
}

/**
 * Passes this request up to the design window.
 *
 * @param x where to place the title
 * @param width how wide it should be
 * @param title the string to display
 * @return the newly-created widget
 * @see DesignWin#addTitleField
 */
public FieldWidget addTitleField(int x, int width, String title) {
    return designWin.addTitleField(x, width, title);
}

/**
 * Passes on to the design window the request to pick up all selected
 * fields for dragging (not just the specified field). Called from
 * <code>FieldWidget.mousePressed</code>.
 *
 * @param f the field in which the mouse has been clicked
 * @see FieldWidget#mousePressed
 * @see DesignWin#pickUp
 */
void pickUp(FieldWidget f) { designWin.pickUp(f); }

/**
 * Passes on to the design window the request to put down all fields
 * being dragged (not just the specified field). Called from
 * <code>FieldWidget.mouseReleased</code>.
 *
 * @param f the field in which the mouse has been clicked
 * @param origScreenPos the original location of the field in screen
 * coordinates
 * @param mouseScreenPos the location of the mouse in screen coordinates
 * @see FieldWidget#mouseReleased
 * @see DesignWin#putDown
 */
void putDown(FieldWidget f, java.awt.Point origScreenPos,
	     java.awt.Point mouseScreenPos)
{
    designWin.putDown(f, origScreenPos, mouseScreenPos);
}

/**
 * Tells the window to drag (move, resize) all selected fields. Called
 * from one field widget when it's being manipulated with the mouse.
 *
 * @param action a <code>FieldWidget.ACTION_*</code> constant
 * @param dx delta x
 * @param dy delta y
 * @see FieldWidget#mouseDragged
 * @see DesignWin#dragSelectedWidgets
 */
void dragSelectedWidgets(int action, int dx, int dy) {
    designWin.dragSelectedWidgets(action, dx, dy);
}

/**
 * Selects or deselcts a field widget, possibly deselecting all others
 * everywhere. Called from field widget itself, this passes the request
 * on to the design window.
 *
 * @param FieldWidget a field widget
 * @param makeSelected if <code>true</code>, select this field; else
 * deselect it
 * @param deselectAll if <code>true</code>, all other fields in all sections
 * are deselected first
 * @see DesignWin#select
 */
void select(FieldWidget fieldWidget, boolean makeSelected,
	    boolean deselectOthers)
{
    designWin.select(fieldWidget, makeSelected, deselectOthers);
}

/**
 * Deselects all fields in all sections. Tells the design window to do
 * so.
 *
 * @see DesignWin#deselectAll
 */
void deselectAll() {
    designWin.deselectAll();
}

/**
 * Adds field widget to panel. Does not affect models. Field retains its
 * selection state.
 *
 * @param fw field widget to add
 */
public void addField(FieldWidget fw) {
    fieldPanel.add(fw.getComponent(), 0); // Add to top of visual stack.
    fw.sectionWidget = this;
    fw.getComponent().setBackground(section.isSuppressed()
				    ? SUPPRESSED_COLOR : NORMAL_COLOR);
    growToFit();
}

/**
 * Removes field widget from panel, but do not change field model's relation
 * with section model. Field retains its selection state.
 * <p>
 * To delete a field widget completely, see
 * <code>DesignWin.deleteSelectedFields</code>.
 *
 * @param fw field widget to remove
 * @see #addField
 * @see DesignWin#deleteSelectedFields
 */
public void removeField(FieldWidget fw) {
    fieldPanel.remove(fw.getComponent());
}

/**
 * Asks the design window to snap the rectangle to it's grid.
 *
 * @param r a rectangle
 */
void snapToGrid(jimm.datavision.Rectangle r) {
    designWin.snapToGrid(r);
}

/**
 * Sets the visibility of all selected fields plus the one passed in.
 * Passes the buck to the design window.
 *
 * @see FieldWidget#toggleVisibility
 */
void setFieldVisibility(boolean newVisiblity, FieldWidget fw) {
    designWin.setFieldVisibility(newVisiblity, fw);
}

/**
 * Asks design window to create and accepts a new text field.
 *
 * @see DesignWin#createNewTextField
 */
void createNewTextField(MouseEvent e) {
    designWin.createNewTextField(this, e);
}

/**
 * Displays popup menu, after enabling and disabling menu items.
 *
 * @param e mouse event that caused popup to do its thing
 */
void showPopup(MouseEvent e) {
    enablePopupMenuItems();
    popup.show(e.getComponent(), e.getX(), e.getY());
}

}
