package jimm.datavision.gui;
import jimm.datavision.*;
import jimm.datavision.gui.cmd.NewSubtotalCommand;
import jimm.datavision.gui.cmd.DeleteSubtotalCommand;
import java.awt.BorderLayout;
import java.awt.GridLayout;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;

/**
 * A subtotals editor for a single field that lets the user add and delete
 * subtotals at each level (group footers, report footer). Never called
 * when multiple fields are selected.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class SubtotalsWin extends EditWin {

class Slot {

FieldWidget subtotal;
boolean existedAlready;
JCheckBox checkBox;

Slot(FieldWidget fw, JCheckBox cb) {
    existedAlready = fw != null;
    setSubtotal(fw);
    checkBox = cb;
}
void setSubtotal(FieldWidget fw) {
    subtotal = fw;
    if (fw != null) {
	SubtotalField sf = (SubtotalField)fw.getField();
	Group group = sf.getGroup();
    }
}
}

protected static final int TEXT_FIELD_COLS = 8;

protected Report report;
protected FieldWidget fieldWidget;
protected HashMap slots;

/**
 * Constructor.
 *
 * @param designWin the window to which this dialog belongs
 * @param fw a field widget
 */
public SubtotalsWin(DesignWin designWin, FieldWidget fw) {
    super(designWin, I18N.get("SubtotalsWin.title_prefix") + ' ' + fw,
	  "SubtotalsWin.command_name");

    fieldWidget = fw;
    report = designWin.report;

    // If we have been handed a subtotal field's widget, find the widget of
    // the original field.
    if (fieldWidget.getField() instanceof SubtotalField) {
	final Field fieldToSubtotal =
	    ((SubtotalField)fieldWidget.getField()).getField();
	final FieldWidget[] list = new FieldWidget[1];

	designWin.withWidgetsDo(new FieldWidgetWalker() {
	    public void step(FieldWidget fw) {
		Field f = fw.getField();
		if (f == fieldToSubtotal)
		    list[0] = fw;
	    }
	    });
	fieldWidget = list[0];
	 // Reset window title
	setTitle(I18N.get("SubtotalsWin.title_prefix") + ' ' + fieldWidget);
    }

    // Create a hash that maps either a group or a report footer section to
    // the associated existing subtotal widget.
    final HashMap subtotals = new HashMap();
    final AbstractList subs =
	designWin.report.getSubtotalFieldsFor(fieldWidget.getField());
    designWin.withWidgetsDo(new FieldWidgetWalker() {
	public void step(FieldWidget fw) {
	    Field f = fw.getField();
	    if (subs.contains(f)) {
		Object key = ((SubtotalField)f).getGroup();
		if (key == null)
		    key = fw.getField().getSection();
		subtotals.put(key, fw);
	    }
	}
	});

    buildWindow(subtotals);
    pack();
    setVisible(true);
}

/**
 * Builds the window contents.
 *
 * @param subtotals a hash that maps either a group or a report footer
 * section to the associated existing subtotal widget
 */
protected void buildWindow(HashMap subtotals) {
    JPanel editorPanel = buildSubtotalsEditor(subtotals);

    // OK, Apply, Revert, and Cancel Buttons
    JPanel buttonPanel = closeButtonPanel();

    // Add title, values, and buttons to window
    getContentPane().add(new JLabel(I18N.get("SubtotalsWin.title_prefix")
				    + ' ' + fieldWidget),
			 BorderLayout.NORTH);
    getContentPane().add(editorPanel, BorderLayout.CENTER);
    getContentPane().add(buttonPanel, BorderLayout.SOUTH);
}

/**
 * Builds the editor widget panel.
 *
 * @param subtotals a hash that maps either a group or a report footer
 * section to the associated existing subtotal widget
 */
protected JPanel buildSubtotalsEditor(HashMap subtotals) {
    JPanel panel = new JPanel();
    panel.setLayout(new GridLayout(0, 1));
    panel.setBorder(BorderFactory.createEmptyBorder(20, 20, 20, 20));
    JCheckBox cb = null;
    slots = new HashMap();

    // For each footer in the report, create a checkbox and a slot.

    // Group footer sections
    int i = report.countGroups();
    for (Iterator iter = report.groupsReversed(); iter.hasNext(); --i) {
	Group g = (Group)iter.next();
	FieldWidget fw = (FieldWidget)subtotals.get(g);

	cb = new JCheckBox(I18N.get("SubtotalsWin.group") + " #" + i + " ("
			   + g.getColumn().fullName() + ")");
	if (fw != null)
	    cb.setSelected(true);
	panel.add(cb);
	slots.put(g, new Slot(fw, cb));
    }

    // Report footer sections
    cb = new JCheckBox(I18N.get("SubtotalsWin.grand_total"));
    panel.add(cb);
    boolean addedSlot = false;
    for (Iterator iter = report.footers(); iter.hasNext() && !addedSlot; ) {
	Section s = (Section)iter.next();
	FieldWidget fw = (FieldWidget)subtotals.get(s);
	if (fw != null) {
	    cb.setSelected(true);
	    slots.put(null, new Slot(fw, cb));
	    addedSlot = true;
	}
    }
    if (!addedSlot)
	slots.put(null, new Slot(null, cb));

    // Add an "all" button
    JButton all = new JButton(I18N.get("GUI.all"));
    all.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent e) {
	    for (Iterator iter = slots.values().iterator();
		 iter.hasNext(); )
		((Slot)iter.next()).checkBox.setSelected(true);
	}
	});
    JPanel buttonPanel = new JPanel();
    buttonPanel.add(all);
    panel.add(buttonPanel);

    return panel;
}

protected void doSave() {
    // Hide existing subtotals that are unchecked and show existing
    // subtotals that are checked.
    for (Iterator iter = slots.keySet().iterator(); iter.hasNext(); ) {
	Object key = iter.next();
	Slot info = (Slot)slots.get(key);
	if (info.checkBox.isSelected()) {
	    if (info.subtotal == null) {
		NewSubtotalCommand cmd =
		    new NewSubtotalCommand(report, fieldWidget, (Group)key);
		cmd.perform();
		commands.add(cmd);

		info.setSubtotal(cmd.getSubtotalWidget());
	    }
	}
	else {			// We don't want a subtotal for this slot
	    if (info.subtotal != null) {
		DeleteSubtotalCommand cmd =
		    new DeleteSubtotalCommand(report, fieldWidget,
					      info.subtotal, (Group)key);
		cmd.perform();
		commands.add(cmd);
		info.setSubtotal(null);
	    }
	}
    }
}

/**
 * Create new subtotal widget and add it to the proper section.
 *
 * @param key group or <code>null</code>, identifying the footer that
 * will contain the subtotal.
 */
protected void createNewSubtotal(Object key) {
    Group group = (key == null) ? null : (Group)key;

    // Find the proper footer. Don't worry; report and group footers always
    // have at least one section.
    Iterator footers =
	(group == null) ? report.footers() : group.footers();
    Section s = (Section)footers.next();

    Field originalField = fieldWidget.getField();
    Long id = originalField.getId();
//      if (originalField instanceof SubtotalField)	// Can't be true
//  	id = ((SubtotalField)originalField).getFieldId();

    // Create the subtotal field. Set the group (OK if null).
    SubtotalField f =
	(SubtotalField)Field.create(null, report, s, "subtotal", id, true);
    f.setGroup(group);
		
    // Copy bounds and format from the field we are subtotaling. Make bold.
    f.setBounds(originalField.getBounds());
    Format originalFormat = originalField.getFormat() != null
	? originalField.getFormat() : Format.DEFAULT;
    Format fmt = (Format)originalFormat.clone();
    fmt.setBold(true);
    f.setFormat(fmt);

    // Add the field to the section
    s.addField(f);

    // Create the widget and add it to the proper section widget.
    SectionWidget sectionWidget = designWin.findSectionWidgetFor(s);
    FieldWidget fw = new FieldWidget(null, f);
    sectionWidget.addField(fw);

    // For some reason we need to force a repaint of the section. Can't
    // just call invalidate.
    sectionWidget.repaint();

    Slot slot = (Slot)slots.get(key);
    slot.setSubtotal(fw);
}

protected void doRevert() {
    // revertInfo is a dummy object; we already have all information we need.
    for (Iterator iter = slots.keySet().iterator(); iter.hasNext(); ) {
	Slot info = (Slot)slots.get(iter.next());
	java.awt.Component parent = info.subtotal.getComponent().getParent();
	if (info.existedAlready) {
	    info.subtotal.getComponent().setVisible(true);
	    info.checkBox.setSelected(true);
	}
	else {
	    if (info.subtotal != null) {
		info.subtotal.doDelete();
		info.subtotal = null;
	    }
	    info.checkBox.setSelected(false);
	}
	// For some reason we need to force a repaint of the section. Can't
	//just call invalidate.
	parent.repaint();
    }
}

}
