package jimm.datavision.gui;
import jimm.datavision.I18N;
import jimm.datavision.Report;
import jimm.datavision.Query;
import jimm.datavision.Join;
import jimm.datavision.sql.*;
import jimm.datavision.gui.cmd.TableJoinCommand;
import java.awt.BorderLayout;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.util.*;
import javax.swing.*;

/**
 * The dialog used for defining joins between database tables. Eventually,
 * this will be nice and graphical.
 * <p>
 * <p>
 * This dialog should only be created if the report uses more than one
 * table. The method <code>DesignWin#enableMenuItems</code> makes sure
 * this is true.
 *
 * @see DesignWin#enableMenuItems
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class VisTableWin extends EditWin implements ActionListener {

/* ================================================================ */
class RevertInfo {
ArrayList joins;
RevertInfo(Query q) {
    joins = new ArrayList();
    for (Iterator iter = q.joins(); iter.hasNext(); ) {
	Join j = (Join)iter.next();
	joins.add(j.clone());
    }
}
}

/* ================================================================ */
class JoinFields {

Join join;			// If null, it's a newly created join
JCheckBox del;
JComboBox from;
JComboBox relation;
JComboBox to;

JoinFields(Join j, Database db) {
    join = j;
    del = new JCheckBox();
    from = buildColDropdown((j == null) ? null : j.getFrom(), db.tables());
    relation = buildRelDropdown((j == null) ? null : j.getRelation());
    to = buildColDropdown((j == null) ? null : j.getTo(), db.tables());
}

protected JComboBox buildColDropdown(Column col, Iterator tables) {
    JComboBox cb = new JComboBox();
    while (tables.hasNext()) {
	Table t = (Table)tables.next();
	for (Iterator cols = t.columns(); cols.hasNext(); ) {
	    Column c = (Column)cols.next();
	    cb.addItem(c.fullName());
	    if (col != null && col.equals(c))
		cb.setSelectedItem(c.fullName());
	}
    }
    return cb;
}

protected JComboBox buildRelDropdown(String rel) {
    JComboBox cb = new JComboBox(Join.RELATIONS);
    if (rel != null)
	cb.setSelectedItem(rel);
    return cb;
}
}

/* ================================================================ */

protected Report report;
protected Query query;
protected ArrayList joinFieldsList;
protected JPanel joinsPanel;
protected Box delCheckBoxPanel;
protected Box fromPanel;
protected Box relationPanel;
protected Box toPanel;
protected JButton deleteButton;

/**
 * Constructor.
 *
 * @param designWin the window to which this dialog belongs
 * @param report the...um...I forgot
 */
public VisTableWin(DesignWin designWin, Report report) {
    super(designWin, I18N.get("VisTableWin.title"), "TableJoinCommand.name");
    this.report = report;
    query = report.getQuery();

    buildWindow();
    pack();
    setVisible(true);
}

/**
 * Builds the window contents.
 *
 * @param whichTab the index of the tab to display when opened
 */
protected void buildWindow() {
    // Joins
    buildJoinsPanel();

    // Add/Delete buttons
    JPanel addDelButtons = new JPanel();
    JButton button = new JButton(I18N.get("VisTableWin.add"));
    button.addActionListener(this);
    addDelButtons.add(button);

    deleteButton = new JButton(I18N.get("VisTableWin.delete_selected"));
    deleteButton.addActionListener(this);
    addDelButtons.add(deleteButton);

    // Panel containing joins and add/delete buttons
    JPanel centerPanel = new JPanel();
    centerPanel.setLayout(new BorderLayout());
    centerPanel.add(joinsPanel, BorderLayout.CENTER);
    centerPanel.add(addDelButtons, BorderLayout.SOUTH);

    // OK, Apply, Revert, and Cancel Buttons
    JPanel buttonPanel = closeButtonPanel();

    // Add tables and buttons to window
    getContentPane().add(centerPanel, BorderLayout.CENTER);
    getContentPane().add(buttonPanel, BorderLayout.SOUTH);
}

protected void buildJoinsPanel() {
    joinsPanel = new JPanel();
    joinsPanel.setBorder(BorderFactory.createEmptyBorder(20, 20, 20, 20));

    Box panel = Box.createHorizontalBox();
    joinsPanel.add(panel);

    delCheckBoxPanel = Box.createVerticalBox();
    fromPanel = Box.createVerticalBox();
    relationPanel = Box.createVerticalBox();
    toPanel = Box.createVerticalBox();

    fillJoinsPanel();

    panel.add(delCheckBoxPanel);
    panel.add(fromPanel);
    panel.add(relationPanel);
    panel.add(toPanel);
}

protected void fillJoinsPanel() {
    joinFieldsList = new ArrayList();
    for (Iterator iter = query.joins(); iter.hasNext(); ) {
	Join j = (Join)iter.next();
	joinFieldsList.add(new JoinFields(j, report.getDatabase()));
    }

    for (Iterator iter = joinFieldsList.iterator(); iter.hasNext(); ) {
	JoinFields jf = (JoinFields)iter.next();
	delCheckBoxPanel.add(jf.del);
	fromPanel.add(jf.from);
	relationPanel.add(jf.relation);
	toPanel.add(jf.to);
    }
}

protected void emptyJoinsPanel() {
    delCheckBoxPanel.removeAll();
    fromPanel.removeAll();
    relationPanel.removeAll();
    toPanel.removeAll();
}

/**
 * Handles add and delete buttons.
 */
public void actionPerformed(ActionEvent e) {
    String cmd = e.getActionCommand();
    if (cmd.equals(I18N.get("VisTableWin.add"))) {
	JoinFields jf = new JoinFields(null, report.getDatabase());
	joinFieldsList.add(jf);

	delCheckBoxPanel.add(jf.del);
	fromPanel.add(jf.from);
	relationPanel.add(jf.relation);
	toPanel.add(jf.to);

	joinsPanel.invalidate();
	pack();
    }
    else if (cmd.equals(I18N.get("VisTableWin.delete_selected"))) {
	ArrayList copy = (ArrayList)joinFieldsList.clone();
	for (Iterator iter = copy.iterator(); iter.hasNext(); ) {
	    JoinFields jf = (JoinFields)iter.next();
	    if (jf.del.isSelected()) {
		joinFieldsList.remove(jf);

		delCheckBoxPanel.remove(jf.del);
		fromPanel.remove(jf.from);
		relationPanel.remove(jf.relation);
		toPanel.remove(jf.to);
	    }
	}
	joinsPanel.invalidate();
	pack();
    }
    else
	super.actionPerformed(e);
}

protected void doSave() {
    ArrayList newJoins = new ArrayList();
    for (Iterator iter = joinFieldsList.iterator(); iter.hasNext(); ) {
	JoinFields jf = (JoinFields)iter.next();
	Join join = jf.join;

	Column from = columnFromDropdown(jf.from);
	String relation = (String)jf.relation.getSelectedItem();
	Column to = columnFromDropdown(jf.to);

	newJoins.add(new Join(from, relation, to));
    }

    TableJoinCommand cmd = new TableJoinCommand(query, newJoins);
    cmd.perform();
    commands.add(cmd);
}

protected Column columnFromDropdown(JComboBox cb) {
    String sel = (String)cb.getSelectedItem();
    return report.findColumn(sel);
}

protected void doRevert() {
    // Rebuild widgets
    emptyJoinsPanel();
    fillJoinsPanel();
    joinsPanel.invalidate();
    pack();
}

}

