package jimm.datavision.gui.cmd;
import jimm.datavision.*;
import jimm.datavision.gui.DesignWin;
import jimm.datavision.gui.FieldWidget;
import jimm.datavision.gui.SectionWidget;
import java.util.Iterator;

/**
 * A command for adding a subtotal to a field for a particular section.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class AbstractSubtotalCommand extends CommandAdapter {

protected Report report;
protected FieldWidget fieldWidget;
protected Group group;
protected FieldWidget subtotalWidget;

/**
 * Constructor.
 *
 * @param report the report containing the field and the subtotal
 * @param fw the field widget to which we are adding a subtotal
 * @param group if <code>null</code>, the subtotal is added to the
 * report footer; else the subtotal is added to the first section in
 * the group's footer.
 * @param name the command name
 */
public AbstractSubtotalCommand(Report report, FieldWidget fw, Group group,
			       String name)
{
    super(name);
    this.report = report;
    fieldWidget = fw;
    this.group = group;
}

public FieldWidget getSubtotalWidget() { return subtotalWidget; }

protected void createSubtotal() {
    // Find the proper footer. Don't worry; report and group footers always
    // have at least one section.
    Iterator footers = (group == null) ? report.footers() : group.footers();
    Section s = (Section)footers.next();

    Field originalField = fieldWidget.getField();
    Long id = originalField.getId();
//      if (originalField instanceof SubtotalField)	// Can't be true
//  	id = ((SubtotalField)originalField).getFieldId();

    // Create the subtotal field. Set the group (OK if null).
    SubtotalField subtotal =
	(SubtotalField)Field.create(null, report, s, "subtotal", id, true);
    subtotal.setGroup(group);
		
    // If we already have a subtotal widget (we already created one, or
    // this is a delete command and we are re-creating the deleted
    // subtotal), copy the format and bounds from that. Else, copy the
    // bounds and format from the field we are subtotaling and make the
    // format bold.
    Format fmt = null;
    Rectangle bounds = null;
    if (subtotalWidget == null) {
	fmt = originalField.getFormat();
	if (fmt == null) fmt = Format.DEFAULT;
	fmt = (Format)fmt.clone();
	fmt.setBold(true);

	bounds = new Rectangle(originalField.getBounds());
    }
    else {
	fmt = subtotalWidget.getField().getFormat();
	if (fmt == null) fmt = Format.DEFAULT;
	fmt = (Format)fmt.clone();

	bounds = new Rectangle(subtotalWidget.getField().getBounds());
    }
    subtotal.setFormat(fmt);
    subtotal.setBounds(bounds);

    // Add the field to the section
    s.addField(subtotal);

    // Create the widget and add it to the proper section widget.
    SectionWidget sectionWidget =
	DesignWin.findWindowFor(report).findSectionWidgetFor(s);

    subtotalWidget = new FieldWidget(null, subtotal);
    sectionWidget.addField(subtotalWidget);

    // For some reason we need to force a repaint of the section. Can't
    // just call invalidate.
    sectionWidget.repaint();
}

protected void deleteSubtotal() {
    java.awt.Component parent = subtotalWidget.getComponent().getParent();
    subtotalWidget.doDelete();
    parent.repaint();
}

}
