package jimm.datavision.layout.swing;
import jimm.datavision.*;
import jimm.datavision.gui.ExportWin;
import jimm.datavision.gui.DesignWin;
import jimm.datavision.gui.MenuUtils;
import jimm.datavision.gui.StatusDialog;
import java.io.*;
import java.awt.*;
import java.awt.print.*;
import java.awt.event.*;
import javax.swing.*;
import java.net.URL;
import java.util.ArrayList;
import java.util.Iterator;

/**
 * <code>SwingLE</code> is a layout engine that creates a Swing window.
 * The window can be printed by selecting the appropriate menu item.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 * @see SwingPageContents
 * @see SwingPage
 * @see SwingField
 * @see SwingPrintBook
 */
public class SwingLE extends LayoutEngine {

protected static final Dimension WINDOW_START_SIZE = new Dimension(600, 400);
protected static final String PRINT_ICON =
    "toolbarButtonGraphics/general/Print16.gif";
protected static final String EXPORT_ICON =
    "toolbarButtonGraphics/general/Export16.gif";
protected static final String FIRST_ICON =
    "toolbarButtonGraphics/navigation/Home16.gif";
protected static final String PREV_ICON =
    "toolbarButtonGraphics/navigation/Back16.gif";
protected static final String NEXT_ICON =
    "toolbarButtonGraphics/navigation/Forward16.gif";
protected static final String LAST_ICON =
    "toolbarButtonGraphics/navigation/Down16.gif";

protected JFrame frame;
protected Dimension pageDim;
protected JPanel cardPanel;
protected int displayPageNum;	// Starts at 1
protected ArrayList pageContents;
protected SwingPageContents pageBeingBuilt;
protected JLabel pageCountLabel;
protected Action printAction;
protected Action closeAction;
protected Action exportAction;
protected Action goFirstAction;
protected Action goPrevAction;
protected Action goNextAction;
protected Action goLastAction;

/**
 * Constructor.
 *
 * @param report the report to display.
 */
public SwingLE() {
    super();
    pageContents = new ArrayList();
}

public void cancel() {
    super.cancel();
    close();
}

/**
 * Creates window and displays a blank page.
 */
protected void doStart() {
    pageDim = new Dimension((int)pageWidth(), (int)pageHeight());
    frame = new JFrame(report.getTitle());
    makeActions();
    makeMenu(frame);
    frame.getContentPane().add(makeToolbar(), BorderLayout.NORTH);

    // Card panel for displaying pages
    cardPanel = new JPanel();
    cardPanel.setPreferredSize(WINDOW_START_SIZE);
    CardLayout cardLayout = new CardLayout(0, 0);
    cardPanel.setLayout(cardLayout);
    frame.getContentPane().add(cardPanel, BorderLayout.CENTER);

    // New, blank, dummy card we can display until the first page is
    // generated.
    JPanel blankPage = new JPanel();
    blankPage.setBackground(Color.white);
    blankPage.setPreferredSize(pageDim);
    cardPanel.add(blankPage, "blank page");
    cardLayout.show(cardPanel, "blank page");

    displayPageNum = 0;

    frame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
    frame.addWindowListener(new WindowAdapter() {
	public void windowClosing(WindowEvent e) {
	    close();
	}
	});

    frame.pack();
    frame.setVisible(true);
}

/**
 * Done loading report.
 */
protected void doEnd() {
    printAction.setEnabled(true);

    if (pageContents.size() > 1) // Start prebuilding the last page
	pageBeingBuilt.prebuildPage();

    // Calling displayPage() will force page 1 to finish building and
    // will display it.
    displayPage(1);
}

/**
 * Creates the actions used by menu items and toolbar widgets.
 */
protected void makeActions() {
    URL url = ClassLoader.getSystemResource(PRINT_ICON);
    String str = I18N.get("SwingLE.action_print");
    printAction = new AbstractAction(str, new ImageIcon(url, str)) {
	public void actionPerformed(ActionEvent e) { printReport(); }
	};
    printAction.putValue(Action.SHORT_DESCRIPTION, str);
    printAction.setEnabled(false);

    str = I18N.get("SwingLE.action_close");
    closeAction = new AbstractAction(str) {
	public void actionPerformed(ActionEvent e) { close(); }
	};

    url = ClassLoader.getSystemResource(EXPORT_ICON);
    str = I18N.get("SwingLE.action_export");
    exportAction = new AbstractAction(str, new ImageIcon(url, str)) {
	public void actionPerformed(ActionEvent e) { export(); }
	};
    exportAction.putValue(Action.SHORT_DESCRIPTION, str);

    url = ClassLoader.getSystemResource(FIRST_ICON);
    str = I18N.get("SwingLE.action_first_page");
    goFirstAction = new AbstractAction(str, new ImageIcon(url, str)) {
	public void actionPerformed(ActionEvent e) { displayFirstPage(); }
	};
    goFirstAction.putValue(Action.SHORT_DESCRIPTION, str);
    goFirstAction.setEnabled(false);

    url = ClassLoader.getSystemResource(PREV_ICON);
    str = I18N.get("SwingLE.action_previous_page");
    goPrevAction = new AbstractAction(str, new ImageIcon(url, str)) {
	public void actionPerformed(ActionEvent e) { displayPrevPage(); }
	};
    goPrevAction.putValue(Action.SHORT_DESCRIPTION, str);
    goPrevAction.setEnabled(false);

    url = ClassLoader.getSystemResource(NEXT_ICON);
    str = I18N.get("SwingLE.action_next_page");
    goNextAction = new AbstractAction(str, new ImageIcon(url, str)) {
	public void actionPerformed(ActionEvent e) { displayNextPage(); }
	};
    goNextAction.putValue(Action.SHORT_DESCRIPTION, str);
    goNextAction.setEnabled(false);

    url = ClassLoader.getSystemResource(LAST_ICON);
    str = I18N.get("SwingLE.action_last_page");
    goLastAction = new AbstractAction(str, new ImageIcon(url, str)) {
	public void actionPerformed(ActionEvent e) { displayLastPage(); }
	};
    goLastAction.putValue(Action.SHORT_DESCRIPTION, str);
    goLastAction.setEnabled(false);
}

/**
 * Creates the window menu.
 *
 * @param frame the window that will contain the menu
 */
protected void makeMenu(JFrame frame) {
    JMenuBar menuBar = new JMenuBar();
    frame.setJMenuBar(menuBar);

    // File menu
    JMenu menu = new JMenu(I18N.get("SwingLE.menu_file"));
    menu.setMnemonic(KeyEvent.VK_F);
    menuBar.add(menu);

    MenuUtils.setKeys(menu.add(printAction), KeyEvent.VK_P);
    menu.addSeparator();
    MenuUtils.setKeys(menu.add(closeAction), KeyEvent.VK_C, KeyEvent.VK_W);

    // View menu
    menu = new JMenu(I18N.get("SwingLE.menu_view"));
    menu.setMnemonic(KeyEvent.VK_V);
    menuBar.add(menu);

    MenuUtils.setKeys(menu.add(goFirstAction), KeyEvent.VK_F, KeyEvent.VK_1);
    MenuUtils.setKeys(menu.add(goPrevAction), KeyEvent.VK_P, KeyEvent.VK_2);
    MenuUtils.setKeys(menu.add(goNextAction), KeyEvent.VK_N, KeyEvent.VK_3);
    MenuUtils.setKeys(menu.add(goLastAction), KeyEvent.VK_L, KeyEvent.VK_4);

    // Report menu
    menu = new JMenu(I18N.get("SwingLE.menu_report"));
    menu.setMnemonic(KeyEvent.VK_R);
    menuBar.add(menu);

    MenuUtils.setKeys(menu.add(exportAction), KeyEvent.VK_E);
}

/**
 * Creates and returns a new tool bar.
 */
protected JToolBar makeToolbar() {
    JToolBar bar = new JToolBar(javax.swing.SwingConstants.HORIZONTAL);
    bar.add(printAction);
    bar.addSeparator();
    bar.add(exportAction);
    bar.addSeparator();
    bar.add(goFirstAction);
    bar.add(goPrevAction);
    bar.add(goNextAction);
    bar.add(goLastAction);
    pageCountLabel = new JLabel(I18N.get("SwingLE.loading_first_page"));
    bar.addSeparator();
    bar.add(pageCountLabel);
    return bar;
}

/**
 * Creates a new page.
 */
protected void doStartPage() {
    int pageNum = pageContents.size() + 1;
    pageBeingBuilt = new SwingPageContents(cardPanel, pageNum, pageDim);
    pageContents.add(pageBeingBuilt);
}

/**
 * At the end of the first page, starts building it in a separate thread.
 * Check to see if the first page is done being built. If so, it is
 * displayed.
 */
protected void doEndPage() {
    int numPages = pageContents.size();
    if (numPages == 1) {	// The first page
	// Start building the page in a separate thread
	pageBeingBuilt.prebuildPage();
	pageCountLabel.setText(I18N.get("SwingLE.building_first_page"));
    }

    // If the first page is done being built and is not yet displayed,
    // display it.
    if (((SwingPageContents)pageContents.get(0)).isPageBuilt()
	&& displayPageNum == 0)
	displayPage(1);		// Updates page count label and nav actions
    else {
	updatePageCountLabel();
	updateNavActions();
    }
}

/**
 * Closes this window. Does not call <code>System.exit</code>.
 */
public void close() {
    if (frame != null) {
	frame.setVisible(false);
	frame.dispose();
    }
    wantsMoreData = false;	// Signal report that we're done

    // Clean up memory a bit
    pageContents = null;
    pageBeingBuilt = null;
}

/**
 * Opens the report export dialog.
 */
protected void export() {
    new ExportWin(DesignWin.findWindowFor(report), report);
}

/**
 * Updates the navigation buttons based on number of pages and the current
 * display page.
 */
protected void updateNavActions() {
    int numPages = pageContents.size();

    boolean canGoBack = numPages > 0 && displayPageNum > 1;
    goFirstAction.setEnabled(canGoBack);
    goPrevAction.setEnabled(canGoBack);

    boolean canGoForward = numPages > 0 && displayPageNum < numPages;
    goNextAction.setEnabled(canGoForward);
    goLastAction.setEnabled(canGoForward);
}

/**
 * Updates the page count label based on the current page number and the
 * total number of pages.
 */
protected void updatePageCountLabel() {
    if (displayPageNum > 0)
	pageCountLabel.setText(I18N.get("SwingLE.page")
			       + ' ' + displayPageNum + ' '
			       + I18N.get("SwingLE.of")
			       + ' ' + pageContents.size());
}

/** Performs the "First Page" command. */
protected void displayFirstPage() {
    if (pageContents.size() > 0 && displayPageNum != 1)
	displayPage(1);
}

/** Performs the "Next Page" command. */
protected void displayNextPage() {
    if (pageContents.size() > displayPageNum)
	displayPage(displayPageNum + 1);
}

/** Performs the "Previous Page" command. */
protected void displayPrevPage() {
    if (displayPageNum > 1)
	displayPage(displayPageNum - 1);
}

/** Performs the "Last Page" command. */
protected void displayLastPage() {
    int numPages = pageContents.size();
    if (numPages > 0 && numPages != displayPageNum)
	displayPage(numPages);
}

/**
 * Fills the window with the contents of the specified page. If necessary,
 * builds the page. Starts building the next and previous pages in separate
 * threads, if they have not already been built.
 *
 * @param num page number, starting at 1
 * @see SwingPageContents
 */
protected void displayPage(int num) {
    if (num == displayPageNum)
	return;

    // Retrieve page. Build if necessary. Show it.
    SwingPageContents contents = (SwingPageContents)pageContents.get(num - 1);
    if (!contents.isPageBuilt())
	pageCountLabel.setText(I18N.get("SwingLE.building_page")
			       + ' ' + num + "...");
    contents.showPage();	// Builds page if necessary

    // If next page not already built, start building it in a separate
    // thread.
    int numPages = pageContents.size();
    if (numPages > num) { // Spawn thread to generate *next* page
	contents = (SwingPageContents)pageContents.get(num);
	if (!contents.isPageBuilt())
	    contents.prebuildPage();
    }

    // If previous page not already built, start building it in a separate
    // thread.
    if (num > 2) {		// We know pages 1 and 2 are already prebuilt
	contents = (SwingPageContents)pageContents.get(num - 2);
	if (!contents.isPageBuilt())
	    contents.prebuildPage();
    }

    displayPageNum = num;
    updatePageCountLabel();
    updateNavActions();
}

/**
 * Creates a new swing field and adds it to the current page.
 *
 * @param field the report field
 * @see SwingField
 */
protected void doOutputField(Field field) {
    String fieldAsString = field.toString();
    if (field.isVisible()
	&& fieldAsString != null && fieldAsString.length() > 0)
    {
	// Set bounds.
	jimm.datavision.Rectangle bounds = field.getBounds();

	// Page footers are anchored to page bottom.
	int y = (currentSectionType == SECT_PAGE_FOOTER)
	    ? (int)(pageHeight() - currentSection.getHeight() + bounds.y)
	    : (int)(pageHeightUsed + bounds.y);
	java.awt.Rectangle fieldBounds =
	    new java.awt.Rectangle((int)bounds.x, y, (int)bounds.width,
				   (int)bounds.height);
	pageBeingBuilt.add(field, fieldAsString, fieldBounds);
    }
}

protected void doOutputImage(ImageField image) {
    doOutputField(image);
}

/**
 * Creates a new line. <em>Unimplemented</em>.
 *
 * @param line a line
 */
protected void doOutputLine(Line line) {}

/**
 * Prints the report.
 */
protected void printReport() {
    final PrinterJob printJob = PrinterJob.getPrinterJob();
    PageFormat format = report.getPaperFormat().getPageFormat();
    SwingPrintBook book = new SwingPrintBook(pageContents, format);
    printJob.setPageable(book);

    if (printJob.printDialog()) {
	final StatusDialog statusDialog =
	    new StatusDialog(frame,
			     I18N.get("SwingLE.print_report_title"),
			     true,
			     I18N.get("SwingLE.print_report_status"));
	book.setStatusDialog(statusDialog);

	new Thread(new Runnable() {
	    public void run() {
		try {
		    printJob.print();
		}
		catch (UserCancellationException uce) {
		    printJob.cancel();
		}
		catch (Exception e) {
		    ErrorHandler.error(e);
		}
		finally {
		    if (statusDialog != null)
			statusDialog.dispose();
		}
	    }
	    }).start();
    }
}

}
