package jimm.datavision.sql;
import jimm.datavision.Nameable;
import jimm.datavision.Draggable;
import java.sql.*;

/**
 * A database column. It knows the table to which it belongs, its name,
 * and other metadata. The id of a column is a string of the form
 * "table_name.column_name".
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class Column implements Nameable, Draggable {

protected String id;
protected Table table;
protected String name;
protected int type;
protected String typeString;
protected int size;
protected boolean isNullable;
protected String defaultValue;
protected Object value;

/**
 * Constructor.
 *
 * @param table the table to which this column belongs
 * @param rset return value from <code>DatabaseMetaData.getColumns</code>
 * @see java.sql.DatabaseMetaData#getColumns
 */
public Column(Table table, ResultSet rset) throws SQLException {
    this.table = table;
    name = rset.getString("COLUMN_NAME");
    id = table.name + "." + name;
    type = rset.getInt("DATA_TYPE");

    // Some drivers don't support some of these values. Since we don't
    // actually use them, it's OK to ignore any errors.
    try {
	typeString = rset.getString("TYPE_NAME");
	size = rset.getInt("COLUMN_SIZE");
	isNullable =
	    !(rset.getInt("NULLABLE") == DatabaseMetaData.columnNoNulls);
	defaultValue = rset.getString("COLUMN_DEF");
    } catch (SQLException e) {}
}

/**
 * Returns the column's name. To retrieve the full "table.column" name,
 * use <code>fullName</code>.
 *
 * @return the column's name
 * @see #fullName
 */
public String getName() { return name; }

/** A column's name is immutable. */
public void setName(String name) { }

/**
 * Returns the type constant (a <code>java.sql.Types</code> value).
 *
 * @return a <code>java.sql.Types</code> value
 * @see java.sql.Types
 */
public int getType() { return type; }

/**
 * Returns the table id. This is a string of the form
 * "table_name.column_name".
 *
 * @return the id string
 */
public String getId() { return id; }

/**
 * Returns the table to which this column belongs.
 *
 * @return the table
 */
public Table getTable() { return table; }

/**
 * Returns the full named of this column. This is a string of the form
 * "table_name.column_name". To retrieve just the column name,
 * use <code>getName</code>.
 *
 * @return the full "table_name.column" name; this is the same as the
 * id string
 * @see #getName
 */
public String fullName() {
    return id;
}

public String dragString() {
    return "column:" + fullName();
}

/**
 * Returns a string representation of this column. Calls
 * <code>fullName</code>.
 *
 * @see #fullName
 */
public String toString() {
    return fullName();
}

}
