package jimm.datavision.sql;
import jimm.datavision.*;
import java.sql.*;
import java.util.*;

/**
 * A database knows about the tables in a database.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class Database implements Writeable {

protected static final String[] DB_OBJECT_TYPES = { "TABLE", "VIEW" };

protected Report report;
protected String driverClassName;
protected String connInfo;
protected String name;
protected String username;
protected TreeMap tables;
protected Connection conn;
private String password;
protected boolean hasPassword;

public Database() {}

/**
 * Constructor.
 *
 * @param driverClassName database driver class name
 * @param connInfo database connection info string
 * @param report the report using this database
 * @param name the database name
 * @param username the user name to use when logging in to the database
 * @param password the database password
 */
public Database(String driverClassName, String connInfo, Report report,
		String name, String user)
    throws SQLException, ClassNotFoundException, InstantiationException,
	   IllegalAccessException, UserCancellationException
{
    this(driverClassName, connInfo, report, name, user, null, false);
}

/**
 * Constructor.
 *
 * @param driverClassName database driver class name
 * @param connInfo database connection info string
 * @param report the report using this database
 * @param name the database name
 * @param username the user name to use when logging in to the database
 * @param password the database password
 */
public Database(String driverClassName, String connInfo, Report report,
		String name, String user, String password)
    throws SQLException, ClassNotFoundException, InstantiationException,
	   IllegalAccessException, UserCancellationException
{
    this(driverClassName, connInfo, report, name, user, password, true);
}


/**
 * Constructor.
 *
 * @param driverClassName database driver class name
 * @param connInfo database connection info string
 * @param report the report using this database
 * @param name the database name
 * @param username the user name to use when logging in to the database
 * @param password the database password
 * @param givenPassword if <code>true</code>, the password was passed in
 * to some other constructor
 */
protected Database(String driverClassName, String connInfo, Report report,
		   String name, String user, String password,
		   boolean givenPassword)
    throws SQLException, ClassNotFoundException, InstantiationException,
	   IllegalAccessException, UserCancellationException
{
    this.driverClassName = driverClassName;
    this.connInfo = connInfo;
    this.report = report;
    this.name = name;
    this.username = (user == null) ? "" : user;
    this.password = password;
    tables = null;
    hasPassword = givenPassword;
    initializeConnection();
    loadAllTables();
}

/**
 * Given an id (a table name), returns the table that has that id. If no
 * table with the specified id exists, returns <code>null</code>.
 *
 * @return a table, or <code>null</code> if no table with the specified
 * id exists
 */
public Table findTable(String id) {
    return tables == null ? null : (Table)tables.get(id.toLowerCase());
}

/**
 * Given an id (a column name), returns the column that has that id. If no
 * column with the specified id exists, returns <code>null</code>. Uses
 * <code>Table.findColumn</code>.
 *
 * @return acolumn, or <code>null</code> if no column with the specified
 * id exists
 * @see Table#findColumn
 */
public Column findColumn(String id) {
    if (tables == null)
	return null;

    for (Iterator iter = tables.values().iterator(); iter.hasNext(); ) {
	Column col = ((Table)iter.next()).findColumn(id);
	if (col != null)
	    return col;
    }
    return null;
}

public Row execute(Query query) throws SQLException {
    return new ResultSetRow(conn, query);
}

/**
 * Initializes the connection to the database.
 *
 * @return a connection to the database
 */
public void initializeConnection()
    throws ClassNotFoundException, InstantiationException,
	   IllegalAccessException, UserCancellationException
{
    // Keep trying until we succeed. We will only exit if we succeed,
    // if the user cancels, or if something bad happens while obtaining
    // the database connection.
    boolean ok = false;
    while (!ok) {
	if (!hasPassword) {
	    // Calling askForPassword sets our user name and password
	    report.askForPassword(this);
	    hasPassword = true;
	}
	if (password == null)
	    throw new UserCancellationException(I18N.get("Database.cancelled"));

	try {
	    if (connInfo == null || connInfo.length() == 0)
		throw new IllegalArgumentException(I18N.get("Database.missing_conn_info"));
	    if (username == null || username.length() == 0)
		throw new IllegalArgumentException(I18N.get("Database.missing_user_name"));
	    if (password == null)
		throw new IllegalArgumentException(I18N.get("Database.null_password"));

	    // Load the database JDBC driver
	    Driver d = (Driver)Class.forName(driverClassName).newInstance();
	    DriverManager.registerDriver(d);

	    // Connect to the database
	    conn = DriverManager.getConnection(connInfo, username, password);

	    ok = true;
	}
	catch (SQLException sqle) {	// Force another password request
	    ErrorHandler.error(sqle);
	    hasPassword = false;
	    report.setDatabasePassword(null);
	}
    }
}


/**
 * Returns a connection to the database.
 *
 * @return a connection to the database
 */
public Connection getConnection() { return conn; }

/**
 * Reset key instance variables, closes current connection, and "reloads"
 * all table information (compares new info with existing info and complains
 * if any existing info is not in the new info).
 *
 * @param driverClassName database driver class name
 * @param connInfo database connection info string
 * @param dbName database name
 * @param username the user name to use when logging in to the database
 */
public void reset(String driverClassName, String connInfo, String dbName,
		  String username, String password)
    throws SQLException, ClassNotFoundException, InstantiationException,
	   IllegalAccessException, UserCancellationException
{
    setDriverClassName(driverClassName);
    setConnectionInfo(connInfo);
    setName(dbName);
    setUserName(username);
    this.password = password;
    hasPassword = true;

    if (conn != null) {
	conn.close();
	conn = null;
    }
    initializeConnection();
    loadAllTables();
}

/**
 * Loads information about all tables in the database. If no tables are
 * found when using the database schema name, try again with a
 * <code>null</code> schema name.
 */
protected void loadAllTables() throws SQLException {
    tables = new TreeMap();

    DatabaseMetaData dbmd = getConnection().getMetaData();

    try {
	// Specify both schema name and DB_OBJECT_TYPES.
	loadTablesUsingSchemaNameAndTypes(dbmd, name, DB_OBJECT_TYPES);
    } catch (SQLException e) {}
    catch (NullPointerException npe) {}

    try {
	if (tables.isEmpty() && name != null) // Only schema name
	    loadTablesUsingSchemaNameAndTypes(dbmd, name, null);
    } catch (SQLException e) {}
    catch (NullPointerException npe) {}

    try {
	if (tables.isEmpty() && name != null) // No schema name, use types
	    loadTablesUsingSchemaNameAndTypes(dbmd, null, DB_OBJECT_TYPES);
    } catch (SQLException e) {}
    catch (NullPointerException npe) {}

    // If no tables found, try again with null database name. This time,
    // throw an exception if there is a problem.
    if (tables.isEmpty() && name != null) // No schema name, no types
	loadTablesUsingSchemaNameAndTypes(dbmd, null, null);
}

/**
 * Loads our list of tables using a database meta data object and a
 * schema name. The schema name may be <code>null</code>.
 *
 * @param dbmd the database meta data object
 * @param objectTypes a list of database object types
 * @param schema the schema name; may be <code>null</code>
 */
protected void loadTablesUsingSchemaNameAndTypes(DatabaseMetaData dbmd,
						 String schema,
						 String[] objectTypes)
    throws SQLException
{
    ResultSet rset = dbmd.getTables(null, schema, "%", objectTypes);
    if (rset == null) return;
    while (rset.next()) {
	String name = rset.getString("TABLE_NAME");
	Table t = new Table(this, dbmd, name);
	tables.put(t.getId().toLowerCase(), t);
    }
    rset.close();
}

/**
 * Returns the driver class name.
 *
 * @return the driver class name
 */
public String getDriverClassName() { return driverClassName; }

/**
 * Sets the driver class name.
 *
 * @param newDriverClassName the driver class name
 */
protected void setDriverClassName(String newDriverClassName) {
    driverClassName = newDriverClassName;
}

/**
 * Returns the connection info string.
 *
 * @return the connection info string
 */
public String getConnectionInfo() { return connInfo; }

/**
 * Sets the connection info string.
 *
 * @param newConnectionInfo the connection info string
 */
protected void setConnectionInfo(String newConnectionInfo) {
    connInfo = newConnectionInfo;
}

/**
 * Returns the database name.
 *
 * @return the database name
 */
public String getName() { return name; }

/**
 * Sets the name.
 *
 * @param newName the new name
 */
protected void setName(String newName) { name = newName; }

/**
 * Returns the user name.
 *
 * @return the user name
 */
public String getUserName() { return username; }

/**
 * Sets the user name.
 *
 * @param newUserName the new user name
 */
public void setUserName(String newUserName) { username = newUserName; }

/**
 * Returns the password.
 *
 * @return the password
 */
public String getPassword() { return hasPassword ? password : null; }

/**
 * Sets the password.
 *
 * @param newPassword the new password
 */
public void setPassword(String newPassword) {
    password = newPassword;
    hasPassword = true;
}

/**
 * Returns an iterator over all the tables in the database.
 *
 * @return an iterator over all the tables in the database
 */
public Iterator tables() { return tables.values().iterator(); }

/**
 * Writes this database and all its tables as an XML tag.
 *
 * @param out a writer that knows about indenting XML tags
 */
public void writeXML(IndentWriter out) {
    out.println("<database driverClassName=\"" + driverClassName + '"');
    out.indent();
    out.indent();
    out.println("connInfo=\"" + connInfo + '"');
    out.println("name=\"" + (name == null ? "" : name) + "\" username=\""
		+ username + "\" />");
    out.outdent();
    out.outdent();
}

}
