package jimm.datavision.sql;
import jimm.datavision.ErrorHandler;
import jimm.datavision.Nameable;
import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.TreeMap;
import java.util.Iterator;

/**
 * Represents a database table.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */
public class Table implements Nameable {

protected Database database;
protected String name;
protected TreeMap columns;
protected boolean usedInReport;

/**
 * Given an id (a column name), returns the column that has that id. If no
 * column with the specified id exists, returns <code>null</code>.
 *
 * @return acolumn, or <code>null</code> if no column with the specified
 * id exists
 */
public Column findColumn(String id) {
    return columns == null ? null : (Column)columns.get(id.toLowerCase());
}

/**
 * Constructor.
 *
 * @param database the database in which this table resides
 * @param dbmd database meta data (used to read column info)
 * @param name the table's name
 */
public Table(Database database, DatabaseMetaData dbmd, String name)
    throws SQLException
{
    this.database = database;
    this.name = name;
    usedInReport = false;
    columns = null;
    loadColumns(dbmd);
}

/**
 * Returns the table id. It's the same as the table name.
 *
 * @return the table id
 */
public String getId() { return name; }

/**
 * Returns the table name.
 *
 * @return the table name
 */
public String getName() { return name; }

/** A table's name is immutable. */
public void setName(String name) { }

/**
 * Flags this table as being used in the database or not.
 *
 * @param isUsed a boolean
 */
public void setUsedInReport(boolean isUsed) { usedInReport = isUsed; }

/**
 * Returns an iterator over the columns in this table.
 *
 * @return an iterator over the columns in this table
 */
public Iterator columns() { return columns.values().iterator(); }

/**
 * Loads all columns in the table.
 *
 * @param dbmd database meta data
 */
void loadColumns(DatabaseMetaData dbmd) throws SQLException {
    columns = new TreeMap();

    ResultSet rset = dbmd.getColumns(null, null, name, "%");
    while (rset.next()) {
	Column c = new Column(this, rset);
	columns.put(c.getId().toLowerCase(), c);
    }
    rset.close();
}

}
