package jimm.datavision.test;
import jimm.datavision.*;
import jimm.datavision.sql.Column;
import jimm.datavision.layout.CharSepLE;
import java.io.*;
import java.util.Date;
import java.sql.SQLException;
import java.text.DecimalFormat;
import java.text.SimpleDateFormat;
import junit.framework.TestCase;
import junit.framework.TestSuite;
import junit.framework.Test;
import org.xml.sax.SAXException;

/**
 * Reads a report from an XML file, runs it, and verifies the output. Uses
 * the <code>CharSepLE</code> layout engine to produce a tab-delimited
 * output file.
 * <p>
 * These tests are tightly coupled with the contents of the file
 * <tt>test.xml</tt> and the contents of the test database generated
 * by the files in <tt>jimm/datavision/testdata</tt>.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 * @see jimm.datavision.layout.CharSepLE
 */
public class ReportRunTest extends TestCase {

protected static final String EXAMPLE_REPORT =
    "jimm/datavision/test/test.xml";
protected static final String PARAMETER_XML_FILE_NAME =
    "jimm/datavision/test/test_parameters.xml";
protected static final String OUT_FILE = "/tmp/datavision_test_out.txt";
// This must match the format string for the report.date field in
// EXAMPLE_REPORT.
protected static final String REPORT_DATE_FORMAT = "yyyy-MM-dd";
// This must be an alphabetically sorted list of office names from the
// offices table.
protected static final String OFFICES[] = {
    "Chicago", "New Jersey", "New York"
};

protected Report report;
protected DecimalFormat dollarFormatter;
protected DecimalFormat lastColFormatter;
protected SimpleDateFormat titleDateFormatter;

public static Test suite() {
    return new TestSuite(ReportRunTest.class);
}

public ReportRunTest(String name) {
    super(name);
}

public void setUp()
    throws FileNotFoundException, IOException, SQLException,
	   ClassNotFoundException, InstantiationException,
	   IllegalAccessException, SAXException
{
    dollarFormatter = new DecimalFormat("$#,###.00");
    lastColFormatter = new DecimalFormat("#,###.##");
    titleDateFormatter = new SimpleDateFormat(REPORT_DATE_FORMAT);

    report = new Report();
    report.setDatabasePassword("");

    File f = new File(OUT_FILE);
    f.deleteOnExit();
    PrintWriter out = new PrintWriter(new FileWriter(f));
    report.setLayoutEngine(new CharSepLE(out, '\t'));

    report.readFile(EXAMPLE_REPORT); // Must come after setting password
}

public void tearDown() {
    new File(OUT_FILE).delete();
}

public void testReportRun() throws IOException, FileNotFoundException {
    // Run report in this thread, not a separate one. Running the
    // report closes the output stream.
    report.runReport();

    // Open the output and look for various things.
    BufferedReader in = new BufferedReader(new FileReader(OUT_FILE));

    expectHeaders(in);

    // Each of the office groups. checkOfficeGroup() returns the total
    // dollar amount (really the total id number amount, but that is proved
    // to be the same thing within checkDetailLine()).
    int total = 0;
    for (int i = 0; i < OFFICES.length; ++i)
	total += checkOfficeGroup(in, OFFICES[i]);

    // The grand total.
    String line = in.readLine();
    assertNotNull(line);
    if (line.startsWith("Page "))
	assertNotNull(line = in.readLine());

    assertEquals("Grand Total:\t" + dollarFormatter.format(total), line);

    // The final page number.
    assertNotNull(line = in.readLine());
    assertEquals("Page ", line.substring(0, 5));

    // Make sure we are at the end of the file.
    assertNull(in.readLine());

    in.close();
}

protected void expectHeaders(BufferedReader in) throws IOException {
    String line;

    // Line 1: report title and formatted date.
    // since that will most definitely be different.
    assertNotNull(line = in.readLine());
    assertEquals(0, line.indexOf("Example Report\t"
				 + titleDateFormatter.format(new Date())));

    // Line 2: the page header.
    assertNotNull(line = in.readLine());
    assertEquals(0, line.indexOf("Job #\tTitle\tHourly Rate"));
}

/**
 * Checks a group (header plus detail lines) and returns the total
 * of the dollar amounts in the group.
 *
 * @param in the input reader
 * @param officeName the group name
 * @return the total dollar amount in the group
 */
protected int checkOfficeGroup(BufferedReader in, String officeName)
    throws IOException
{
    String line;
    int subtotal = 0;

    // The office name. Skip page delimiters ("Page" at foot and
    // page headers at top).
    assertNotNull(line = in.readLine());
    while (line.startsWith("Page ") || line.startsWith("Job #\t"))
	assertNotNull(line = in.readLine());
    assertEquals(officeName, line);

    Object postDateGroupEnd;
    while ((postDateGroupEnd = checkPostDateGroup(in)) instanceof Integer)
	subtotal += ((Integer)postDateGroupEnd).intValue();
    line = (String)postDateGroupEnd;

    assertEquals("Total:\t" + dollarFormatter.format(subtotal), line);

    return subtotal;
}

/**
 * Checks a subgroup (header plus detail lines) and returns the total
 * of the dollar amounts in the group. If the next report section is
 * not a subgroup, it will be a group total line; we return the line
 *
 * @param in the input reader
 * @param officeName the group name
 * @return either an Integer containing the total or the next line read
 * from the report
 */
protected Object checkPostDateGroup(BufferedReader in) throws IOException {
    int subtotal = 0;
    String line;

    // Read either post date or next (super)group name. Return false if we
    // see a non-date group name; that means we're done. Skip page
    // delimiters ("Page" at foot and page headers at top).
    while (true) {
	assertNotNull(line = in.readLine());
	if (line.startsWith("Page ") || line.startsWith("Job #\t"))
	    continue;
	if (line.startsWith("Total:"))
	    return line;

	// Primitive date check
	assertTrue(Character.isDigit(line.charAt(0)) && line.length() == 10
		   && line.charAt(4) == '-' && line.charAt(7) == '-');
	break;
    }

    // The detail lines. Collect the id number from the beginning of each
    // detail line. It is the same as the dollar amount value. Add that
    // amount to the subtotal.
    while (true) {
	assertNotNull(line = in.readLine());
	if (line.startsWith("Post Date Total:"))
	    break;
	if (line.startsWith("Page ") || line.startsWith("Job #\t"))
	    continue;

	subtotal += checkDetailLine(line);
    }
    assertEquals("Post Date Total:\t" + dollarFormatter.format(subtotal),
		 line);

    return new Integer(subtotal);
}

/**
 * Checks the format of a detail line and returns the integer id found
 * at the beginning of line. This is the same as the dollar amount.
 *
 * @param line a detail line
 * @return the id number
 */
protected int checkDetailLine(String line) {
    // Read job id from the beginning of the line.
    int id = Integer.parseInt(line.substring(0, line.indexOf("\t")));

    // The description, at random, may be repeated ("job 3 job 3"). We
    // check for the first full description.
    String str = "" + id + "\tThis is the short description of job " + id;
    assertTrue("line does not start with \"" + str + "\"",
	       line.startsWith(str));

    // Check for the remaining columns at the end of the line. We are also
    // assuring that the dollar amount is the same as the id number.
    str = "\t" + (id * 100) + "\t" + dollarFormatter.format(id) + "\t"
	+ lastColFormatter.format(id);
    assertTrue("line does not end with \"" + str + "\"", line.endsWith(str));

    return id;
}

public void testWhereClause() throws IOException, FileNotFoundException {
    // Set where clause
    report.getQuery()
	.setEditableWhereClause("office.name = '" + OFFICES[0] + "'");

    // Run report in this thread, not a separate one. Running the
    // report closes the output stream.
    report.runReport();

    // Open the output and look for various things.
    BufferedReader in = new BufferedReader(new FileReader(OUT_FILE));

    expectHeaders(in);

    // The single office group we asked for. checkOfficeGroup() returns
    // the total dollar amount (really the total id number amount, but
    // that is proved to be be the same thing within checkDetailLine()).
    int total = checkOfficeGroup(in, OFFICES[0]);

    // The grand total.
    String line = in.readLine();
    assertNotNull(line);
    if (line.startsWith("Page "))
	assertNotNull(line = in.readLine());

    assertEquals("Grand Total:\t" + dollarFormatter.format(total), line);

    in.close();
}

public void testParamInWhereClause() throws IOException, FileNotFoundException
{
    // Set where clause and parameter value file
    report.getQuery().setEditableWhereClause("office.name = {?String Param}");
    report.setParameterXMLFile(PARAMETER_XML_FILE_NAME);

    // Run report in this thread, not a separate one. Running the
    // report closes the output stream.
    Date now = new Date();	// Get date as near report run time as possible
    report.runReport();

    // Open the output and look for various things.
    BufferedReader in = new BufferedReader(new FileReader(OUT_FILE));
    String line;

    // Line 1. Look for report title and the date (ignore the time,
    // since that will most definitely be different.
    assertNotNull(line = in.readLine());

    // There is an infinitessimally small chance the next assertion will
    // fail even though the report is correct. Imagine the date "now"
    // retrieved above is created just before midnight. Further imagine
    // that the report.date special field's value isn't retrieved until
    // just after midnight. Not terribly likely, but I'm in a CYA mood.
    assertEquals(0, line.indexOf("Example Report\t"
				 + titleDateFormatter.format(now)));

    // Line 2: the page header.
    assertNotNull(line = in.readLine());
    assertEquals(0, line.indexOf("Job #\tTitle\tHourly Rate"));

    // The single office group we asked for. checkOfficeGroup() returns
    // the total dollar amount (really the total id number amount, but
    // that is proved to be be the same thing within checkDetailLine()).
    int total = checkOfficeGroup(in, OFFICES[0]);

    // The grand total.
    assertNotNull(line = in.readLine());
    if (line.startsWith("Page "))
	assertNotNull(line = in.readLine());

    assertEquals("Grand Total:\t" + dollarFormatter.format(total), line);

    in.close();
}

public static void main(String[] args) {
    junit.textui.TestRunner.run(suite());
}

}
