package jimm.datavision.test;
import jimm.datavision.*;
import jimm.datavision.sql.Column;
import jimm.datavision.layout.CharSepLE;
import java.io.*;
import java.util.Date;
import java.util.List;
import java.util.Iterator;
import java.sql.SQLException;
import java.text.DecimalFormat;
import junit.framework.TestCase;
import junit.framework.TestSuite;
import junit.framework.Test;
import org.xml.sax.SAXException;

/**
 * Reads a report from an XML file, tests its structure, and tests various
 * pieces like parameter and formula substitution.
 *
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 * @see jimm.datavision.layout.CharSepLE
 */
public class ReportTest extends TestCase {

protected static final String EXAMPLE_REPORT =
    "jimm/datavision/test/test.xml";
protected static final String PARAMETER_XML_FILE_NAME =
    "jimm/datavision/test/test_parameters.xml";
//  protected static final String REPORT_FILE = "/tmp/datavision_test_report.xml";

protected Report report;
protected DecimalFormat dollarFormatter;
protected DecimalFormat lastColFormatter;

public static Test suite() {
    return new TestSuite(ReportTest.class);
}

public ReportTest(String name) {
    super(name);
}

public void setUp()
    throws FileNotFoundException, IOException, SQLException,
	   ClassNotFoundException, InstantiationException,
	   IllegalAccessException, SAXException
{
    report = new Report();
    report.setDatabasePassword("");
    report.readFile(EXAMPLE_REPORT); // Must come after setting password
    report.setParameterXMLFile(PARAMETER_XML_FILE_NAME);
}

public void tearDown() {
    report = null;
}

public void testReportStructure() {
    // Attributes
    assertEquals("example1", report.getName());
    assertEquals("Example Report", report.getTitle());
    assertEquals("Jim Menard", report.getAuthor());
    assertEquals(new Long(7), report.generateNewParameterId());

    // Paper format
    PaperFormat pf = report.getPaperFormat();
    assertNotNull(pf);
    assertEquals(PaperFormat.PORTRAIT, pf.getOrientation());
    assertEquals("US-Letter", pf.getName());

    // Columns
    Column col = report.findColumn("jobs.id");
    assertNotNull(col);
    assertEquals("jobs.id", col.getId());

    // Formulas
    Formula formula = report.findFormula("1");
    Long id = new Long(1);
    assertNotNull(formula);
    assertEquals("hourly rate / 100", formula.getName());
    assertEquals("{jobs.hourly_rate} 100.0 /", formula.getEvalString());
    assertEquals("formula:1", formula.dragString());
    assertEquals("{@hourly rate / 100}", formula.designLabel());
    assertEquals("{@1}", formula.formulaString());
    // We test Utils.formulaToDisplay() and Utils.displayToFormula()
    // in FormulaTest.java.

    formula = report.findFormula(id);
    assertNotNull(formula);
    assertEquals(id, formula.getId());

    // Sections
    assertEquals(1, report.testCountHeaders());
    assertEquals(1, report.testCountFooters());
    assertEquals(1, report.testCountPageHeaders());
    assertEquals(1, report.testCountPageFooters());
    assertEquals(2, report.countGroups());
    assertEquals(1, report.testCountDetails());

    // Add a new detail section
    Section detail = report.getFirstSectionByArea(Report.AREA_DETAIL);
    assertNotNull(detail);
    report.insertSectionBelow(detail);
    assertEquals(2, report.testCountDetails());
}

public void testParameters() {
    Parameter param = report.findParameter("5");
    Long id = new Long(5);
    assertNotNull(param);
    assertEquals("parameter:5", param.dragString());
    assertEquals("{?5}", param.formulaString());
    assertEquals("{?Yes/No}", param.designLabel());
    assertEquals(Parameter.TYPE_BOOLEAN, param.getType());
    assertEquals(Parameter.ARITY_ONE, param.getArity());
    assertEquals(id, param.getId());
    assertEquals("Yes/No", param.getName());
    assertEquals("Do you breathe regularly?", param.getQuestion());

    param = report.findParameter(id);
    assertNotNull(param);
    assertEquals(id, param.getId());

    // Get ready to read values from parameter values file when asked.
    // We have to add a parameter to the report somewhere. If we don't,
    // the report won't read the XML file.
    report.getQuery().setEditableWhereClause("office.name = {?String Param}");

    id = new Long(2);
    param = report.findParameter(id);
    assertNotNull(param);

    Object obj = report.getParameterValue(id); // Reads values from XML
    assertNotNull("range param value should not be null", obj);

    assertTrue("range param value is not a list; class = "
	       + obj.getClass().getName(),
	       obj instanceof List);
    List list = (List)obj;

    assertTrue("range param value list does not have length 2; it has length "
	       + list.size(), list.size() == 2);

    assertEquals(new Integer(50), list.get(0));
    assertEquals(new Integer(75), list.get(1));
}

public void testNullFieldIds() {
    // Empty reports start with one detail section
    Section detail = report.getFirstSectionByArea(Report.AREA_DETAIL);
    assertNotNull(detail);

    // Add fields.
    TextField f = new TextField(null, report, detail, "field 1", true);
    assertNotNull(f.getId());
    Long id = f.getId();
    assertTrue("field id is not > 1", id.longValue() >= 1);
    detail.addField(f);

    f = new TextField(null, report, detail, "field 2", true);
    assertNotNull(f.getId());
    assertEquals(new Long(id.longValue() + 1), f.getId());
    id = f.getId();
    detail.addField(f);

    f = new TextField(null, report, detail, "field 3", true);
    assertNotNull(f.getId());
    assertEquals(new Long(id.longValue() + 1), f.getId());
    detail.addField(f);
}

protected void preparedQueryTest(Query q, String whereClause, String answer) {
    q.setEditableWhereClause(whereClause);
    String sql = q.toPreparedStatementString();
    assertTrue("expected \"" + answer + "\" in where clause; sql = " + sql,
	       sql.indexOf(answer) >= 0);
}

public void testQueryRangeSubstitution() {
    String answer = "jobs.id between ? and ?";
    String notAnswer = "jobs.id not between ? and ?";

    Query q = report.getQuery();
    preparedQueryTest(q, "jobs.id in {?Number Range}", answer);
    preparedQueryTest(q, "jobs.id in{?Number Range}", answer);
    preparedQueryTest(q, "jobs.id between {?Number Range}", answer);
    preparedQueryTest(q, "jobs.id = {?Number Range}", answer);
    preparedQueryTest(q, "jobs.id={?Number Range}", answer);
    preparedQueryTest(q, "jobs.id ={?Number Range}", answer);
    preparedQueryTest(q, "jobs.id= {?Number Range}", answer);

    preparedQueryTest(q, "jobs.id not in {?Number Range}", notAnswer);
    preparedQueryTest(q, "jobs.id not in{?Number Range}", notAnswer);
    preparedQueryTest(q, "jobs.id not between {?Number Range}", notAnswer);
    preparedQueryTest(q, "jobs.id != {?Number Range}", notAnswer);
    preparedQueryTest(q, "jobs.id!={?Number Range}", notAnswer);
    preparedQueryTest(q, "jobs.id !={?Number Range}", notAnswer);
    preparedQueryTest(q, "jobs.id!= {?Number Range}", notAnswer);
    preparedQueryTest(q, "jobs.id <> {?Number Range}", notAnswer);
    preparedQueryTest(q, "jobs.id<>{?Number Range}", notAnswer);
}

public void testQueryListSubstitution() {
    String answer = "jobs.id in (?,?,?)";
    String notAnswer = "jobs.id not in (?,?,?)";

    Query q = report.getQuery();
    preparedQueryTest(q, "jobs.id in {?Pick One}", answer);
    preparedQueryTest(q, "jobs.id = {?Pick One}", answer);
    preparedQueryTest(q, "jobs.id={?Pick One}", answer);
    preparedQueryTest(q, "jobs.id not in {?Pick One}", notAnswer);
    preparedQueryTest(q, "jobs.id != {?Pick One}", notAnswer);
    preparedQueryTest(q, "jobs.id <> {?Pick One}", notAnswer);
}

public void testQueryDateParam() {
    String answer = "jobs.post_date >= ?";

    Query q = report.getQuery();
    preparedQueryTest(q, "jobs.post_date >= {?Date}", answer);
}

public void testQueryUserColPrep() {
    String answer = "substr(jobs.title, 1, 8)";
    UserColumn uc = new UserColumn(null, report, "my user col",
				   "substr({jobs.title}, 1, 8)");
    Query q = report.getQuery();
    assertEquals(answer, q.userColumnForPreparedStatement(uc));
}

public void testCaseInsensitivity() {
    assertNotNull(report.findTable("jobs"));
    assertNotNull(report.findTable("JOBS"));
    assertNotNull(report.findTable("jOBs"));

    assertNotNull(report.findColumn("jobs.id"));
    assertNotNull(report.findColumn("JOBS.ID"));
    assertNotNull(report.findColumn("jOBs.id"));

    assertNotNull(report.findFormulaByName("refs f1"));
    assertNotNull(report.findFormulaByName("REFS F1"));
    assertNotNull(report.findFormulaByName("rEFs f1"));

    assertNotNull(report.findParameterByName("Yes/No"));
    assertNotNull(report.findParameterByName("yes/no"));
    assertNotNull(report.findParameterByName("yES/No"));
}

public void testFormat() {
    Format f = new Format();
    Format f2 = Format.DEFAULT;
    assertTrue("New and DEFAULT formats should be equal", f.equals(f2));

    assertNotNull(f.getColor());

    Field field = Field.createFromDragString(report, "column:jobs.id");
    assertNotNull(field);
    assertNull(field.getFormat());
}

public void testCloning() {
    // Make sure whatever we create here does not have a value that
    // depends upon running the report and having data.
    Section s = report.getFirstSectionByArea(Report.AREA_REPORT_HEADER);
    SpecialField f =
	(SpecialField)Field.create(null, report, s, "special",
				   "report.date", true);

    f.setFormat(new Format());
    f.setBounds(new Rectangle(1.0, 2.0, 3.0, 4.0));
    f.setBorder(new Border(f));

    SpecialField f2 = (SpecialField)f.clone();
    assertNotNull(f2);

    // assertEquals does too much, and we don't need Field.equals
    // anywhere else, so let's do it piece by piece

    assertEquals(f.getBounds().x, f2.getBounds().x, 0.00001);
    assertEquals(f.getBounds().y, f2.getBounds().y, 0.00001);
    assertEquals(f.getBounds().width, f2.getBounds().width, 0.00001);
    assertEquals(f.getBounds().height, f2.getBounds().height, 0.00001);
    assertEquals(f.getFormat(), f2.getFormat());
    assertEquals(f.getBorder(), f2.getBorder());
    assertEquals(f.getValue(), f2.getValue());
}

public static void main(String[] args) {
    junit.textui.TestRunner.run(suite());
}

}
