package jimm.datavision.testdata;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import uk.co.wilson.xml.MinML;
import org.xml.sax.*;

/**
 * Skeleton for creating a schema.sql file by reading an XML description
 * of a schema.
 * 
 * @author Jim Menard, <a href="mailto:jimm@io.com">jimm@io.com</a>
 */

public abstract class SchemaGen extends MinML {

protected String tableName;
protected String columnName;
protected String type;
protected int size;
protected boolean notNull;
protected boolean isFirstColumn;

public void run(String schemaXMLFile) {
    try {
	parse(new FileReader(schemaXMLFile));
    }
    catch (Exception e) {
	e.printStackTrace();
	System.exit(1);
    }
}

public void startElement(String name, AttributeList attributes) {
    if ("table".equals(name)) table(attributes);
    else if ("column".equals(name)) column(attributes);
}

public void endElement(String name) {
    if ("table".equals(name)) endTable();
}

/**
 * Parses a table XML tag and calls <code>makeTable</code>.
 *
 * @param line a line of XML
 * @see #makeTable
 */
protected void table(AttributeList attributes) {
    tableName = attributes.getValue("name");
    makeTable();
    isFirstColumn = true;
}

/**
 * Parses a column XML tag and calls <code>printColumn</code>. Also handles
 * commas and indentation.
 *
 * @param line a line of XML
 * @see #makeTable
 */
protected void column(AttributeList attributes) {
    columnName = attributes.getValue("name");
    type = attributes.getValue("type");
    String sizeStr = attributes.getValue("size");
    if (sizeStr == null)
	size = 0;
    else
	size = Integer.parseInt(sizeStr);
    notNull = (attributes.getValue("not-null") != null);

    if (!isFirstColumn)
	System.out.print(",");
    System.out.println();
    System.out.print("\t");

    printColumn();
    isFirstColumn = false;
}

/**
 * Outputs the SQL needed to create a database table. Optionally prints
 * the SQL needed to destroy the table first.
 */
protected abstract void makeTable();

/**
 * Outputs the SQL needed to close a create table statement.
 */
protected abstract void endTable();

/**
 * Prints the SQL needed to create a database column within a create table
 * statement.
 */
protected void printColumn() {
    System.out.print(columnName);
    printType();
    if (notNull) printNotNull();
}

/**
 * Prints the SQL that defines a column's type.
 */
protected abstract void printType();

/**
 * Prints the SQL that defines a column as NOT NULL.
 */
protected abstract void printNotNull();

}
